function blaze_function, lambda_in, m, k_emp=k_emp_in, lambda_offset=lambda_offset_in, $
                         alpha_emp=alpha_emp_in, lambda_peak=lambda_peak_in, silicon=silicon, $
                         modfunction=modfuction_in, instrument_type=instrument_type_in, $
                         angle_blaze=angle_blaze_in, angle_in=angle_in_in, $
                         angle_reflect=angle_reflect_in, angle_out=angle_out_in, $
                         groove_width=groove_width_in, print=print, lambda_corr=lambda_corr_in, output
  
  
;BLAZE_FUNCTION.PRO: calculate the blaze function for echelle
;                    gratings, following formula 1 in Barker et
;                    al. (1984) and formula in
;                    esomidas_1999_midasusermanual.pdf This function
;                    does not include instrument specifics, other than
;                    the type of instrument (Littrow, pseudo-Littrow,
;                    etc,).
;
;
; out=blaze_function(
;
;                    lambda_in FLTARR(n) wavelength in micron
;
;                    m INTEGER echelle order
;
;                    /silicon KEYWD include wavelength-dependent
;                                   optical constants of silicon for
;                                   immersion grating [DEFAULT:
;                                   grating in air or vacuo, i.e.,
;                                   optical constant is 1.0]
;
;                    instrument_type=.. STRING what kind of instrument
;                                    is this for? Currrently only
;                                    pseudo-Littrow is implemented. See
;                                    a good description of different
;                                    designs in John Rayner's "iSHELL
;                                    Optical Design Document" [DEFAULT:
;                                    'pseudo-littrow']
;
;                    /print KEYWD show used parameters on IDL window
;                                 [DEFAULT: do not print parameters]
;
;                    output=.. FLTARR() return the variables used in
;                                 the calculated blaze function:
;                                 [m, lambda_peak, alpha]
;                                 [DEFAULT: none]
;
;                    The user provides either empirical grating
;                    constants, or theoretical ones:
;
;                    EMPIRICAL CONSTANTS:
;
;                    k_emp FLOAT empirical grating constant slope
;                              without taking into account optical
;                              constant for immersion grating because
;                              that's wavelength dependent, while
;                              k_emp should not be.  Lambda_peak
;                              calculated as lambda_offset+k_emp*n/m
;                              [not sure why this isn't simply
;                              k_emp*n/m]
;
;                    lambda_offset FLOAT empirical grating constant
;                              offset
;                              lambda_peak=lambda_offset+k_emp/m
;                              [unclear to me why needed]
;
;                    alpha_emp FLOAT empirical grating constant
;                                usually between 0.5 and 1.0 (?) that
;                                sets the width of the blaze
;                                function. It is defined as
;                                (l/d)*COS(gamma) with l=grating facet
;                                width 1/d=spatial frequency of facets
;                                gamma=blaze angle.  [DEFAULT:
;                                calculated as (COS(angle_blaze))^2]
;
;                    lambda_corr FLTARR(2) polynomial coefficients of a
;                                correction function to multiply
;                                theoretical lambda with as function
;                                of 1/m
;                                [DEFAULT: 1., i.e., no such correction]
;
;                    lambda_peak FLOAT ampirical peak wavelength of
;                        the blaze function for order m [optional in
;                        case k is not known]
;
;                    modfunction=... FLTARR(n) empirical correction
;                        that will be multiplied with the pure blaze
;                        shape. This can be a result of imperfect
;                        groove shape, etc. Values correspond to the
;                        input lambda_in array.  [DEFAULT: no such
;                        correction]
;
;                    THEORETICAL CONSTANTS:
;
;                    angle_blaze=... FLOAT blaze angle in degrees,
;                                          usually referred to as
;                                          delta
;
;                    angle_in=... FLOAT angle incoming light
;                                w.r.t. grating normal in degrees,
;                                usually referred to as alpha (but
;                                different alpha than instrument
;                                constant below). [DEFAULT: same as
;                                blaze angle for pseudo-littrow]
;
;                    angle_reflect=... FLOAT angle reflected light
;                                  w.r.t. grating normal in degrees,
;                                  usually referred to as
;                                  beta. [DEFAULT: same as blaze angle
;                                  for pseudo-littrow]
;
;                    angle_out=... FLOAT out of plane angle of the
;                                grating in degrees, usually referred
;                                to as gamma
;
;                    groove_width=... FLOAT grating groove width in
;                                micron, referred to as 'sigma' in
;                                many equations.
; 
;
;
; v0.0 05/10/18 A. Boogert: created
;

IF keyword_set(print) THEN print, 'echelle order: ', m

;instrument type. Different grating functions or constants apply for
;different instrument designs. Currently only pseudo-Littrow (as for
;iSHELL) implemented.
  IF keyword_set(instrument_type_in) THEN instrument_type=instrument_type_in ELSE $
     instrument_type='pseudo-littrow'

  IF instrument_type NE 'pseudo-littrow' THEN BEGIN
     print, '***ABORTED: only instrument_type of pseudo-littrow implemented'
     return, lambda_in*0.
  ENDIF ELSE BEGIN
     IF keyword_set(print) THEN print, 'instrument type: ', instrument_type
  ENDELSE

;take into account material grating is immersed in. Get real part of
;refractive index n as function of wavelength. For gratings not
;immersed, n=1.
  IF keyword_set(silicon) THEN BEGIN
     n=n_silicon(lambda_in)
     IF keyword_set(print) THEN print, 'silicon-immersed grating'
  ENDIF ELSE BEGIN
     n=FLTARR(n_elements(lambda_in))+1.0
     IF keyword_set(print) THEN print, 'grating in air or vacuum'
  ENDELSE
  ;get mean of n over lambda, which we'll need in some cases
  n_mean=MEAN(n,/nan)
  IF keyword_set(print) THEN print, 'mean n: ', n_mean 
  
;wavelength of peak response of blaze function
  ;
  ;user-provided value
  IF keyword_set(lambda_peak_in) THEN BEGIN
       lambda_peak=lambda_peak_in[0]
       IF keyword_set(k_emp_in) THEN BEGIN
          print, 'cannot give both lambda_peak and k_emp'
          return, lambda_in*0. 
       ENDIF
  ENDIF
  ;
  ;correction to central wavelength may be needed both for
  ;theoretical method and user-provided constants method
  IF NOT keyword_set(lambda_offset_in) THEN lambda_offset=0. ELSE lambda_offset=lambda_offset_in
  ;
  ;determination from user-provided constants
  IF keyword_set(k_emp_in) THEN BEGIN
    lambda_peak=lambda_offset_in+k_emp_in*n_mean/m  ;theoretically it should just be k_emp/m. Don't understant why lambda_offset is needed.
    IF keyword_set(print) THEN BEGIN
       print, 'k_emp: ', k_emp_in
       print, 'k_emp*n_mean: ', k_emp_in*n_mean
       print, 'lambda_offset: ', lambda_offset
    ENDIF
  ENDIF
  ;
  ;theoretical determination
  IF keyword_set(angle_blaze_in) AND keyword_set(angle_out_in) AND keyword_set(groove_width_in) THEN BEGIN
     ;theoretical k for pseudo-Littrow instrument like iSHELL
     IF instrument_type EQ 'pseudo-littrow' THEN BEGIN
        angle_blaze=angle_blaze_in ; blaze angle in degrees, usually referred to as delta

        angle_in=angle_blaze    ;angle incoming light w.r.t. grating normal in degrees,
                                ;usually referred to as alpha (but
                                ;different alpha than instrument
                                ;constant below). Same as blaze angle
                                ;for pseudo-littrow.

        angle_reflect=angle_blaze ;angle reflected light w.r.t. grating normal in
                                  ;degrees, usually referred to as
                                  ;beta. Same as blaze angle for
                                  ;pseudo-littrow.

        angle_out=angle_out_in  ; out of plane angle of the grating in degrees,
                                ; usually referred to as gamma

        groove_width=groove_width_in ;grating groove width in micron

        IF keyword_set(print) THEN BEGIN
           print, 'angle_blaze: ', angle_blaze
           print, 'angle_in: ', angle_in
           print, 'angle_reflect: ', angle_reflect
           print, 'angle_out: ', angle_out
           print, 'groove_width: ', groove_width
        ENDIF

        ;needs to be multiplied with n_mean, but do that later
        k_theo=groove_width*COS(angle_out*!PI/180.)*2.*SIN(angle_blaze*!PI/180.)
     ENDIF

     IF keyword_set(print) THEN BEGIN
        print, 'k_theo: ', k_theo
        print, 'k_theo*n_mean: ', k_theo*n_mean
        print, 'lambda_offset: ', lambda_offset
     ENDIF

     IF keyword_set(lambda_corr_in) THEN lambda_corr=poly(1./m, lambda_corr_in) ELSE lambda_corr=m/m

     lambda_peak=lambda_corr*k_theo*n_mean/m+lambda_offset
  ENDIF

  IF keyword_set(print) THEN print, 'lambda_peak: ', lambda_peak
  
;second grating constant alpha (in addition to k discussed above)
  ;it should be this following Ake et al. (1981)
  ;alpha=(COS(angle_blaze*!PI/180.))^2 
  ;...but I found that this gives a very good fit to the data
  IF NOT keyword_set(alpha_emp_in) THEN BEGIN
     angle_blaze=angle_blaze_in
     alpha=(COS((90-angle_blaze)*!PI/180.))
  ENDIF ELSE BEGIN
     alpha=alpha_emp_in
  ENDELSE

  IF keyword_set(print) THEN print, 'alpha: ', alpha
  
;part of argument in blaze function
X=m*(1.-lambda_peak/(lambda_in))

;full argument in blaze function
d=!PI*alpha*X

;blaze function
r_lambda=(SIN(d)/d)^2

;normalize blaze function
r_lambda=r_lambda/MAX(r_lambda)

output=[m,lambda_peak,alpha]
print, output
return, r_lambda

end
