;#> inter_pol.dc2
; Identifier     inter_pol
;
; Purpose        Perform different kinds of interpolation to resample a 
;                spectrum 
;
; Synopsis       status = inter_pol([algo], xarr, yarr, sarr, refarr, 
;                                   yarr_out, sarr_out)
;
;
; Arguments      Name      I/O Type       Description
;                -----------------------------------------------------
;                algo       I  string     indicate interpolation algorithm;    
;                                         default = 'poly1 ' (linear)
;                xarr       I  Fltarr     x-scale (e.g. wavelength array) that
;                                         is to be resampled
;                yarr       I  Fltarr     y-scale (e.g. flux array) that
;                                         is to be resampled
;                sarr       I  Fltarr     sigma array (stdev) for yarr
;                refarr     I  Fltarr     reference scale to which x-scale
;                                         is to be resampled
;                sarr_out   O  Fltarr     the propagated errors of yarr_out
;                yarr_out   O  Fltarr     the rebinned yarr (interpolated
;                                         flux array)
;
; Returns        error status, 1 for fatal, 0 otherwise       
;
; Description    xarr has to be strictly monotonically increasing (or decreasing).
;                If the reference array contains multiple points at the 'same'
;                value, the interpolation provides each of these points
;                with the same value. This means that the reference array
;                has to be monotonically but need not to be strictly 
;                monotonically increasing (or decreasing).
;                The supported methods of interpolation are:
;                  1) cubic spline interpolation 
;                     Value of algo: 'spline'
;
;                  2) polynomial interpolation 
;                     Value of algo: 'polyx' where x stands for the grade of
;                     the polynomial. Example: poly3 for third order polynomial. 
;
;                The reference scale can be locally coarser or finer than the
;                data. Both may be unequally sampled.
;
;                The error for each interpolated data point (flux) is computed 
;                as follows.
;                weighted mean of the standard deviations of the points
;                used in interpolation (For cubic spline, 5 points are
;                used, the closest point, and 2 additional points on each
;                side). Weights are  computed as the seperations between
;                the reference point and the points in input xarr array.
;
; Comment        A more sphisticated method to comput error propagation will 
;                be used in later versions.   Also, linear interpolation 
;                (flux conserving) will be available in upcoming version.
;
;
; Example        1) status = inter_pol(lsanwav, lsanflx, lsanflxu,lsanwav2,
;                                      lsanflx2, lsanflxu2)
;                   new grid is given by lsanwav2 (the wavelength scale
;                   of a second spectrum). By default linear interpolation is
;                   performed.
;
; Dependencies   CALLS:  avgsig, binsrch, kwd_notset, polyint, sap_error
;                CALLED FROM: sap_rebin, tippex
;
; Category       ISAP
;
; Filename       inter_pol.pro
;
; Author         Iffat R. Khan
;
; Version        0.3
;
; History             27.7.95   definition of requirements - first draft   E. Sturm 
;                      8.8.95   revision of requirements                   E. Sturm
;                      1.9.95   revision of requirements and code          J. Mazz
;                0.1   8.9.95   Design and coding                          I. Khan
;                0.2  17.10.95  revised code (include "linear" for algo)   I. Khan
;                0.3  02.11.95  Revised parameters                         I. Khan
;                               checking/Error messages
;                1.0  24-05-96  change usage to FUNCTION from PRO          I. Khan
;                               to handle error status flag
;
; ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; Copyright (C) 1995, California Institute of Technology.  ;
; U.S. Government Sponsorship under NASA Contract NAS7-918 ;
; is acknowledged.                                         ;
; ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;#<

FUNCTION inter_pol, ALGO = algo, xarr, yarr, sarr, refarr, yarr_out, sarr_out


; check the validity of input arrays and parameters
  chk_stat = KWD_NOTSET(xarr, "xarr", "num", array= "required", /required)
  chk_stat = KWD_NOTSET(yarr, "yarr", "num", array= "required", /required)
  chk_stat = KWD_NOTSET(sarr, "sarr", "num", array= "required", /required)
  chk_stat = KWD_NOTSET(refarr, "refarr", "num", array= "optional", /required)
  IF ( KWD_NOTSET(algo, "algo", "str")) THEN interp = 1 ELSE interp = 0

; get the type of algorithm

  IF interp EQ 0 THEN BEGIN
     IF strpos(strlowcase(algo), 'spline') EQ 0 THEN interp = 0 $
     ELSE IF strpos(strlowcase(algo), 'poly')  EQ 0 THEN interp = strmid(algo, 4, 1) $
     ELSE IF strpos(strlowcase(algo), 'linear')  EQ 0 THEN interp =  1 $
     ELSE  BEGIN
       SAP_ERROR, 'F', 'Unknown algorithm = ' + algo
       return, 1
     ENDELSE
  ENDIF

; points in refarr, and declare output yarr_out, sarr_out
    np = n_elements(refarr)
    yarr_out = fltarr(np) & sarr_out = yarr_out

; array to hold the points for polynomial interpolation & sigma computations
    xint = fltarr(10) & yint = xint & sint = xint & delx = xint

; temp arrays
    stemp = xint & dels = xint

; get the indices in xarr for the points in refarr (binary search)
    nin = n_elements(xarr)
    status  = binsrch(xarr, refarr,index_arr)
    IF status EQ 1 THEN return, 1

  IF interp EQ 0 THEN BEGIN
; cubic spline
    yarr_out = spline(xarr, yarr, refarr)

; compute sigma values for spline case
    
; for each point of ref point, find the index in input array 
    FOR i = 0, np-1 DO BEGIN
       curr = index_arr(i)
; get 2 additional points on both sides of the current index
       left = curr -2
       right = curr + 2

; adjust for lhs endpoint
       IF (left LT 0) THEN BEGIN
	 right = right + abs(left)
	 left = 0
       ENDIF ELSE IF (right GE nin-1) THEN BEGIN

; adjust for rhs endpoint
	 left = left - (right - nin +1)
	 right = nin -1
       ENDIF   

; fill up the array for sigma computations (cubic spline)
; xint, yint contain the points used in interpolation

       xint = xarr(left : right)
       sint = sarr(left : right)
       delx = abs(refarr(i) - xint)
       zerochk = where(delx LT 1.e-20, zcnt)
       IF( zcnt GT 0 ) THEN delx (zerochk)  = 1.e-8
; get mean value of sigma
       stat = avgsig(sint, delx, /disp)
       sinfo = size(stat)
       IF sinfo(sinfo(0)+1) NE 8 THEN return, 1
       sarr_out(i) = stat.mean

    ENDFOR
  ENDIF ELSE BEGIN

; polynomial interpolation starts here

; get the points in xarr and yarr to do the polynomial
; fitting. For nth order we need n+1 points.

; no of points needed on both sides of xarr points (defined by index_arr)
    nsrch = interp/2                            

; check whether degree is odd or even. if odd, we need one additional
; point.
    odd = 0
    oddtst = float(interp)/2.0 - fix( float(interp)/2.0)
    IF oddtst NE 0.0 THEN odd = 1

; start polynomial interpolation for each point in refarr
    FOR i = 0., np-1. DO BEGIN
    excess = 0 & endpoint = 0

; find the range of points
       curr = index_arr(i)
       left = curr - nsrch
       right= curr + nsrch

; left endpoint of data, no points are available, get the points on the right
       IF (left LE 0) THEN BEGIN
	 excess = abs(left)
	 left = 0
	 right = right + excess
	 IF (odd) THEN right = right + 1             ; get one additional point
         IF (right GT nin-1) THEN right = nin - 1
	 endpoint = 1
       ENDIF

; right endpoint of data, no points are available, get the points on the left 
       IF (right GE nin-1) THEN BEGIN
	 excess = right - (nin-1)
	 left = left - excess
	 IF (odd) THEN left = left - 1             ; get one additional point
	 IF (left LT 0) THEN left = 0
	 right = nin - 1
	 endpoint = 1
       ENDIF

; if not endpoint, remember to add one more for odd
; for lin interpolation (nsrch = 0), take one point on the right

       IF odd AND endpoint EQ 0 THEN $
         IF nsrch  THEN   $
           IF (abs(xarr(curr) - xarr(left - 1)) GE   $
            abs(xarr(curr) - xarr(right + 1))) THEN right = right +1 $
           ELSE left = left -1 $
         ELSE right = right + 1


; xint, yint contain the points used in interpolation
       xint = xarr(left : right)
       yint = yarr(left : right)
       sint = sarr(left : right)
       delx = abs(refarr(i) - xint)
       zerochk = where(delx LT 1.e-20, zcnt)
       IF( zcnt GT 0 ) THEN delx (zerochk)  = 1.e-8

; do the interpolation now, and fill up the output y array
       polyint ,xint, yint, refarr(i), yval 
       yarr_out(i) = yval

; compute sigma for error propagation
       stat = avgsig(sint, delx, /disp)
       sinfo = size(stat)
       IF sinfo(sinfo(0)+1) NE 8 THEN return, 1
       sarr_out(i) = stat.mean
    ENDFOR
  ENDELSE

 return, 0
 end
