function ishell_blaze_function, spectrum, orders, configuration, plot=plot, $
                                invert=invert, outparams, theory=theory, $
                                empirical_constants=empirical_constants, $
                                empirical_lambda=empirical_lambda, lambda_corr=lambda_corr_in

;ISHELL_BLAZE_FUNCTION.PRO: calculates the iSHELL blaze function
;
; Calling: out=ishell_blaze_function(
;
;                   spectrum FLTARR object spectrum containing
;                                   wavelengths for which blaze
;                                   function is to be corrected. Flux
;                                   column will be replaced by blaze
;                                   function in output spectrum.
;
;                   orders INDGEN array of iSHELL order numbers
;
;                   configuration STRING iSHELL configuration, such as
;                                 'K3', 'M2', etc...
;
;                   /theory KEYWD use default instrument constants and
;                                 formulae, but do allow for a small
;                                 correction using an empirically
;                                 determined polynomial given in the
;                                 lambda_corr input parameter
;
;                   /empirical_constants KEYWD use empirical
;                              instrument constants with theoretical
;                              formulae
;  
;                   /empirical_lambda KEYWD use empirical blaze peak
;                              wavelengths and no formulae for that
;
;                    lambda_corr FLTARR(2) polynomial coefficients of a
;                                correction function to multiply
;                                theoretical lambda with as function
;                                of 1/m
;                                [DEFAULT: 1., i.e., no such correction]
;
;                                   )
;  
;
; Output: spectral array with blaze function for each order. It's in
;         fact the input array with the flux column replaced by blaze
;         function.
;
;
; v0.0 A. Boogert 05/21/2018: created
; v0.1 A. Boogert 07/18/2018: added defaults, not specific for a configuration
;
;  

IF keyword_set(plot) THEN BEGIN
 cleanplot
 !p.multi=[0,1,2]
ENDIF
   
;output spectrum
spectrum_blaze=spectrum
  
;Blaze constants

;defaults

;theoretical values
groove_width=80.0 ;79.4 ;in micron, from 12.5 grooves/mm, see caption of Figure 10 in Rayner et al. 2016; very accurate..to 0.001 um
angle_blaze=71.6 ; 70.35;71.6 degrees, see caption of Figure 10 in Rayner et al. 2016, accurate from 71.4-71.8
angle_out=1.025;1.025 ;degrees from Rayner et al. 2016, not that accurate

;specifically for each configuration (really just needed for testing)

IF strlowcase(configuration) EQ 'k3' THEN BEGIN
 ;empirical values peak wavelengths
 lambda_peak=[2.5457516, 2.5334653, 2.5213274, 2.5093080, 2.4974055, $
           2.4856182, 2.4739445, 2.4623827, 2.4509312, 2.4395885, $
           2.4283531, 2.4172234, 2.4061979, 2.3952753, 2.3844540, $
           2.3737328, 2.3631101, 2.3525848, 2.3421554, 2.3318207, $
           2.3215793, 2.3114302, 2.3013719, 2.2914033, 2.2815233, $
           2.2717305, 2.2620241]
 ;
 ;empirical constants
 ;
 ;grating constants determined (below) by fitting lambda_c/n
 ;(wavelength of peak intensity divided by wavelength-dependent optical
 ;constant of the immersion material) to 1/m with function
 ;lambda_c/n=k_0+k_1/m
 ;I don't understand why k_0 is needed, though
 lambda_offset=0.0131           ; iSHELL K3 echelle grating constant offset in micron
 k_emp=150.65127     ; iSHELL K3 echelle grating constant slope in air/vacuum
 ;alpha, another grating constant that determines the width of the
 ;blaze function
 alpha=0.97;0.948876
 lambda_corr=[1.0032746, -1.15599581]
 ;
 ;theoretical values
 groove_width=80.0 ;79.4 ;in micron, from 12.5 grooves/mm, see caption of Figure 10 in Rayner et al. 2016; very accurate..to 0.001 um
 angle_blaze=71.6 ; 70.35;71.6 degrees, see caption of Figure 10 in Rayner et al. 2016, accurate from 71.4-71.8
 angle_out=1.025;1.025 ;degrees from Rayner et al. 2016, not that accurate
ENDIF

IF strlowcase(configuration) EQ 'm1' THEN BEGIN
 ;empirical values peak wavelengths
 lambda_peak=[5.2153718, 5.1640313, 5.1133154, 5.0637918, 5.0150318, $
              4.9666073, 4.9198917, 4.8731593, 4.8280853, 4.7839458, $
              4.7402179, 4.6970794, 4.6551089, 4.6137860, 4.5732907, $
              4.5347039]
 ;
 ;empirical constants
 ;
 ;grating constants determined (below) by fitting lambda_c/n
 ;(wavelength of peak intensity divided by wavelength-dependent optical
 ;constant of the immersion material) to 1/m with function
 ;lambda_c/n=k_0+k_1/m
 ;I don't understand why k_0 is needed, though
 lambda_offset=0.045;0.0131;0.044;  ; iSHELL K3 echelle grating constant offset in micron
 k_emp=150.65127     ; iSHELL K3 echelle grating constant slope in air/vacuum
 ;alpha, another grating constant that determines the width of the
 ;blaze function
 alpha=0.97;0.948876
 lambda_corr=[1.0050621,-0.345511311]
 ;
 ;theoretical values
 groove_width=80.0 ;79.4 ;in micron, from 12.5 grooves/mm, see caption of Figure 10 in Rayner et al. 2016; very accurate..to 0.001 um
 angle_blaze=71.6 ; 70.35;71.6 degrees, see caption of Figure 10 in Rayner et al. 2016, accurate from 71.4-71.8
 angle_out=1.025;1.025 ;degrees from Rayner et al. 2016, not that accurate
ENDIF

IF strlowcase(configuration) EQ 'm2' THEN BEGIN
 ;empirical values peak wavelengths
 lambda_peak=[5.2153718, 5.1640313, 5.1133154, 5.0637918, 5.0150318, $
              4.9666073, 4.9198917, 4.8731593, 4.8280853, 4.7839458, $
              4.7402179, 4.6970794, 4.6551089, 4.6137860, 4.5732907, $
              4.5347039]
;   lambda_peak=[5.2133938, 5.1618549, 5.1113394, 5.0618162, 5.0132565, $
;                4.9656325, 4.9189173, 4.8730853, 4.8281116, 4.7839725, $
;                4.7406450, 4.6981068, 4.6563367, 4.6153142, 4.5750192, $
;                4.5354327]
 ;
 ;empirical constants
 ;
 ;grating constants determined (below) by fitting lambda_c/n
 ;(wavelength of peak intensity divided by wavelength-dependent optical
 ;constant of the immersion material) to 1/m with function
 ;lambda_c/n=k_0+k_1/m
 ;I don't understand why k_0 is needed, though
 lambda_offset=0.045;0.0131;0.044;  ; iSHELL K3 echelle grating constant offset in micron
 k_emp=150.65127     ; iSHELL K3 echelle grating constant slope in air/vacuum
 ;alpha, another grating constant that determines the width of the
 ;blaze function
 alpha=0.97                     ;0.948876
 lambda_corr=[1.0050621,-0.345511311]
 ;
 ;theoretical values
 groove_width=80.0 ;79.4 ;in micron, from 12.5 grooves/mm, see caption of Figure 10 in Rayner et al. 2016; very accurate..to 0.001 um
 angle_blaze=71.6 ; 70.35;71.6 degrees, see caption of Figure 10 in Rayner et al. 2016, accurate from 71.4-71.8
 angle_out=1.025;1.025 ;degrees from Rayner et al. 2016, not that accurate
ENDIF

;overrride lambda_corr in this file with user input
IF NOT keyword_set(lambda_corr_in) THEN lambda_corr=[1.,0] ELSE lambda_corr=lambda_corr_in

;return m, lambda_peak, and alpha
outparams=FLTARR(3,n_elements(orders))

FOR ord=0,n_elements(orders)-1,1 DO BEGIN
   output=1
   ;
   IF keyword_set(theory) THEN BEGIN
      blaze_function=blaze_function(spectrum_blaze[*,0,ord], orders[ord], lambda_corr=lambda_corr, alpha_emp=alpha, $
       groove_width=groove_width,angle_blaze=angle_blaze,angle_out=angle_out,/silicon,/print,output)
   ENDIF
   ;
   IF keyword_set(empirical_constants) THEN BEGIN   
      blaze_function=blaze_function(spectrum_blaze[*,0,ord], orders[ord], alpha_emp=alpha, $
        k_emp=k_emp, lambda_offset=lambda_offset, angle_blaze=angle_blaze,/silicon,/print,output)
   ENDIF
   ;
   IF keyword_set(empirical_lambda) THEN BEGIN   
      blaze_function=blaze_function(spectrum_blaze[*,0,ord], orders[ord], alpha_emp=alpha, $
        lambda_peak=lambda_peak[ord], angle_blaze=angle_blaze,/silicon,/print,output)
   ENDIF
   ;just to transfer the blaze function parameters to output
   outparams[*,ord]=output

   IF keyword_set(invert) THEN spectrum_blaze[*,1,ord]=1./spectrum_blaze[*,1,ord]

   IF keyword_set(plot) THEN BEGIN
     loadct, 2, /silent
     plot, spectrum_blaze[*,0,ord], spectrum_blaze[*,1,ord]/max(medsmooth(spectrum_blaze[*,1,ord],50)), title='order '+STRING(orders[ord]),yr=[0,1.2]

     oplot, spectrum_blaze[*,0,ord], blaze_function, color=80
     plot, spectrum_blaze[*,0,ord], spectrum_blaze[*,1,ord]/max(medsmooth(spectrum_blaze[*,1,ord],50))/blaze_function, yr=[0.8,1.2]
     oplot, [-100,100],[1,1],linestyle=2
     oplot, [0,0]+output[1], [-100,100], color=30
     oplot, [0,0]+output[1]+output[2]*0.0014, [-100,100], color=30
     oplot, [0,0]+output[1]+output[2]*0.0045, [-100,100], color=30          
     stopcont
   ENDIF
   spectrum_blaze[*,1,ord]=blaze_function
ENDFOR

return, spectrum_blaze

end
