function ishell_derive_blaze_peaks, flats, orders, configuration, plot=plot_in, dir=dir_in

;ISHELL_DERIVE_BLAZE_PEAKS.PRO: derive the wavelengths of the peak
;          response of the blaze response at each order of an iSHELL
;          spectrum.
;
;Calling: output=ishell_derive_blaze_peaks(
;
;              flats FLTARR(n,m,o) arrray with extracted flat spectra,
;                                  reduced using Spextool, but not
;                                  order-merged.
;
;              orders INTARR(o) echelle order numbers corresponding to
;                               input flats
;                
;              configuration STRING iSHELL configuration
;
;              plot=... STRING name of file with diagnostic plots
;                              [DEFAULT: no plotting]
;                
;              dir_in=... STRING directory of file ishell_good_flat_blaze_orders.dat
;                              [DEFAULT: ./, i.e., current directory]
;                
;                                         )
;
;           output: FLTARR(3,o) wavelength of the peak of each echelle
;                             order
;                             output[0,*]: echelle order
;                             output[1,*]: lambda_peak
;                             output[1,*]: alpha (a grating constant)
;
; v0.0 A.Boogert 07/Jun/2018: created
; v0.1 A.Boogert 18/Jul/2018: added plot of polynomial fit to lambda_peak(measured)/lambda_peak(theory)
; v0.2 A.Boogert 27/Aug/2018: exclude NaN's from polynomial fit
; v0.3 A.Boogert 03/Oct/2018: added dir= input
; v0.4 A.Boogert 29/Nov/2018: -take into account NaN's in normalization
;                              and corrected bug in interpolating over
;                              them
;                             -always exclude first and last order
;                              (could be partially on array)
; v0.5 A.Boogert 09/Nov/2020: determine peak of order response by
;                             excluding edges
;

  IF keyword_set(dir_in) THEN dir=dir_in ELSE dir='./'
  
  n_orders=n_elements(orders)

  ;theoretical blaze functions for each order
  flat_theory=ishell_blaze_function(flats,orders,configuration,/theory,outparams,lambda_corr=0);, lambda_corr=[1.0032746, -1.15599581])
  lambda_theo=outparams[1,*]
  alpha_theo=outparams[2,*]

  ;FOR i=0,n_orders-1 DO BEGIN
  ;   cleanplot
  ;   window, 1
  ;   plot, flats[*,0,i], flats[*,1,i]/max(medsmooth(flats[*,1,i],50)),yr=[-0.1,1.1]
  ;   oplot, flat_theory[*,0,i], flat_theory[*,1,i], color=2
  ;   stop
  ;ENDFOR
  
  ;find wavelength offset between flat and theoretical blaze function
  ;for suitable orders (i.e., not affected by molecular absorption in
  ;air between flat lamp and instrument)
  ;
  ;orders suitable to derive blaze peak from
  dum=''
  s_orders=0  
  s_orders_qual=0   ;quality of each order
  openr,1, dir+'ishell_good_flat_blaze_orders.dat'
  while not eof(1) DO BEGIN
     readf, 1, dum
     IF strmid(dum,0,1) NE '#' THEN BEGIN        
        s_orders=[s_orders, fix(STRMID(dum,0,3))]
        s_orders_qual=[s_orders_qual, fix(STRMID(dum,4,20))]        
     ENDIF
  endwhile
  close, 1

  ;always exclude first and last order as they may be partially on the
  ;array (and often they have molecular lines, being at atm
  ;transmission band edges)
  ind=where((s_orders EQ min(orders)) OR (s_orders EQ max(orders)))
  s_orders_qual[ind]=0
  
  ind=where(s_orders NE 0 AND s_orders_qual NE 0)
  s_orders=s_orders[ind]
  s_orders_qual=s_orders_qual[ind]  
  ;
  ;initialize array for shifts between observation and theory
  shifts=FLTARR(n_orders)-1.
  ;initialize arra for actual peak wavelengths
  lambda_peak=FLTARR(n_orders)
  lambda_peak_plot=FLTARR(n_orders)  
  ;initialize diagnostic plots
  IF keyword_set(plot_in) THEN BEGIN
    set_plot, 'ps'
    device, filename=plot_in,/color, bits_per_pixel=8,$
            /landscape, scale_factor=1.0
    print, plot_in+' will be written to disk'
  ENDIF
       
  FOR i=0, n_orders-1, 1 DO BEGIN
     order=orders[i]
     ind=where(order EQ s_orders, count)
     IF count GT 0 THEN BEGIN
       ;warn if this order was not checked
       IF s_orders_qual[ind] EQ 2 THEN print, '***WARNING: order ', order, ' not checked for molecular lines'
       
       w=flats[*,0,i]
       f=flats[*,1,i]
       b=flats[*,3,i]

       ;excluded data points with any non-zero flag
       ind=where(b EQ 0.)
       w=w[ind]
       f=f[ind]
       ;order in increasing wavelength
       ind=sort(w)
       w=w[ind]
       f=f[ind]       
       ;put on equidistant grid, without gaps
       we=FINDGEN(2000)/2000*(max(w)-min(w))+min(w)
       fe=interpol(f,w,we)
       ;use heavily smoothed flat to get rid of bad pixels
       fe=medsmooth(fe,50)
       ;and normalize it, excluding edges (avoiding spurious peaks
       ;near edge)
       indmax=FINDGEN(1000)+250
       fe=fe/max(fe[indmax],/nan)

       ;put on blaze x-scale using first order blaze wavelength from
       ;maximum signal
       lambda_peak_temp=we[(where(fe EQ 1.0))[0]]
       x=!PI*alpha_theo[i]*order*(1.-lambda_peak_temp/we)
       
       ;;and then refine observed blaze wavelength by cross correlation
       ;;with theoretical curve
       ;wt=flat_theory[*,0,i]
       ;ft=flat_theory[*,1,i]
       ;;order in increasing wavelength
       ;ind=sort(wt)
       ;wt=wt[ind]
       ;ft=ft[ind]
       ;;convert to blaze x scale
       ;xt=!PI*alpha_theo[i]*order*(1.-lambda_theo[i]/wt)
       ;;put on same x grid as observation
       ;fti=interpol(ft,xt,x)

       fti=(SIN(x)/x)^2

       ;interpolate over any NaNs or the cross correlation won't worj
       ind=where(finite(fti) EQ 0, count)
       IF count GT 0 THEN BEGIN
         ind=where(finite(fti), count)
         fti=interpol(fti[ind],x[ind],x)
       ENDIF
       ind=where(finite(fe) EQ 0, count)
       IF count GT 0 THEN BEGIN
         ind=where(finite(fe), count)
         fe=interpol(fe[ind],x[ind],x)
       ENDIF

       ;determine response of observed blaze
       ;function on short x side which is
       ;most similar to theoretical curve
       anchors=min(x)*[0.3,0.4,0.5,0.6,0.7,0.8,0.9,1.0]
       r_anchors=interpol(fe, x, anchors)

       ;determine x with same response in theoretical curve
       ;x_theo=!PI*alpha_theo[i]*order*(1.-lambda_theo[i]/we)       
       ind=where(x LT 0. AND x GT -2.0)
       anchors_theo=interpol(x[ind], fti[ind], r_anchors)

       ;shifts for each ancor point
       w_shifts_anchor=lambda_peak_temp/(1.-anchors/(!PI*alpha_theo[i]*order))- $
                       lambda_peak_temp/(1.-anchors_theo/(!PI*alpha_theo[i]*order))

       ;median shift over anchor points is the shift we're looking for
       median_shifts_anchor=median(w_shifts_anchor)
       ;best peak wavelength
       lambda_peak[i]=lambda_peak_temp+median_shifts_anchor
       ;lambda_peak for plotting, taking into
       ;account that data plotted using lambda_peak_temp
       lambda_peak_plot[i]=lambda_peak_temp-median_shifts_anchor
       shifts[i]=median_shifts_anchor
       ;for plotting
       shiftx=median(anchors_theo-anchors)

       IF keyword_set(plot_in) THEN BEGIN       
          plot, x, fe, title='order '+string(order), ytitle='B=(SIN(x)/x)!U2!N', $
                xtitle='x=!7p*a*!Xorder*(1-!7k!X!Bpeak!N/!7k!X)',xr=[-2.0,2.0], $
                yr=[0,1.05],xs=1,ys=1
         ;print, fe
         ;print, order
         ;stop
         oplot, x, fti, color=2
         xnew=!PI*alpha_theo[i]*order*(1.-lambda_peak_plot[i]/we)
         ftinew=(SIN(x)/x)^2
         oplot, xnew, ftinew, color=3
         FOR anch=0,n_elements(anchors)-1,1 DO BEGIN
          oplot, [0,0]+anchors[anch], [0,1.1], color=4
         ENDFOR
         xyouts, -0.2,0.50, 'black: data with !7k!X!Bpeak!N=!7k!X(B=1)'
         xyouts, -0.2,0.45, 'red: blaze function with !7k!X!Bpeak!N=!7k!X(B=1)', color=2
         xyouts, -0.2,0.40, 'green: shifted to fit anchor wavelengths (blue lines)', color=3
         xyouts, -0.0,0.35,        'x!Bpeak!N='+string(-shiftx), color=3
         xyouts, -0.2,0.30, 'blue: anchor wavelengths', color=4
         xyouts, -0.2,0.20, 'correct blaze peak found if green line follows black line'
       ENDIF
        
    ENDIF     
  ENDFOR

  ;fit polynomial to array of shifts beteen observation and theory
  ind=where(shifts NE -1.)
  indfit=where((shifts NE -1.) AND (finite(lambda_peak/lambda_theo) EQ 1))
  
  coef=poly_fit(1./orders[indfit], lambda_peak[indfit]/lambda_theo[indfit], 1,/double)

  print, 'coefficients:'
  print, coef
  
  IF keyword_set(plot_in) THEN BEGIN         
    plx=FINDGEN(1000)/1000*max(1./orders[ind])
    plot, 1./orders[ind], lambda_peak[ind]/lambda_theo[ind], ps=4,ys=1, $
          ytitle='!7k!X!Bpeak!N (measured)/!7k!X!Bpeak!N (theory)', $
          xtitle='1/order', title='Any strong outliers could be due to molecular lines in flat field', $
          yr=[min(lambda_peak[ind]/lambda_theo[ind]),max(lambda_peak[ind]/lambda_theo[ind])]
    oplot, plx, poly(plx, coef), color=2
  ENDIF
  ;stopcont
  
  ;then calculate actual blaze peaks for all orders
  dum=ishell_blaze_function(flats,orders,configuration,/theory,outparams, lambda_corr=coef)

  ;and overplot the result to see if optimal blaze function indeed
  ;fits the observed one
  IF keyword_set(plot_in) THEN BEGIN         
    FOR i=0, n_orders-1, 1 DO BEGIN
       order=orders[i]
       ind=where(order EQ s_orders, count)
       IF count GT 0 THEN BEGIN
         x=!PI*alpha_theo[i]*order*(1.-outparams[1,i]/flats[*,0,i])
         fbest=dum[*,1,i]
         plot, x, flats[*,1,i]/max(medsmooth(flats[*,1,i],50)), title='order '+string(order), $
               yr=[0,1.05], xr=[-2,2],xs=1,ys=1, ytitle='B=(SIN(x)/x)!U2!N', $
                xtitle='x=!7p*a*!Xorder*(1-!7k!X!Bpeak!N/!7k!X)'
         oplot, x, fbest, color=2
         oplot, [-2,2],[1,1],linestyle=1
         xyouts, -0.2,0.50, 'black: data using optimal !7k!X!Bpeak!N='+string(outparams[1,i])+' !7l!Xm'
         xyouts, -0.2,0.45, 'red: blaze function with !7k!X!Bpeak!N='+string(outparams[1,i])+' !7l!Xm', color=2
         
      ENDIF
    ENDFOR
  ENDIF

  ;end of plotting
  IF keyword_set(plot_in) THEN BEGIN
       device, /close
       set_plot, 'x'
  ENDIF
  
  return, outparams
  
end

