function ishell_derive_flatblaze, flats, blazepeaks, orders, alpha, configuration, $
                                  plot=plot_in, write=write_in, dir=dir_in
  
;ISHELL_DERIVE_FLATBLAZE.PRO: derive median blaze function from 
;                             extracted flat field orders.
;
;Calling: output=ishell_derive_flatblaze(
;
;              flats FLTARR(n,m,o) array with extracted flat spectra,
;                                  reduced using Spextool, but not
;                                  order-merged.
;              
;              blazepeaks FLTARR(o) previously determined precise
;                                   wavelengths of the peak signal of
;                                   the flat spectra for each order
;
;              orders INTARR(o) echelle order numbers corresponding to
;                               input flats
;
;              alpha FLOAT a grating constant
;                
;              configuration STRING iSHELL configuration
;
;              plot=... STRING name of postscript file for diagnostic
;                              plots [DEFAULT: no such plot]
;                
;              write=... STRING name of data file with median blaze
;                               function to be written to disk
;                               [DEFAULT: no such writing]
;                
;              dir_in=... STRING directory of file ishell_good_flat_blaze_orders.dat
;                              [DEFAULT: ./, i.e., current directory]
;                
;                                         )
;
;           output: FLTARR(2,n) blaze function derive from flats
;                               output[0,*] Blaze wavelength scale ~PI*alpha*order*(1-blazepeak/lambda)
;                               output[1,*] blaze function
;
; v0.0 A.Boogert 07/Jun/2018: created
; v0.1 A.Boogert 03/Oct/2018: added dir= input
; v0.2 A.Boogert 03/Oct/2018: always exclude first and last order
;                             (could be partially on array)
; v0.3 A.Boogert 30/Nov/2018: added wavelengths of all orders to
;                             medianblaze output data file
;
;

  IF keyword_set(dir_in) THEN dir=dir_in ELSE dir='./'

  n_orders=n_elements(orders)

  ;the median blaze from flat field to be divided out first put in
  ;common array, on common grid
  n=6000
  x=FINDGEN(n)/n*6.-3.
  flatarr=FLTARR(n,n_orders)-1.
  flatarr_med=FLTARR(n,2+n_orders) ; x, med_flat, w_ord1, ...,w_ordn
  flatarr_med[*,0]=x
  ;cleanplot
  ;!p.multi=[0,1,2]
  ;rfile, 'ishell_k3_medianblazeflat.dat', flatarr_medf, path='$HOME/Soft/Spextool_ishell/Mylib/'
  ;
  ;orders suitable to derive median blaze function from
  dum=''
  s_orders=0 
  s_orders_qual=0   ;quality of each order
  openr,1, dir+'ishell_good_flat_blaze_orders.dat'
  while not eof(1) DO BEGIN
     readf, 1, dum
     IF strmid(dum,0,1) NE '#' THEN BEGIN        
        s_orders=[s_orders, fix(STRMID(dum,0,3))]
        s_orders_qual=[s_orders_qual, fix(STRMID(dum,4,20))]        
     ENDIF
  endwhile
  close, 1

  ;always exclude first and last order as they may be partially on the
  ;array (and often they have molecular lines, being at atm
  ;transmission band edges)
  ind=where((s_orders EQ min(orders)) OR (s_orders EQ max(orders)))
  s_orders_qual[ind]=0
  
  ind=where(s_orders NE 0 AND s_orders_qual NE 0)
  s_orders=s_orders[ind]
  s_orders_qual=s_orders_qual[ind]
    
  ;which suitable orders are in these data
  match, s_orders, orders, dum, ind_orders
  good_orders=orders[ind_orders]

  ;warn if this order was not checked
  match, s_orders[where(s_orders_qual EQ 2)], orders, dum, ind_orders2
  IF ind_orders2[0] NE -1 THEN print, '***WARNING: order ', orders[ind_orders2], ' not checked for molecular lines'       

  FOR i=0,n_orders-1,1 DO BEGIN
    ind=where(orders[i] EQ good_orders, count)
    IF count GT 0 THEN BEGIN 

     ;;Multiply small slope on wavelengths to get orders to match better.
     ;;This is expected as the flat field spectrum is probably not
     ;;intrinsically flat.
     ;wf=flats[*,0,i];-blazepeaks[i]
     ;rico=[1,1.e16]
     ;funct=rico[0]+rico[1]*wf^(4)
     ;functb=interpol(funct,wf,blazepeaks[i])
     ;funct=funct/functb
     ;print, funct
     ;flat=flats[*,1, i]/max(medsmooth(flats[*,1,i],50))*funct
     flat=flats[*,1, i]/max(medsmooth(flats[*,1,i],50))
     xd=!PI*alpha[i]*orders[i]*(1.-blazepeaks[i]/flats[*,0, i])
     ;IDL 7 does not allow /nan keyword in interpol
     ;flati=interpol(flat,xd,x,/nan) 
     ind=where(finite(flat) EQ 1)
     flati=interpol(flat[ind],xd[ind],x)
     
     ;assign NaN to wavelengths for which there was no data
     ind=where((x LT min(xd) OR x GT max(xd)))
     flati[ind]=0/0.
     flatx=medsmooth(flati,20)
     flatarr[*,i]=flatx
    ENDIF
  ENDFOR
  ;only median over orders not affected by molecular lines
  ind_med=where(flatarr[0,*] NE -1.)
  FOR i=0,n-1 DO BEGIN
    flatarr_med[i,1]=median(flatarr[i,ind_med])
  ENDFOR
  flatarr_med=transpose(flatarr_med)
  ;flatarr_med[1,*]=medsmooth((flatarr_med[1,*])[*],50)

  ;remove data points with NaNs
  ind=where(finite(flatarr_med[1,*]) EQ 1)  
  flatarr_med=flatarr_med[*,ind]
  
  ;end of plotting
  IF keyword_set(plot_in) THEN BEGIN
        set_plot, 'ps'
        device, filename=plot_in,/color, bits_per_pixel=8,$
                /landscape, scale_factor=1.0
        print, plot_in+' will be written to disk'
        FOR i=0,n_orders-1,1 DO BEGIN
          ind=where(orders[i] EQ good_orders, count)
          IF count GT 0 THEN BEGIN 
           plot, x, flatarr[*,i], title='order '+string(orders[i]), $
                 ytitle='B=(SIN(x)/x)!U2!N', xs=1,ys=1,xr=[-2,2],yr=[0.0,1.05], $
                 xtitle='x=!7p*a*!Xorder*(1-!7k!X!Bpeak!N/!7k!X)'
           oplot, flatarr_med[0,*], flatarr_med[1,*], color=2
           xyouts, -1.3,0.30, 'black: data using optimal !7k!X!Bpeak!N='+string(blazepeaks[i])+' !7l!Xm'
           xyouts, -1.3,0.20, 'red: blaze function derived from median over good orders', color=2           
          ENDIF
        ENDFOR
       device, /close
       set_plot, 'x'
  ENDIF

  ;multiply small slope to get orders to match better
  x=flatarr_med[0,*]
  rico=[1.,0.006] ;1st coefficient is for normalization and has to stay 1.0, 2nd coefficient needs to be optimized. 0.006 seems best for K3 
  funct=poly(x, rico)
  flatarr_med[1,*]=flatarr_med[1,*]*funct

  ;add wavelength (in micron) scale to flatarr_med
  FOR i=0,n_orders-1 DO BEGIN
     flatarr_med[2+i,*]=blazepeaks[i]/(1.-flatarr_med[0,*]/(!PI*alpha[i]*orders[i]))
  ENDFOR
  
  IF keyword_set(write_in) THEN BEGIN
     openw, 1, write_in
     printf, 1, '#     x       blaze    wavelength_micron_for_orders.....................', format='(a71)'
     printf, 1, '#            median', orders, format='(a19,'+STRING(n_orders)+'i10)'     
     printf, 1, flatarr_med, format='(f10.6,f11.6,'+STRING(n_orders)+'f10.6)'
     close, 1
     print, '***median blaze function written to file ', write_in
  ENDIF

  return, flatarr_med
  
end

