;#> kwd_notset.dc2
; Indentifier   kwd_notset
;
; Purpose       Check if a parameter is set and of the right type
;
; Synopsis      status = kwd_notset, param, name, type, [required]
;
; Arguments     Name      I/O  Type     Description
;               -----------------------------------------------------------
;
;               param     I    (any)    Parameter to be checked
;
;               name      I    string   Name of parameter (for messsages)
;
;               type      I    string   One of:
;                                        "int": integer (long or short)
;                                        "flo": floating or double
;                                        "num": int, long int, float, double
;                                        "str": string
;                                        "aar": AAR structure (calls is_struct)
;                                        "out": ouptut parameter (anything)
;                                        "any": anything
;
;               array     I    string   Optional specifier for arrays.  Must
;                                       be on of: "optional", "required",
;                                       if it is optional or required that the 
;                                       parameter be an array.
;
;               required  I    int      Optional flag for required parameter
;
; Returns       1 if set,   0 if not set  If parameter is not valid,
;                             sap_error does RETALL in command more,
;                             but returns with a 1 in GUI mode.
;                            (This is an obsolete error handling method
;                             and will be changed in the next version.)
;
; Description   This subroutine checks that the parameter is of the type
;               type indicated and prints a message if not.  For efficiency,
;               no error checking is done on the call to kwd_notset itself.
;
; Examples:
;
;               To check if detno is an integer.   ....
;
;                   IF (kwd_notset(detno, "detno", "int")) detno = 1
;
; Dependencies   CALLS: is_struct, sap_keyword_set, sap_error
;                CALLED FROM: bin_struct, chk_multix, disp_aar, flux_conserv,
;                       glue, inter_pol, mask, normal, poly_int, sap_rebin,
;                       sap_rfits, sap_shift, sap_wfits, scale_define, tippex
;
; Category      ISAP
;
; Filename      kwd_notset.pro
;
; Author        R. Narron (IPAC)
;
; Version       1.2
;
; History       1.0   951021  --> R. Narron (IPAC)
;               1.1   960129      E. Sturm (MPE) new call to is_struct, added
;                                 dependencies
;               1.2   960129      RN - change handling of invalid parameters:
;                                      RETALL    for comand line
;                                      return,1  for GUI
;                            
; ****************************************************************************
; Copyright (C) 1995, California Institute of Technology.
; U.S. Government Sponsorship under NASA Contract NAS7-918 is acknowledged.
; ****************************************************************************
;#<

FUNCTION kwd_notset, param, name, type, array=array, required=required

;--------------------------------------------------------------
; process parameters to 'kwd_notset' subroutine
;--------------------------------------------------------------
if (keyword_set(required) EQ 0) THEN required = 0
if (keyword_set(array) EQ 0)    THEN array = ""
type3 = strmid(strlowcase(type),0,3)

;--------------------------------------------------------------
; check if param is really really set (set and not "NONE")
;--------------------------------------------------------------
set = sap_keyword_set(param)

;--------------------------------------------------------------
; process if not set
;--------------------------------------------------------------
IF (set EQ 0) THEN BEGIN
  IF (required NE 0 AND type3 NE "out") THEN BEGIN
   sap_error, "FATAL", "Required parameter not given: " + name
   GOTO, fatal_e
  ENDIF ELSE RETURN, 1  ; not set
ENDIF

;--------------------------------------------------------------
; get info about parameter
;--------------------------------------------------------------
siz_info = size(param)
ndims = siz_info(0)
type_n = siz_info(ndims+1)
n_elts = siz_info(ndims+2)

;print, "SIZE= ", siz_info

;--------------------------------------------------------------
; check if right type
;--------------------------------------------------------------
ok = 1
CASE type3 OF
 "": 
 "any":
 "out":
 "aar": ok = is_struct(param,'AAR')
 "int": IF (type_n LT 2 OR type_n GT 3) THEN ok = 0
 "flo": IF (type_n LT 4 OR type_n GT 5) THEN ok = 0
 "num": IF (type_n LT 2 OR type_n GT 5) THEN ok = 0
 "str": IF (type_n NE 7) THEN ok = 0
 ELSE: ok = -1
ENDCASE

IF (ok EQ -1) THEN BEGIN
 sap_error, "PROG ERR", "kwd_notset: illegal 'type' option " + type
 GOTO, fatal_e
ENDIF

IF (ok EQ 0) THEN BEGIN
 sap_error, "FATAL", "Parameter '" + name + "' must be of type: " + type
 GOTO, fatal_e
ENDIF

;--------------------------------------------------------------
; check for array
;--------------------------------------------------------------
array3 = strmid(strlowcase(array),0,3)
CASE array3 OF
 "": if (n_elts GT 1) THEN BEGIN
      sap_error, "FATAL", "Parameter '" + name + "' may not be an array."
      GOTO, fatal_e
     ENDIF
 "opt":
 "req": if (ndims EQ 0) THEN BEGIN
         sap_error, "FATAL", "Parameter '" + name + "' must be an array."
         GOTO, fatal_e
        ENDIF
 ELSE: BEGIN
  sap_error, "PROG ERR", "kwd_notset: illegal 'array' option " + array
  GOTO, fatal_e
 ENDELSE
ENDCASE

;--------------------------------------------------------------
; return: Yes, it is set
;--------------------------------------------------------------
RETURN, 0

;--------------------------------------------------------------
; fatal error handling
;--------------------------------------------------------------
fatal_e:
 gui_mode = 0
 defsysv, "!sap_gui", exists=e
 IF e NE 0 THEN BEGIN
  cmd='gui_mode=!SAP_GUI'
  r=execute(cmd)         ;execute is necessary to fool the foolish compiler
 ENDIF
IF (gui_mode eq 0) THEN RETALL
RETURN, 1

;--------------------------------------------------------------
END

; size(xxxx)  returns   NDIMs  DIM1  DIM2  ...  TYPE  NELTS
;     s = size(xxx)
;     ndims = s[0]
;     type = s[ndims+1]
; 0 undefined
; 1 byte
; 2 int
; 3 long int
; 4 float
; 5 double
; 6 complex
; 7 string
; 8 structure
