;#> polyint.dc2
; Identifier     polyint   
;
; Purpose        Interpolate a set of N points by fitting a polynomial of 
;                degree N-1
;
; Synopsis       polyint, xp , yp , x, y    
;                             
;
; Arguments      Name      I/O       Type           Description
;                -----------------------------------------------------
;                xp        I   any except string   input x array 
;                yp        I   any except string   input y array 
;                x         I   any except string   x value to be interpolated
;                y         O   any except string   y value corresponding to x
;
; Returns        y - Scalar, interpolated value in (xp,yp) corresponding to x
;
; Description    Interpolate a set of N points by fitting a polynomial of 
;                degree N-1.
;                Adapted from algorithm in Numerical Recipes, Press et al. 
;                (1986) Section 3.1 (p82).
;                Uses Neville's algorithm to iteratively build up the correct
;                polynomial, with each iteration containing one higher order.
;
; Comment        
;
; Example   	 Find sin(2.5) by polynomial interpolation on sin(indgen(10))
;
;		 IDL> xp = indgen(10)
;		 IDL> yp = sin( xp )
;		 IDL> polyint, xp, yp, 2.5, y ,dy
;             
;                The above method gives  y = .5988 and dy = 3.1e-4  a close
;	         approximation to the actual sin(2.5) = .5985
;
; Dependencies   CALLS: kwd_notset, sap_error
;                CALLED FROM: flux_conserv, inter_pol
;
; Category       ISAP
;
; Filename       polyint.pro 
;
; Author         I. R. Khan       
;
; Version        0.2
;
; History        0.1 27-08-95  Design and code       --> irk@ipac
;                0.2 01-11-95  Revised parameters    --> irk@ipac
;                              checking/Error messages
;
; ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; Copyright (C) 1995, California Institute of Technology.  ;
; U.S. Government Sponsorship under NASA Contract NAS7-918 ;
; is acknowledged.                                         ;
; ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;#<

PRO polyint, xp , yp , x, y, dy

; check the validity of input arrays and parameters
  chk_stat = KWD_NOTSET(xp, "xp", "num", array= "required", /required)
  chk_stat = KWD_NOTSET(yp, "yp", "num", array= "required", /required)

; check array sizes first
 n = n_elements( xp ) 
 IF n_elements( yp ) NE n THEN BEGIN
      SAP_ERROR, 'F', 'Input X and Y vectors must have same number of elements'
      return
 ENDIF

; Find the index of xp which is closest to x

 dif = min( abs(x-xp), ns )

; initialise correction vectors c's and d's.
 c = yp    
 d = yp

; initial approximation to y
 y = yp(ns)
 ns = ns - 1

; for each element of the given array
 FOR m = 1,n-1 DO BEGIN

; compute to update current c's and d's
    ho = xp(0:n-m-1) - x
    hp = xp(m:n-1) - x
    w = c(1:n-m) - d(0:n-m-1)
    den = ho - hp

    IF min( abs(den) ) EQ 0 THEN BEGIN 
      SAP_ERROR, 'F', 'All input x vector values must be distinct'
      return
    ENDIF

    den = w / den

; here they are actually updated
    c = ho * den
    d = hp * den

; decide which value to add to the accumulating value of y
    IF ( 2*ns LT n-m-1 ) THEN dy = c(ns+1) $
    ELSE BEGIN
         dy = d(ns)
         ns = ns - 1
    ENDELSE
    y = y + dy
 ENDFOR

 return
 end
