; this file contains the following IDL programs:
;
; -GenerateTrans: get atmospheric transmission spectrum from
;                 GSFC's PSG server. This program is provided
;                 by Geronimo Villanueva and modified where needed by
;                 Adwin Boogert.
;
; -teltool: correct observed spectra for atmospheric
;             absorption. Initially optimized for IRTF/iSHELL
;             spectra. It (may) call GenerateTrans to get atmospheric
;             model spectra from PSG at GSFC, but in the future it
;             will also be possible to call the IRTF weather
;             server. Written by Adwin Boogert


function teltool_conv_instr, waven_in, flux_in, atmstruct

; TELTOOL_CONV_INSTR.PRO: convolve atmospheric model spectrum with an
;                         instrument profile.
;
; Calling: out=teltool_conv_instr(  waven_in DBLARR(n) array of wavenumbers
;                                          [cm-1] of spectrum to be convolved
;  
;                                 , flux_in DBLARR(n) array of flux
;                                        [any unit] of spectrum to be convolved
;
;                                 , atmstruct STRUCTURE IDL structure
;                                        containing all information
;                                        needed for the convolution,
;                                        such as shape and width of
;                                        instrumental profile
;
;                                )
;
; Output: out DBLARR(2,n) with n the same number of elements as input
;                         waven and flux, specifically:
;
;             out[0,*]  array of wavenumbers [cm-1] of convolved spectrum
;
;             out[1,*]  array of flux [arbitrary units] of convolved spectrum
;  
;         
; v0.0 A. Boogert 05/Jul/2018: created

; sort input spectrum 
indsort=sort(waven_in)
waven=waven_in[indsort]
flux=flux_in[indsort]

;put input spectrum on equidistant grid in wavenumber space
sampling=(max(waven)-min(waven))/(n_elements(waven)-1)
waven_i=FINDGEN(n_elements(waven))*sampling+min(waven)
flux_i=interpol(flux, waven, waven_i)
median_wn=median(waven_i)

; define kernel array

;keep track of if line_spread_function is defined
lsf_present=0                   

;the 'measured' lsf is not yet available for all configs and slits
IF atmstruct.line_spread_function EQ 'measured' AND $
   ((float(atmstruct.slit_width) NE 0.375 AND $
    float(atmstruct.slit_width) NE 0.75) OR $
    STRMID(strlowcase(atmstruct.configuration),0,1) EQ 'l'  OR $
    STRMID(strlowcase(atmstruct.configuration),0,1) EQ 'm' $
   ) $
   THEN atmstruct.line_spread_function='gaussian' 

;purely Gaussian LSF with FWHM defined in resolving_power
IF atmstruct.line_spread_function EQ 'gaussian' THEN BEGIN
  ; define x and y values for kernel, the line spread function
  fwhm=median_wn/float(atmstruct.resolving_power)
  sigma=fwhm/(2.*SQRT(2.*ALOG(2.)))
  n=ROUND(6.*fwhm/sampling)
  lsf_waven=sampling*(FINDGEN(n)-n/2)
  peak=0.
  x=lsf_waven
  ;lsf_flux = mygauss(lsf_waven, fwhm=fwhm, peak=0.)
  lsf_flux = EXP(-(x-peak)^2/(2.*sigma^2))/(SQRT(2.*!PI)*sigma)
  lsf_present=1
ENDIF

;formalization of LSF measured from arc lines
IF atmstruct.line_spread_function EQ 'measured' THEN BEGIN
  ;iSHELL:
  ;
  ;Christian has this in python:
  ;from astropy.convolution import convolve
  ;from astropy.modeling.models import Voigt1D, Box1D
  ;
  ;fwhm_Lorentzian=...
  ;fwhm_Gaussian=...
  ;fwhm_Box=...
  ;
  ;Voigt1 = Voigt1D(x_0=x_0, amplitude_L=1., fwhm_L=fwhm_Lorentzian,
  ;                 fwhm_G=fwhm_Gaussian)
  ;VoigtProfile1 = voigt1(wave_array)
  ;box1 = Box1D(amplitude=1., width=fwhm_Box)
  ;bbox = box1(wave_array)[:-1]
  ;Convolved_data = convolve(VoigtProfile1, bbox, boundary='extend')
  ;normalized_Voigt_box = Convolved_data / np.nanmax(Convolved_data)

  IF float(atmstruct.slit_width) EQ 0.375 THEN BEGIN
    IF STRMID(strlowcase(atmstruct.configuration),0,1) EQ 'j' THEN BEGIN
      fwhm_Lorentzian=0.03e-4*median_wn^2/1.e4 ;in cm-1...0.03e-4 micron determined for j2
      fwhm_Gaussian=0.13e-4*median_wn^2/1.e4  ;in cm-1...0.13e-4 micron determined for j2
      fwhm_Box=0.               ;determined for j2
    ENDIF
    IF STRMID(strlowcase(atmstruct.configuration),0,1) EQ 'h' THEN BEGIN
      fwhm_Lorentzian=0.04e-4*median_wn^2/1.e4  ;in cm-1...0.04e-4 micron determined for h2
      fwhm_Gaussian=0.19e-4*median_wn^2/1.e4 ;in cm-1...0.19e-4 micron determined for h2
      fwhm_Box=0.               ;determined for h2
    ENDIF
    IF STRMID(strlowcase(atmstruct.configuration),0,1) EQ 'k' THEN BEGIN
      fwhm_Lorentzian=0.04e-4*median_wn^2/1.e4 ;in cm-1...0.04e-4 micron determined for k2
      fwhm_Gaussian=0.28e-4*median_wn^2/1.e4 ;in cm-1...0.28e-4 micron determined for k2
      fwhm_Box=0.               ;determined for k2
    ENDIF
  ENDIF

  IF float(atmstruct.slit_width) EQ 0.75 THEN BEGIN
    IF STRMID(strlowcase(atmstruct.configuration),0,1) EQ 'j' THEN BEGIN
      fwhm_Lorentzian=0.026e-4*median_wn^2/1.e4 ;in cm-1...0.026e-4 micron determined for j2
      fwhm_Gaussian=0.094e-4*median_wn^2/1.e4  ;in cm-1...0.094e-4 micron determined for j2
      fwhm_Box=0.246e-4*median_wn^2/1.e4         ;in cm-1...0.246e-4 micron determined for j2
      ;same Gaussian as 0.375 slit and then adjust box
      ;fwhm_Gaussian=0.13e-4*median_wn^2/1.e4 ;in cm-1...0.13e-4 micron determined for j2
      ;fwhm_Box=0.24d-4*median_wn^2/1.e4         ;in cm-1...0.246e-4 micron determined for j2
    ENDIF
    IF STRMID(strlowcase(atmstruct.configuration),0,1) EQ 'h' THEN BEGIN
      fwhm_Lorentzian=0.026e-4*median_wn^2/1.e4  ;in cm-1...0.026e-4 micron determined for h2
      fwhm_Gaussian=0.195e-4*median_wn^2/1.e4  ;in cm-1...0.195e-4 micron determined for h2
      fwhm_Box=0.297e-4*median_wn^2/1.e4         ;in cm-1...0.297e-4 micron determined for h2
      ;same Gaussian as 0.375 slit and then adjust box
      ;fwhm_Gaussian=0.19e-4*median_wn^2/1.e4 ;in cm-1...0.19e-4 micron determined for h2
      ;fwhm_Box=0.298e-4*median_wn^2/1.e4         ;in cm-1...0.297e-4 micron determined for h2
    ENDIF
    IF STRMID(strlowcase(atmstruct.configuration),0,1) EQ 'k' THEN BEGIN
      fwhm_Lorentzian=0.043e-4*median_wn^2/1.e4  ;in cm-1...0.043e-4 micron determined for k2
      fwhm_Gaussian=0.246e-4*median_wn^2/1.e4  ;in cm-1...0.246e-4 micron determined for k2
      fwhm_Box=0.449e-4*median_wn^2/1.e4         ;in cm-1...0.449e-4 micron determined for k2
      ;same Gaussian as 0.375 slit and then adjust box
      ;fwhm_Gaussian=0.28e-4*median_wn^2/1.e4 ;in cm-1...0.28e-4 micron determined for k2
      ;fwhm_Box=0.440e-4*median_wn^2/1.e4         ;in cm-1...0.449e-4 micron determined for k2
      ;central wavelength off by -0.015e-4 micron
   ENDIF
  ENDIF
  
  ;x values for kernel (line spread function)
  fwhm=median_wn/float(atmstruct.resolving_power) ;sigma=fwhm/(2.*SQRT(2.*ALOG(2.)))
  n=ROUND(10.*fwhm/sampling) ;define line spread function over 10 resolution elements
  lsf_waven=sampling*(FINDGEN(n)-n/2)
  x=lsf_waven
  ;y values for kernel (line spread function)  
  peak=0.
  lsf_flux = cnb_voigt(x, fwhm_Gaussian/(2.*SQRT(2.*ALOG(2.))), fwhm_Lorentzian/(2.*SQRT(2.*ALOG(2.))))
  ;convolve with box (representing slit width)
  IF fwhm_Box NE 0. THEN BEGIN
     box_flux=x*0.
     ind=where(x GE -fwhm_Box/2. AND x LE fwhm_Box/2.)
     box_flux[ind]=1.
     lsf_flux1=CONVOL(box_flux, lsf_flux, TOTAL(lsf_flux),/edge_wrap)   
     lsf_flux=lsf_flux1
  ENDIF
     
  lsf_present=1
ENDIF

IF lsf_present EQ 0 THEN BEGIN
   print, '***STOP: line spread function ', STRTRIM(atmstruct.line_spread_function), ' not implemented'
   print, '         please change line_spread_function in parameter file ', atmstruct.paramfile   
   stop
ENDIF

fluxcon=CONVOL(flux_i, lsf_flux, TOTAL(lsf_flux),/edge_wrap)

;put back on original grid
wavenout=waven_in
fluxout=interpol(fluxcon, waven_i, waven_in)

;IF keyword_set(plot) THEN BEGIN
; loadct, 2
; plot, waven, flux, xr=[min(waven),max(waven)],xs=1
; oplot, wavenout1, fluxout1,color=40
; stopcont
;ENDIF

n=n_elements(wavenout)
wavenout=reform(wavenout,1,n)
fluxout=reform(fluxout,1,n)

return, [wavenout, fluxout]

end

;
;---keep the following procedure, although it is just for testing the---
;---line spread function------------------------------------------------
;

pro test_teltool_conv_instr

  cleanplot 
  cleanplot
  window, 1

  sw='0.75';strong of iSHELL slit width in arcsec
  cf='K2' ;iSHELL configuration

  IF cf EQ 'J2' THEN cw=1.25  
  IF cf EQ 'H2' THEN cw=1.65  
  IF cf EQ 'K2' THEN cw=2.2
  
  ;create delta function
  wave_in=FINDGEN(10000.)/1000000.+cw 
  wave_in=1.e4/wave_in  ;wavenumber
  wave_in=wave_in[sort(wave_in)] ;sort
  flux_in=wave_in*0.
  flux_in[5000]=1.
  wave_peak=wave_in[5000]
  atmstruct={line_spread_function: 'measured', $
             paramfile: 'afile.txr', $
             configuration: cf, $             
             resolving_power: 80000., $
             slit_width: float(sw) $
            }

  result=teltool_conv_instr(wave_in, flux_in, atmstruct)

  ;does this fit the observed line profile?
  close, 1
  openr, 1, atmstruct.auxdir+'Stacked_'+cf+'_'+sw+'_with_uncertainties.txt'
  w_lsf=-100
  f_lsf=-100
  vals=fltarr(3)
  while (not eof(1)) do begin
     readf,1, vals
     w_lsf=[w_lsf,vals[0]]
     f_lsf=[f_lsf,vals[1]]
  endwhile
  close, 1
  ;shift to line defined above taking into account measured profile is
  ;in Angstrom and wave_peak in cm-1
  w_lsf=w_lsf[1:*]+1.e8/wave_peak
  f_lsf=f_lsf[1:*]  
  ;and convert from Angstrom to cm-1
  w_lsf=1.e4/(w_lsf*1.e-4)
  ind=sort(w_lsf)
  w_lsf=w_lsf[ind]
  f_lsf=f_lsf[ind]  

  plot, wave_in, flux_in,xr=wave_peak+[-0.2,0.2],xs=1,ps=10,yr=[0,0.1], title='Line spread function: measured (green) and functional (red)'
  oplot, result[0,*], result[1,*], color=2,ps=10
  oplot, w_lsf, f_lsf/max(f_lsf)*max(result[1,*]), color=3
  ;oplot, wave_in, mygauss(wave_in, peak=2.0050,fwhm=0.28e-4,/maxnorm)*0.034, color=3
  
  print, 'integrated original flux: ', TOTAL(flux_in)
  print, 'integrated convolved flux: ',   TOTAL(result[1,*])

  
end
  

;**************************************************
; GenerateTrans() function
; IDL driver for atmospheric transmittance synthesis - Using PUMAS
; Version 2.0 G. Villanueva - NASA GSFC - June/2017
; Example: IDL>trn=generatetrans(3025.0,3045.0)
;
; Modified by A. Boogert March 2018 for use with teltool
;**************************************************

function GenerateTrans_outputfiles, stemfile, atmstruct, fit, nmols, timestart

; GENERATETRANS_OUTPUTFILES.PRO: generate output files from PSG API
;                         ".all" file produced by generatetrans.pro
;
; dat=GENERATETRANS_OUTPUTFILES(stemfile STRING name of file
;                                               containing PSG output
;                              )
;
; output: array with the following....
;
; V0.0 03/06/2018, ACAB: taken out of generatetrans.pro

npts = file_lines(atmstruct.obsdir+stemfile + '.all') & ipt=0
trf = dblarr(nmols+2,npts) & rtrf=dblarr(nmols+2)
IF strlen(fit) NE 0 THEN ret = dblarr(4,npts) & rret=dblarr(4) ;ACAB  
rad = dblarr(2,npts)       & rrad=dblarr(2)
log = strarr(20)
cfg = strarr(1000) ;ACAB
rcfg=STRARR(1) ;ACAB
openr, 1, atmstruct.obsdir+stemfile + '.all' & type=0 & str='' & h2ocol=0.0 & amcol=1.0 & tcol=0.0
while not eof(1) do begin
  readf,1,str
  ;if the number of lines times the number of layers reaches a limit
  ;(defined remotely at PSG server) then the not all transitions are
  ;included and this spectrum is useless so give a clear warning
  IF STRMID(str,0,64) EQ 'WARNING | PUMAS | Maximum number of [lines x layers] reached for' THEN $
     print, 'ERROR: too many lines*layers. Reduce one of these. ', str
  if strcmp(str,'results',7) then begin
    if type eq 1 or type eq 2 then npts=ipt
    type = 0 & ipt=0l
    if strmid(str,strlen(str)-7,3) eq 'atm' then type=1
    if strmid(str,strlen(str)-7,3) eq 'trn' then type=2
    if strmid(str,strlen(str)-7,3) eq 'log' then type=3
    if strmid(str,strlen(str)-7,3) eq 'lyr' then type=4
    if strmid(str,strlen(str)-7,3) eq 'dat' then type=5 ;ACAB
    if strmid(str,strlen(str)-7,3) eq 'cfg' then type=6 ;ACAB    
    continue
  endif
  if type eq 4 then begin & if strcmp(str, '# Integrated atmospheric columns',32) then begin & reads,strmid(str,50),amcol,tcol,h2ocol & h2ocol=h2ocol/amcol & endif & endif
  if strcmp(str,'#',1) then continue
  if type eq 1 then begin & reads,str,rrad & rad[*,ipt]=rrad & endif 
  if type eq 2 then begin & reads,str,rtrf & trf[*,ipt]=rtrf & endif
  IF strlen(fit) NE 0  THEN BEGIN
     if type eq 5 then begin & reads,str,rret & ret[*,ipt]=rret & endif ;ACAB
  ENDIF      
  if type eq 6 then begin
     reads,str,rcfg
     cfg[ipt]=rcfg
  endif
  if type eq 3 then log[ipt]=str
  ipt = ipt + 1
endwhile
close,1

rad=reverse(rad[*,0:npts-1],2)
trf=reverse(trf[*,0:npts-1],2)
IF strlen(fit) NE 0 THEN ret=reverse(ret[*,0:npts-1],2)  ;ACAB

; Store the results in compatible format
; 0: Frequency vector
; 6: Fully resolved radiance
; 7: Convolved transmittance
; 8: Convolved radiance
; 9: Fully resolved transmittance
; 10-9+nmols: Convolved transmittance per molecule
dat=dblarr(10+nmols,npts)
dat[0,*] = rad[0,*]
dat[6,*] = rad[1,*]
dat[7,*] = trf[1,*];ACAB
dat[8,*] = rad[1,*];ACAB
dat[9,*] = trf[1,*]
dat[10:10+nmols-1,*] = trf[2:nmols+2-1,*]
openw, 1, atmstruct.obsdir+stemfile+'.trf'
printf,1,trf[0:nmols+2-1,*], FORMAT='('+STRTRIM(2+nmols,2)+'f10.6)'
close,1
;ACAB
IF strlen(fit) NE 0 THEN BEGIN
  openw, 1, atmstruct.obsdir+stemfile+'.ret'
  printf,1,ret[0:3,*]
  close,1
ENDIF
;ACAB
openw, 1, atmstruct.obsdir+stemfile+'.cfg'
FOR i=0,n_elements(cfg)-1 DO BEGIN
   printf,1, cfg[i]
ENDFOR
close,1

if atmstruct.convolve EQ 'yes' AND strlen(fit) EQ 0 then begin
  ;convolve with instrument line spread function
  ;out1=conv_gauss(trf[0,*],trf[1,*],median(trf[0,*])/float(atmstruct.resolving_power)) ;ACAB
  out=teltool_conv_instr(trf[0,*],trf[1,*], atmstruct) ;ACAB
  trn=out[0:1,*]                ;ACAB
  openw, 1, atmstruct.obsdir+stemfile+'.trn'
  printf,1,trn
  close,1
  dat[7,*] = interpol(trn[1,*], trn[0,*], dat[0,*])
  ;do the same for transmission by each molecule
  FOR nm=0,nmols-1 DO BEGIN
     ;out=conv_gauss(trf[0,*],trf[nm+2,*],median(trf[0,*])/float(atmstruct.resolving_power)) ;ACAB
     out=teltool_conv_instr(trf[0,*],trf[nm+2,*], atmstruct) ;ACAB
     trn=out[0:1,*]                
     dat[10+nm,*] = interpol(trn[1,*], trn[0,*], dat[0,*])
  ENDFOR
  ;and now for emission spectrum
  ;out=conv_gauss(rad[0,*],rad[1,*],median(rad[0,*])/float(atmstruct.resolving_power)) ;ACAB
  out=teltool_conv_instr(rad[0,*], rad[1,*], atmstruct) ;ACAB 
  rdf=out[0:1,*]                ;ACAB    
  openw, 1, atmstruct.obsdir+stemfile+'.rad'
  printf,1,rad
  close,1
  dat[8,*] = interpol(rdf[1,*], rdf[0,*], dat[0,*])
endif 

;; --------------------------------------------------
;; LOG - Create a log file
;; --------------------------------------------------
;logfile = string(atmstruct.obsdir+stemfile+'.log')
;openw, 1, atmstruct.obsdir+logfile
;printf,1,'---------------------------------------------------------------'  
;printf,1,'PSG/PUMAS Transmittance model, ',systime()
;file_delete,'uid.txt',/ALLOW_NONEXISTENT & usrname=''
;spawn,'whoami > uid.txt'
;openr, 2, atmstruct.obsdir+'uid.txt'
;readf,2,usrname
;close,2
;file_delete,'uid.txt',/ALLOW_NONEXISTENT
;printf,1,'User: ',usrname
;printf,1,'Radiative transfer took [secs]', systime(1)-timestart
;printf,1,npts,dat(0,0),dat(0,npts-1),format='(I-8," points were calculated between ",F9.2," and ",F9.2," cm-1")'
;printf,1,'Synthesis resolving power:',round(dat[0,0]/(dat[0,1]-dat[0,0]))
;printf,1,'---------------------------------------------------------------'  
;printf,1,'Airmass:',airmass
;printf,1,'Telescope altitude [m]:',altitude
;printf,1, fH2O, fCO2,fO3,fN2O,fCO,fCH4, format='("Abundances, H2O, CO2, O3, N2O, CO, CH4:", 6F7.4)'
;printf,1, ff[5],ff[6],ff[7],            format='("Abundances, O2, NH3, C2H6:", 3F7.4)'
;printf,1,'---------------------------------------------------------------'  
;printf,1,'Precipitable millimeter of water (PWV):', h2ocol/3.34d25
;printf,1,'---------------------------------------------------------------'  
;printf,1,'Fully-resolved transmittance spectrum: ',string(atmstruct.obsdir+stemfile,'.trf')
;printf,1,'Convolved transmittance spectrum: ',string(atmstruct.obsdir+stemfile,'.trn')
;printf,1,'Convolved radiance spectrum: ',string(atmstruct.obsdir+stemfile,'.rad')
;if showplot gt 0 then printf,1,'Postscript plot of the transmittance spectra: ',string(atmstruct.obsdir+stemfile,'.ps')
;printf,1,'---------------------------------------------------------------'  
;close,1

return, dat

end

;----------------------------------------------------------------

pro GenerateTrans_plot  

; GENERATETRANS_PLOT.PRO: plot output produced by generatetrans.pro
;
; V0.0 03/06/2018, ACAB: taken out of generatetrans.pro

; --------------------------------------------------
; Plot the results
; --------------------------------------------------
if showplot gt 0 then begin

  ; Plot of molecules
  mingas = dblarr(nmols) & for i=0,nmols-1 do mingas[i] = min(trf[i+2,*])
  active = where(mingas lt 0.95, nactive)
  asort  = sort(mingas[active])
  offs   = where(mingas[active] gt 0.1, noff)
  
  entry_device = !d.name
  tvlct,tvr,tvg,tvb,/get
  for ps=0,1 do begin
    if ps eq 1 then begin
      pm=4
      set_plot, 'PS'
      psfile = atmstruct.obsdir+string(stemfile+'.ps')
      device, filename=psfile,/helvetica
      device, XOFFSET=pm, YOFFSET=pm, XSIZE=7*pm, ysize=4*pm, /color, /ENCAPSULATED, /PORTRAIT
      !p.font=0
      !p.thick=2.0
      !p.charsize=1.0
      pcl=0 & pcn='black'
    endif else begin
      ;window,1,xs=700,ys=400 ;ACAB commented out
      !p.font=-1
      !p.charsize=1.0
      !p.thick=0.5
      device,decomposed=0
      pcl=255 & pcn='white'
    endelse
    
    ; Create 16-colors pallette
    colarr=strarr(17)
    tvlct,pcl,pcl,pcl,01   &  colarr(01)=pcn
    tvlct,255,0,0    ,02   &  colarr(02)='Red'
    tvlct,0,200,0    ,03   &  colarr(03)='Green'
    tvlct,0,0,255    ,04   &  colarr(04)='Blue'
    tvlct,200,200,0  ,05   &  colarr(05)='Olive'
    tvlct,255,0,255  ,06   &  colarr(06)='Fuchsia'
    tvlct,0,255,0    ,07   &  colarr(07)='Lime'
    tvlct,192,192,192,08   &  colarr(08)='Silver'
    tvlct,128,128,128,09   &  colarr(09)='Gray'
    tvlct,255,255,0  ,10   &  colarr(10)='Yellow'
    tvlct,128,0,128  ,11   &  colarr(11)='Purple'
    tvlct,128,0,0    ,12   &  colarr(12)='Maroon'
    tvlct,0,255,255  ,13   &  colarr(13)='Aqua'
    tvlct,0,255,0    ,14   &  colarr(14)='Lime'
    tvlct,0,128,128  ,15   &  colarr(15)='Teal'
    tvlct,0,0,128    ,16   &  colarr(16)='Navy'
    
    spc=0.1
    plot,dindgen(10), xrange=[wn2,wn1],xs=1, /nodata, yrange=[0.0, 1.0+(noff+1)*spc],ys=1, $
          title='PSG/PUMAS: Terrestrial transmittance spectra (convolved is ' + pcn + ')',$ 
          thick=!p.thick, xthick=!p.thick, ythick=!p.thick, $
          xtitle='Wavenumber [cm-1]', ytitle='Transmittance'
    oplot,dat[0,*], dat[7,*], color=1, thick=!p.thick*2.0 & ioff=0 & ibase=0
    for i=0,nactive-1 do begin
      id = active[asort[i]]
      icolor=(i mod 15) + 2
      if mingas[id] lt 0.1 then begin 
        ibase = ibase + 1
        ypos=0d
        xpos=wn2-(wn2-wn1)*ibase/20d
      endif else begin
        ioff = ioff + 1
        ypos=ioff*spc
        xpos=wn2-(wn2-wn1)/20d
      endelse
      oplot,trf[0,*],trf[id+2,*] + ypos,color=icolor
      xyouts,xpos,1.03 + ypos,mols[id],color=icolor
    endfor
    if ps eq 1 then device, /close_file
  endfor
  set_plot, entry_device
  tvlct,tvr,tvg,tvb
  
endif 
  
end

;----------------------------------------------------------------

function GenerateTrans,   $
               wn1, wn2,  $ ; Wavenumber range [cm-1]
               lineatlas, $ ; O Flag for calculating the line-atlas and freq. vector (O:optional)
               airmass,   $ ; O Atmospheric airmass
               altitude,  $ ; O Altitude of the telescope [m], or string with "YYYY/MM/DD-HH:MM,LAT,LON" (e.g."2016/04/01-10:00,-29.90453,-71.24894")
               respower,  $ ; O Resolving power of the instrument
               T1,        $ ; O T1 factor
               T2,        $ ; O T2 factor
               fP,        $ ; O Pressure factor
               fH2O,      $ ; O Water content factor relative to a tropical atm.
               fCO2,      $ ; O Carbon dioxide factor relative to a tropical atm.
               fO3,       $ ; O Ozone factor relative to a tropical atm.                     
               fN2O,      $ ; O Nitrous oxide factor relative to a tropical atm.
               fCO,       $ ; O Carbon monoxide factor relative to a tropical atm.
               fCH4,      $ ; O Methane factor relative to a tropical atm.
               fOCS,      $ ; O OCS factor relative to a tropical atm.
               npoints,   $ ; O number of points
               stemfile,  $ ; O Stem of output files
               showplot,  $ ; O Show plots?
               ff,        $ ; O Vector with fitting factors
               atmstruct, $ ; ACAB: IDL structure to transfer other variables
               fit,       $ ; ACAB: fit column of this molecule (e.g., 'H2O'), if atmstruct.fit_method='remote'
               fitvalue     ; ACAB: initial column of molcule 'fit', if atmstruct.fit_method='remote'

  ;ACAB: initial running code generates file pumas.atm, showing that
  ;PSG server has moved
  ;psgserver = 'https://ssed.gsfc.nasa.gov/psg'
  psgserver = 'https://psg.gsfc.nasa.gov/'
  ;psgserver = 'http://localhost:8888'
  cd,current=mydir
  timestart=systime(1)
  
  ; Set defaults
  if n_elements(airmass)    eq 0 then airmass=2.0
  if n_elements(altitude)   eq 0 then altitude='4200.0'
  if n_elements(respower)   eq 0 then respower=25000.0
  if n_elements(T1)         eq 0 then T1=1.0
  if n_elements(fP)         eq 0 then fP=1.0
  if n_elements(fH2O)       eq 0 then fH2O=1.0
  if n_elements(fCO2)       eq 0 then fCO2=1.0
  if n_elements(fO3)        eq 0 then fO3=1.0
  if n_elements(fN2O)       eq 0 then fN2O=1.0
  if n_elements(fCO)        eq 0 then fCO=1.0
  if n_elements(fCH4)       eq 0 then fCH4=1.0
  if n_elements(fOCS)       eq 0 then fOCS=1.0  
  if n_elements(stemfile)   eq 0 then stemfile='pumas1'
  if n_elements(showplot)   eq 0 then showplot=1
  if n_elements(ff)         lt 2 then ff=dblarr(10)+1.0
  if n_elements(fit)        eq 0 then fit=''
  
  ; Establish the atmosphere for this run
  if file_test(atmstruct.obsdir+stemfile + '.atm',/read) eq 0 or file_test(atmstruct.obsdir+stemfile + '.atm',/zero_length) eq 1 then begin
    ;ACAB
    print, '***creating new atmosphere (latitude, longitude, altitude, date) for this run (file: '+atmstruct.obsdir+stemfile+'.atm)' ; ACAB
    if showplot then print,'PSG/PUMAS: Extracting atmospheric information ...'
    lat='-29.9045300' & lon='-71.2489400' & date='2016/04/01-10:00' & alt=strtrim(string(altitude/1d3),2)
    if size(altitude, /type) eq 7 then begin
      sst = strsplit(altitude,',',/extract, count=count)
      if count ge 3 then begin 
        date=strtrim(sst[0],2) 
        lat=strtrim(sst[1],2) 
        lon=strtrim(sst[2],2)
        if count ge 4 then alt=strtrim(sst[3],2) else alt='0.0'
      endif
    endif
    openw, 2, atmstruct.obsdir+stemfile + '.cfg'
    printf,2,'<OBJECT>Earth'
    printf,2,'<OBJECT-DATE>' + date
    printf,2,'<OBJECT-OBS-LATITUDE>' + lat
    printf,2,'<OBJECT-OBS-LONGITUDE>' + lon
    printf,2,'<GEOMETRY>Lookingup'
    printf,2,'<GEOMETRY-OBS-ALTITUDE>' + alt
    printf,2,'<GEOMETRY-ALTITUDE-UNIT>km'
    ;not working yet
    ;printf,2,'<ATMOSPHERE-LAYERS>' + strtrim(atmstruct.nlayers,2) ;default is 72
    close,2
    ccommand='curl -d key='+atmstruct.psg_key+' -d type=cfg -d watm=y --data-urlencode file@' + atmstruct.obsdir+stemfile + '.cfg ' + psgserver + '/api.php > ' + atmstruct.obsdir+stemfile + '.atm 2>/dev/null'
    ;ccommand='curl -d type=cfg -d watm=y --data-urlencode file@' + atmstruct.obsdir+stemfile + '.cfg ' + psgserver + '/api.php > ' + atmstruct.obsdir+stemfile + '.atm 2>/dev/null'
    IF atmstruct.debug EQ 'yes' THEN print, ccommand
    print, '***calculating atmosphere'
    spawn, ccommand, /sh
    if file_test(atmstruct.obsdir+stemfile + '.atm',/read) eq 0 or file_test(atmstruct.obsdir+stemfile + '.atm',/zero_length) eq 1 then begin
      print,'PSG/PUMAS: No internet connection / failure to derive atmospheric information.'
      stop
    endif
  endif else begin
     ;ACAB
     IF atmstruct.debug EQ 'yes' THEN print, '***using existing atmosphere (latitude, longitude, altitude date) for this run (file: '+atmstruct.obsdir+stemfile+'.atm)' 
  endelse

  ; Modify the molecular abundances and prepare simulation
  mols=atmstruct.psg_molecules  
  ;    ['H2O','CO2','O3','N2O','CO','CH4','O2',  'NH3', 'HCl', 'N2','C2H6', 'OCS']
  moli=[ 1,    2,    3,   4,    5,   6,     7,    11,    15,   22,   27,     19  ]
  molx=[fH2O, fCO2, fO3, fN2O, fCO, fCH4, ff[5], ff[6], ff[8], 1.0, ff[7],  fOCS ]
  sgas='' & stype='' & sabun='' & sunit='' & nmols=n_elements(mols)
  for i=0,nmols-1 do begin
    if i eq nmols-1 then sp='' else sp=','
    sgas  = sgas  + mols[i] + sp
    stype = stype + 'HIT[' + strtrim(string(moli[i]),2) + ']' + sp
    sabun = sabun + strtrim(string(molx[i]),2) + sp
    sunit = sunit + 'scl' + sp
  endfor
  file_copy,atmstruct.obsdir+stemfile + '.atm',atmstruct.obsdir+stemfile + '.cfg',/overwrite
  openw, 2, atmstruct.obsdir+stemfile + '.cfg',/append
  printf,2,'<ATMOSPHERE-NGAS>' + strtrim(n_elements(mols),2)
  printf,2,'<ATMOSPHERE-GAS>' + sgas
  printf,2,'<ATMOSPHERE-TYPE>' + stype
  printf,2,'<ATMOSPHERE-ABUN>' + sabun
  printf,2,'<ATMOSPHERE-UNIT>' + sunit
  ;,NOT WORKING YET
  ;printf,2,'<ATMOSPHERE-LAYERS>' + strtrim(atmstruct.nlayers,2) ;defaultis 72
  printf,2,'<GEOMETRY-OBS-ANGLE>' + strtrim(string((180d/!PI)*acos(1.0/airmass)),2)
  printf,2,'<GENERATOR-RANGE1>' + strtrim(string(wn1),2)
  printf,2,'<GENERATOR-RANGE2>' + strtrim(string(wn2),2)
  printf,2,'<GENERATOR-RANGEUNIT>cm'
  printf,2,'<GENERATOR-RESOLUTION>' + strtrim(string(respower),2)
  printf,2,'<GENERATOR-RESOLUTIONUNIT>RP'
  printf,2,'<GENERATOR-CONT-MODEL>Y' ;important for generating transmission spectrum
  printf,2,'<GENERATOR-CONT-STELLAR>N'
  printf,2,'<GENERATOR-RADUNITS>rif' ;normalize continuum 
  close,2

  ;ACAB: added retrieval 26/Feb/2018
  IF strlen(fit) NE 0 THEN BEGIN
     openw, 2, atmstruct.obsdir+stemfile + '.cfg',/append
     ;generate transmission spectrum toward an object
     printf,2,'<GEOMETRY>'+'LookingObject'
     ;but what this continuum emitting object is, is not important as
     ;transmission spectrum will be normalized using
     printf,2,'<GEOMETRY-STELLAR-TYPE>'+'U'
     printf,2,'<GEOMETRY-STELLAR-TEMPERATURE>'+strtrim(string(6000.),2)
     printf,2,'<GEOMETRY-STELLAR-MAGNITUDE>'+'0'
     printf,2,'<GENERATOR-RESOLUTION>'+strtrim(string(atmstruct.resolving_power),2) 
     printf,2,'<RETRIEVAL-RANGEUNIT>'+strtrim(string('cm'),2) ;wavenumber
     printf,2,'<RETRIEVAL-RESOLUTION>'+strtrim(string(atmstruct.resolving_power),2) 
     printf,2,'<RETRIEVAL-RESOLUTIONUNIT>'+'RP'  ;resolution as resolving power
     printf,2,'<RETRIEVAL-FLUXSCALER>'+strtrim(string(1.0),2)
     printf,2,'<RETRIEVAL-FITTELLURIC>'+'N'
     printf,2,'<RETRIEVAL-FITGAIN>'+strtrim(string(-1),2)
     printf,2,'<RETRIEVAL-REMOVEOFFSET>'+strtrim(string(-1),2)
     printf,2,'<RETRIEVAL-REMOVEFRINGE>'+strtrim(string(0),2)
     printf,2,'<RETRIEVAL-FITSTELLAR>'+'N'
     printf,2,'<RETRIEVAL-FITFREQ>'+'Y'  ;ACAB: we will want a flag to optionally refine wavelength calibration
     printf,2,'<RETRIEVAL-FITRESOLUTION>'+'N' ;ACAB: would be interesting to see if this works
     printf,2,'<RETRIEVAL-FREQSHIFT>'+strtrim(string(0.000e+00),2)
     printf,2,'<RETRIEVAL-NVARS>'+strtrim(string(1),2) ;ACAB, if we only fir the H2O column, then NVARS=1
     printf,2,'<RETRIEVAL-FLUXLABELS>'+strtrim(string('Data,Model'),2) ;ACAB:is this for the output data file?
     printf,2,'<RETRIEVAL-VARIABLES>'+strtrim(string('ATMOSPHERE-'+fit),2) ;ACAB: options are atmosphere-H2O, atmosphere-CH4, atmosphere-CO2 [not allowed here are atmosphere-pressure, surface-temperature, surface-albedo, surface-emissivity]
     printf,2,'<RETRIEVAL-VALUES>'+strtrim(string(fitvalue),2) ;ACAB initial values for variables to be fitted
     printf,2,'<RETRIEVAL-MIN>'+strtrim(string(0.05),2) ;ACAB: lower boundary of variable permitted
     printf,2,'<RETRIEVAL-MAX>'+strtrim(string(5),2) ;ACAB: upper boundary of variable permitted
     printf,2,'<RETRIEVAL-UNITS>'+strtrim(string('scl'),2) ;ACAB: units of initial and boundary values entered (scaler, ppm etc)
     printf,2,'<RETRIEVAL-STATUS>'+strtrim(string('OK'),2) ;ACAB: not sure what's this for, values could be OK, RUNNING, and ...
     printf,2,'<DATA>'
     ;1.640180e+0 1.002222e+0 3.933286e-2
     FOR i=0,n_elements(atmstruct.fit_wavenumber)-1 DO BEGIN
       printf, 2, STRTRIM(atmstruct.fit_wavenumber[i],2)+' '+STRTRIM(atmstruct.fit_flux[i],2)
     ENDFOR
     printf,2,'</DATA>'
     close,2
  ENDIF
  ; Call the program and read results
  if showplot gt 0 then print,'PSG/PUMAS: Computing radiative transfer ...'
  IF strlen(fit) NE 0 THEN BEGIN
     cmd = 'curl -d key='+atmstruct.psg_key+' -d wcon=n -d wgeo=n -d type=ret --data-urlencode file@' + atmstruct.obsdir+stemfile + '.cfg ' + psgserver + '/api.php > ' + atmstruct.obsdir+stemfile + '.all 2>/dev/null'
     ;cmd = 'curl -d wcon=n -d wgeo=n -d type=ret --data-urlencode file@' + atmstruct.obsdir+stemfile + '.cfg ' + psgserver + '/api.php > ' + atmstruct.obsdir+stemfile + '.all 2>/dev/null'
     IF atmstruct.debug EQ 'yes' THEN print, cmd
     spawn, cmd, /sh
     if file_test(atmstruct.obsdir+stemfile + '.all',/read) eq 0 or file_test(atmstruct.obsdir+stemfile + '.all',/zero_length) eq 1 then begin
       print,'PSG/PUMAS: No internet connection / failure to compute transmittances.'
       stop
     endif
     dat=GenerateTrans_outputfiles(stemfile, atmstruct, fit, nmols, timestart)
  ENDIF ELSE BEGIN
     cmd = 'curl -d key='+atmstruct.psg_key+' -d wcon=n -d wgeo=n -d type=all --data-urlencode file@' + atmstruct.obsdir+stemfile + '.cfg ' + psgserver + '/api.php > ' + atmstruct.obsdir+stemfile + '.all 2>/dev/null'
     ;cmd = 'curl -d wcon=n -d wgeo=n -d type=all --data-urlencode file@' + atmstruct.obsdir+stemfile + '.cfg ' + psgserver + '/api.php > ' + atmstruct.obsdir+stemfile + '.all 2>/dev/null'
     IF atmstruct.debug EQ 'yes' THEN print, cmd
     spawn, cmd, /sh
     if file_test(atmstruct.obsdir+stemfile + '.all',/read) eq 0 or file_test(atmstruct.obsdir+stemfile + '.all',/zero_length) eq 1 then begin
        print,'PSG/PUMAS: No internet connection / failure to compute transmittances.'
       stop
     endif
     dat=GenerateTrans_outputfiles(stemfile, atmstruct, fit, nmols, timestart)
  ENDELSE
  ;GenerateTrans_plot    

  ;print,'PSG/PUMAS: number of points and precision [cm-1]: ', npts, abs(dat[0,0]-dat[0,1]),format='(A, I8, F9.5)'
  ;print,'PSG/PUMAS: Radiative transfer took [secs]: ', systime(1)-timestart  
  ;print,'PSG/PUMAS: Active molecules within range: ', mols[active[asort]]

  return, dat 

end

;----------------------------------------------------------------

function get_psg, atmstruct, fit, fitvalue

;define observatory parameters
IF atmstruct.observatory EQ "NASA IRTF" THEN BEGIN
    ;IRTF/Mauna Kea: 19 49 34.38594
    ;(http://irtfweb.ifa.hawaii.edu/IRrefdata/telescope_ref_data.php)
    latitude=19.826218316666665
    
    ;IRTF/Mauna Kea: 155 28 19.19564
    ;(http://irtfweb.ifa.hawaii.edu/IRrefdata/telescope_ref_data.php)
    longitude=155.4719987888889

    ;IRTF/Mauna Kea: 13674.76 feet= 4.1680668 km
    ;(http://irtfweb.ifa.hawaii.edu/IRrefdata/telescope_ref_data.php)
    ;give at most 1 meter accuracy
    obs_altitude=4.168
ENDIF ELSE BEGIN
   print, '***Observatory not defined. Cannot run PSG.***'
   stop
ENDELSE

; psggenerator parameter 'altitude' defined as string
; "YYYY/MM/DD-HH:MM,LAT,LON"
; (e.g."2016/04/01-10:00,-29.90453,-71.24894")
; ...although I think altitude may be the last part, so
;actually "YYYY/MM/DD-HH:MM,LAT,LON,ALT"
;(e.g."2016/04/01-10:00,-29.90453,-71.24894,4200.")
;Note that atmstruct.utc has format '2016-09-16T14:30:00.00'
;Also not that psggenerator instructions say altitude is in meter,
;while it's actually in km in <GEOMETRY-ALTITUDE-UNIT>
altitude=STRMID(atmstruct.utc,0,4)+'/'+STRMID(atmstruct.utc,5,2)+'/'+ $
         STRMID(atmstruct.utc,8,2)+'-'+STRMID(atmstruct.utc,11,5)+','+ $
         STRTRIM(latitude,2)+','+STRTRIM(longitude,2)+','+STRTRIM(obs_altitude,2)

;get model spectrum from PSG

;get models from GSFC's Planetary Spectrum Generator (PSG)
;this part of the code was provided by Geronimo Villanueva

;convert atmstruct input parameters to PSG's generatetrans input
;parameters. Defaults are 0
; Wavenumber range [cm-1]
wn1=atmstruct.wavenumber[0]
wn2=atmstruct.wavenumber[1]
; Flag for calculating the line-atlas and freq. vector (O:optional)
lineatlas=1
; Atmospheric airmass
;airmass=1./COS(atmstruct.za/180.*!PI)
airmass=atmstruct.airmass
; Altitude of the telescope [m], or string with "YYYY/MM/DD-HH:MM,LAT,LON" (e.g."2016/04/01-10:00,-29.90453,-71.24894")
; ...I think for string, altitude may be the last part, so
;actually "YYYY/MM/DD-HH:MM,LAT,LON,ALT" (e.g."2016/04/01-10:00,-29.90453,-71.24894, 4200") 
altitude=altitude
;Resolving power to use in PSG...give something very high, as we
;will convolve with line spread function ourselves
;For R=5.e5, convolution with R=1.e5 yields 0.98e5, i.e., 2% reduction
respower=atmstruct.resolving_power_pre
; T1 factor
T1=1.0
; T2 factor
T2=1.0
; Pressure factor
fP=1.0
; molecular abundance scaling factors
molecules_in=atmstruct.molecules
scale_molecules_in=atmstruct.scale_molecules
;molecules that PSG is able to use (more can easily be added,
;though. The actually used molecules are defined in
;atmstruct.molecules, though. Anything not in atmstruct.molecules will
;not be included in the model. This part of the code needs improvement.
;IMPORTANT: DEFINITION NEEDS TO MATCH PSG_MOLECULES IN GENERATETRANS!
psg_molecules=['H2O','CO2','O3','N2O','CO','CH4','O2',  'NH3', 'HCl', 'N2','C2H6', 'OCS']
;store the molecules that PSG generatetrans uses for later use
IF tag_exist(atmstruct, 'psg_molecules') EQ 0 THEN BEGIN
  atmstruct2={psg_molecules : psg_molecules}
  atmstruct1=atmstruct
  atmstruct=create_struct(atmstruct1, atmstruct2)
ENDIF
psg_scale_molecules=FLTARR(n_elements(psg_molecules))
FOR i=0,n_elements(psg_molecules)-1 DO BEGIN
   index=where(molecules_in EQ psg_molecules[i], count)
   IF count NE 0 THEN psg_scale_molecules[i]=scale_molecules_in[index]
ENDFOR

; Water content factor relative to a tropical atm.
fH2O=psg_scale_molecules[0]
; Carbon dioxide factor relative to a tropical atm.
fCO2=psg_scale_molecules[1]
; Ozone factor relative to a tropical atm.                     
fO3=psg_scale_molecules[2]
; Nitrous oxide factor relative to a tropical atm.
fN2O=psg_scale_molecules[3]
; Carbon monoxide factor relative to a tropical atm.
fCO=psg_scale_molecules[4]
; Methane factor relative to a tropical atm.
fCH4=psg_scale_molecules[5]
; OCS factor relative to a tropical atm.
fOCS=psg_scale_molecules[11]
; number of points (not used by generatetrans)
; !NULL not defined in IDL 7.1
npoints=0
; Stem of output files
; !NULL not defined in IDL 7.1
stemfile=atmstruct.stemfile
; Show plots?
showplot=atmstruct.do_plot EQ 4

IF atmstruct.debug EQ 'yes' THEN print,  fH2O,      $ ; O Water content factor relative to a tropical atm.
        fCO2,      $ ; O Carbon dioxide factor relative to a tropical atm.
        fO3,       $ ; O Ozone factor relative to a tropical atm.                     
        fN2O,      $ ; O Nitrous oxide factor relative to a tropical atm.
        fCO,       $ ; O Carbon monoxide factor relative to a tropical atm.
        fCH4,      $ ; O CH4 factor relative to a tropical atm.    
        fOCS         ; O OCS factor relative to a tropical atm.    
                                         
atmod=generatetrans(wn1, wn2,  $ ; Wavenumber range [cm-1]
        lineatlas, $ ; O Flag for calculating the line-atlas and freq. vector (O:optional)
        airmass,   $ ; O Atmospheric airmass
        altitude,  $ ; O Altitude of the telescope [m], or string with "YYYY/MM/DD-HH:MM,LAT,LON" (e.g."2016/04/01-10:00,-29.90453,-71.24894")
        respower,  $ ; O Resolving power of the instrument
        T1,        $ ; O T1 factor
        T2,        $ ; O T2 factor
        fP,        $ ; O Pressure factor
        fH2O,      $ ; O Water content factor relative to a tropical atm.
        fCO2,      $ ; O Carbon dioxide factor relative to a tropical atm.
        fO3,       $ ; O Ozone factor relative to a tropical atm.                     
        fN2O,      $ ; O Nitrous oxide factor relative to a tropical atm.
        fCO,       $ ; O Carbon monoxide factor relative to a tropical atm.
        fCH4,      $ ; O Methane factor relative to a tropical atm.
        fOCS,      $ ; O OCS factor relative to a tropical atm.                    
        npoints,   $ ; O number of points
        stemfile,  $ ; O Stem of output files
        showplot,  $ ; O Show plots?
        0.,        $ ; ACAB: some other variables, TBD
        atmstruct, $ ; ACAB: propagate other variables
        fit,       $ ; ACAB: fit column of this molecule (e.g., 'H2O'), but only if atmstruct.fit_method='remote'        
        fitvalue)    ; ACAB: initial column of molecule 'fit'

return, atmod

end

;----------------------------------------------------------------

function get_param_from_stemfile, stemfile, param, atmstruct

;GET_PARAM_FROM_STEMFILE.PRO: extract a variable from PSF output
;                             stemfile.all
;
;Calling: values=get_param_from_stemfile(variable STRING e.g., <RETRIEVAL-VALUES>
;                                       )        
;
;Output: array of values

nchar=strlen(param)+2

values='nothing'
str=''
openr, 1, atmstruct.obsdir+stemfile + '.all'
while (not eof(1)) AND (values EQ 'nothing') do begin
  readf,1,str
  IF strcmp(str,'<'+param+'>',nchar) THEN BEGIN
    values=STRMID(str,nchar,strlen(str)-nchar)
  ENDIF
endwhile
close,1

return, values

end

;----------------------------------------------------------------

function get_lblrtm, atmstruct

;get model spectrum from LBLRTM

;spawn ....

atmod=1
  
return, atmod

end

;----------------------------------------------------------------

function teltool_normlineflux, atmstruct, wave, flux, error, order, w_fit, region, contin, nfit, c_reg1=c_reg1, c_reg2=c_reg2, w_reg=w_reg, const=const, rico=rico

  ;TELTOOL_NORMLINEFLUX.PRO: fit baseline and use that to normalize
  ;                            flux of a given spectrum
  ;
  ;Calling: out=teltool_normlineflux(atmstruct, 
  ;
  ;            wave,    wavelengths of spectrum (cm-1)
  ;
  ;            flux,    flux of spectrum (any unit)
  ;
  ;            error,   uncertainty of spectrum (same unit as flux)
  ;
  ;            order,   echelle order
  ;
  ;            w_fit,   central wavelengths of lines (cm-1)
  ;
  ;            region,  region left and
  ;                     right of w_fit[nfit] to select
  ;                     (in resolution units)
  ;
  ;            contin,  start of
  ;                     continuum left and right
  ;                      of w_fit[nfit] (in resolution units)
  ;
  ;            nfit,    index of w_fit to use
  ;
  ;            c_reg1=c_reg1, optional output left continuum range
  ;
  ;            c_reg2=c_reg2, optional output right continuum range
  ;
  ;            const=const,  optional output constant of straight line baseline
  ;
  ;            rico=rico   optional output slope of straight line continuum
  ;            )
  ; 
  ;
  ;
  ;Output: normalized flux over only the baseline-fitted wavelength
  ;        range, as in
  ;
  ;            wave_norm=out[0,*]
  ;            flux_norm=out[1,*]  input flux/baseline
  ;            error_norm=out[2,*]  input error/baseline
  ;            flux=out[3,*]  input flux for wave_norm
  ;            error=out[4,*]  input error for wave_norm
  ;            continuum=out[5,*]  baseline for wave_norm
  ;
  ;
  ; v0.0 04/18/2018 A. Boogert: created
  ; v0.1 12/04/2018 A. Boogert: added error propagation
  ;  
  
  w_reg=[w_fit[nfit]-region[nfit]*w_fit[nfit]/float(atmstruct.resolving_power), w_fit[nfit]+region[nfit]*w_fit[nfit]/float(atmstruct.resolving_power)]
  atmstruct.wavenumber=w_reg
  c_reg1=[w_reg[0], w_fit[nfit]-contin[nfit]*w_fit[nfit]/float(atmstruct.resolving_power)]
  c_reg2=[w_fit[nfit]+contin[nfit]*w_fit[nfit]/float(atmstruct.resolving_power), w_reg[1]]      
  index=where(wave GT atmstruct.wavenumber[0] AND wave LT atmstruct.wavenumber[1], count)
  ;if 2 orders overlap at these wavelengths, take the one
  ;with the most signal. We assume that not more than 2
  ;orders will overlap
  ord=(order[index])[UNIQ(order[index], SORT(order[index]))]
  IF n_elements(ord) EQ 2 THEN BEGIN
     index1=where(wave GT atmstruct.wavenumber[0] AND wave LT atmstruct.wavenumber[1] AND order EQ ord[0])
     index2=where(wave GT atmstruct.wavenumber[0] AND wave LT atmstruct.wavenumber[1] AND order EQ ord[1])                  
     ordrat=MEDIAN(flux[index1])/MEDIAN(flux[index2])
     IF ordrat GT 1. THEN do_ord=ord[0] ELSE do_ord=ord[1]
  ENDIF ELSE do_ord=ord
  index=where(wave GT atmstruct.wavenumber[0] AND wave LT atmstruct.wavenumber[1] AND order EQ do_ord, count)
  med1=MEDIAN(flux[where(wave GT c_reg1[0] AND wave LT c_reg1[1] AND order EQ do_ord)])
  med2=MEDIAN(flux[where(wave GT c_reg2[0] AND wave LT c_reg2[1] AND order EQ do_ord)])
  wmed1=MEDIAN(wave[where(wave GT c_reg1[0] AND wave LT c_reg1[1] AND order EQ do_ord)])
  wmed2=MEDIAN(wave[where(wave GT c_reg2[0] AND wave LT c_reg2[1] AND order EQ do_ord)])
  rico=(med2-med1)/(wmed2-wmed1)
  const=med1-wmed1*rico
  cont=poly(wave[index],[const,rico])
  ;cont=medsmooth(flux[index],15*round((MEDIAN(wave[index])/float(atmstruct.resolving_power))/channelwidth))
  ;cont=contfit(wave[index],flux[index],2)
  flux_norm=flux[index]/cont
  error_norm=error[index]/cont  

  out=transpose([[wave[index]], [flux_norm], [error_norm], [flux[index]], [error[index]], [cont]])
  
return, out                     

end
                     
;----------------------------------------------------------------
                 

function fit_atmod_func, p

; FIT_BGSTAR_FUNC.PRO: function to be used in amoeba minimization of
;     atmospheric model spectra to a stellar observation. It is called
;     from teltool.pro
;

common teltool_commonblock, atmstruct, indexmol, atmod

;sanity check
IF p[0] GT 0. THEN BEGIN

  atmstruct.scale_molecules[indexmol]=p[0]

  IF atmstruct.server EQ 'psg' THEN atmod=get_psg(atmstruct)
  IF atmstruct.server EQ 'lblrtm' THEN atmod=get_lblrtm(atmstruct)

  ;difference in peak depth should be minimized, but we're happy once
  ;we're within 0.5% of peak depth
  diff=ABS(MIN(atmstruct.fit_flux)-MIN(atmod[7,*]))
  rat=MIN(atmstruct.fit_flux)/MIN(atmod[7,*])
  IF rat GT 0.995 AND rat LT 1.005 THEN diff=0.

  IF atmstruct.debug EQ 'yes' THEN print, p[0], diff, rat
  
ENDIF ELSE diff=1.e10
  
return, diff

end

;----------------------------------------------------------------

pro teltool_fitpeak_column, wave, flux, order, w_fit, region, contin, nfit, best_atm=best_atm_in

; TELTOOL_FITPEAK_COLUMN.PRO: Determine column density scaling factor
;        by fitting the peak depths of observed atmospheric lines with
;        models. It is called from teltool.pro
;
; Optional output
;
;    best_atm=..  FLTARR(4,N)  best-fitting atmospheric line with
;                     column 1: wavenumber
;                            2: normalized flux
;                            3: flux
;                            4: continuum  
;
;
; v0.1 04/29/2018 ACAB: added best_atm output
;

common teltool_commonblock, atmstruct, indexmol, atmod

  ;initial model with scaling factor 1
  atmstruct.scale_molecules[indexmol]=1.0
  ;initialize ratio of observed and model peak optical depths
  rat=1000.
  WHILE rat LT 0.995 OR rat GT 1.005 DO BEGIN
    IF atmstruct.server EQ 'psg' THEN atmod=get_psg(atmstruct)
    IF atmstruct.server EQ 'lblrtm' THEN atmod=get_lblrtm(atmstruct)

    ;divide over same local baseline as was done for the data
    ;because there can be overlying absorption by strong line wings
    out=teltool_normlineflux(atmstruct, atmod[0,*], atmod[7,*], indgen(n_elements(atmod[0,*]))*0, indgen(n_elements(atmod[0,*]))*0+1, w_fit, region, contin, nfit)
    transm=out[1,*]
    ;cleanplot
    ;window, 5
    ;plot, out[0,*], out[3,*], yr=[0,1.1]
    ;oplot, out[0,*], out[5,*], color=80
    ;oplot, out[0,*], out[1,*]
    ;stop
    ;ratio of peak optical depths
    index=where(FINITE(atmstruct.fit_flux) EQ 1)
    rat=MAX(-ALOG(atmstruct.fit_flux[index]))/MAX(-ALOG(transm))
    atmstruct.scale_molecules[indexmol]=atmstruct.scale_molecules[indexmol]*rat
    IF atmstruct.debug EQ 'yes' THEN print, 'scaling factor: '+ STRTRIM(atmstruct.scale_molecules[indexmol],2)+ ' tau_obs/tau_model: '+STRTRIM(rat,2)
  ENDWHILE

  best_atm_in=out

end

    
;----------------------------------------------------------------

pro teltool_fitprofile_column, wave, flux, error, order, w_fit, region, contin, nfit, best_atm=best_atm_in

; TELTOOL_FITPROFILE_COLUMN.PRO: Determine column density scaling
;        factor by fitting the profiles of observed atmospheric lines
;        with models. It is called from teltool.pro.
;
; Algorithm: determine ratio of observed peak optical depth over model
;    optical depth, roughly, that's the inital scaling
;    factor. Download 5 model spectra with that scaling factor times
;    0.50, 0.75, 1.00, 1.25, 1.50. Use amoeba to optimize using a
;    function that determines scaling factor and wavelength shift.
;
;
; Optional output
;
;    best_atm=..  FLTARR(4,N)  best-fitting atmospheric line with
;                     column 1: wavenumber
;                            2: normalized flux
;                            3: flux
;                            4: continuum  
;
;
; v0.1 11/03/2018 ACAB: created
;

common teltool_commonblock, atmstruct, indexmol, atmod
common teltool_atmfitblock, mgrid, sfac

  ;initial model with scaling factor 1
  atmstruct.scale_molecules[indexmol]=1.0
  
  ;get initial model to get rough idea of scaling factor this needs to
  ;be done in 2 steps because line depth and column not always linear
  FOR i=0,1 DO BEGIN
    IF atmstruct.server EQ 'psg' THEN atmod=get_psg(atmstruct)
    IF atmstruct.server EQ 'lblrtm' THEN atmod=get_lblrtm(atmstruct)

    ;divide over same local baseline as was done for the data
    ;because there can be overlying absorption by strong line wings
    out=teltool_normlineflux(atmstruct, atmod[0,*], atmod[7,*], indgen(n_elements(atmod[0,*]))*0, indgen(n_elements(atmod[0,*]))*0+1, w_fit, region, contin, nfit)
    transm=out[1,*]
    ;determine optical depth scaling factor taking into account that
    ;data can be noise. This is just to get within ~20% of best-fit.
    index=where(FINITE(atmstruct.fit_flux) EQ 1)
    rat=MAX(-ALOG(smooth(atmstruct.fit_flux[index],5,/nan)))/MAX(-ALOG(transm))
    atmstruct.scale_molecules[indexmol]=atmstruct.scale_molecules[indexmol]*rat
  ENDFOR
  IF atmstruct.debug EQ 'yes' THEN print, 'initial scaling factor: '+ STRTRIM(atmstruct.scale_molecules[indexmol],2)+ ' tau_obs/tau_model: '+STRTRIM(atmstruct.scale_molecules[indexmol],2)
    
  ;retrieve X models around this initial scaling factor
  sfac=[0.80,0.90,1.00,1.10,1.20]
  scale_orig=atmstruct.scale_molecules[indexmol]
  ;grid of observations and models
  mgrid=FLTARR(n_elements(sfac)+3,n_elements(atmstruct.fit_wavenumber))
  mgrid[0,*]=atmstruct.fit_wavenumber
  mgrid[1,*]=atmstruct.fit_flux
  mgrid[2,*]=atmstruct.fit_error
  FOR i=0,n_elements(sfac)-1 DO BEGIN
    atmstruct.scale_molecules[indexmol]=scale_orig*sfac[i]
    ;print, atmstruct.scale_molecules[indexmol]
    ;get initial model to get rough idea of scaling factor
    IF atmstruct.server EQ 'psg' THEN atmod=get_psg(atmstruct)
    IF atmstruct.server EQ 'lblrtm' THEN atmod=get_lblrtm(atmstruct)
    ;divide over same local baseline as was done for the data
    ;because there can be overlying absorption by strong line wings
    out=teltool_normlineflux(atmstruct, atmod[0,*], atmod[7,*], indgen(n_elements(atmod[0,*]))*0, indgen(n_elements(atmod[0,*]))*0+1, w_fit, region, contin, nfit)
    t=interpol(out[1,*],out[0,*],mgrid[0,*])
    mgrid[i+3,*]=t
  ENDFOR

  ;cleanplot
  ;window, 1
  ;plot, mgrid[0,*], mgrid[1,*]
  ;oplot, mgrid[0,*], mgrid[2,*], color=2
  ;oplot, mgrid[0,*], mgrid[3,*], color=3
  ;oplot, mgrid[0,*], mgrid[4,*], color=4
  ;oplot, mgrid[0,*], mgrid[5,*], color=5
  ;oplot, mgrid[0,*], mgrid[6,*], color=6
  ;stop
  
  ;fitting
  accur=0.001
  fitparams=amoeba(accur, p0=[1.,0.0],scale=[max(sfac)-1., (median(mgrid[0,*])/float(atmstruct.resolving_power))/5.], function_name='fit_atmgrid')
  atmstruct.scale_molecules[indexmol]=scale_orig*fitparams[0]
  ;atmstruct.waveshift=fitparams[0]
  
  ;get model spectrum with best_best scaling factor and wavelength shift
  IF atmstruct.server EQ 'psg' THEN atmod=get_psg(atmstruct)
  IF atmstruct.server EQ 'lblrtm' THEN atmod=get_lblrtm(atmstruct)
  ;divide over same local baseline as was done for the data
  ;because there can be overlying absorption by strong line wings
  out=teltool_normlineflux(atmstruct, atmod[0,*], atmod[7,*], indgen(n_elements(atmod[0,*]))*0, indgen(n_elements(atmod[0,*]))*0+1, w_fit, region, contin, nfit)

  best_atm_in=out

end

;----------------------------------------------------------------

function fit_atmgrid, p

; FIT_ATMGRID.PRO: function to be used in amoeba chi-square
;     minimization of atmospheric transmission line, called from
;     teltool_fitprofile_column.pro
;

common teltool_atmfitblock, mgrid, sfac

;sanity check
IF p[0] LT 0. THEN BEGIN
   ;02/18/2020..bug, should be r_chisq
   ;chisq=1.e5
   r_chisq=1.e5    
ENDIF ELSE BEGIN   
  ;observed wavelength
  wo=mgrid[0,*]
  ;observed transmission
  to=mgrid[1,*]
  ;observed uncertainty
  eo=mgrid[2,*]
  ;interpolate to atm transmission at p[0]
  t=to*0.
  FOR i=0,n_elements(wo)-1 DO BEGIN
    ;linear interpolation 
    t[i]=interpol(mgrid[3:*,i],sfac,p[0])
    ;question is whether quadratic interpolation is better
    ;t[i]=interpol(mgrid[3:*,i],sfac,p[0],/quadratic)
    ;t[i]=interpol(mgrid[3:*,i],sfac,p[0],/lsquadratic)
    ;cleanplot
    ;window, 1
    ;plot, sfac, mgrid[3:*,i],ys=1,ps=-5
    ;stop
  ENDFOR
  ;wavelength shift
  w=mgrid[0,*]+p[1]
  
  ;compare to observations via reduced chi-square
  r_chisq=TOTAL((to-interpol(t,w,wo))^2/(eo)^2)/(n_elements(wo)-2)

  ;;print, p
  ;plot, wo, to, title=r_chisq
  ;oplot, w, t, color=2
  ;stop

ENDELSE

;do not iterate further if reduced chi-square less than 1.01
IF r_chisq LT 1.01 THEN r_chisq=1.01

return, r_chisq

end

;----------------------------------------------------------------
                       
function fit_atmod_wave_func, p

; FIT_BGSTAR_WAVE_FUNC.PRO: function to be used in amoeba minimization
;     of residual as a function of wavelength shift between observed
;     data and model. It is called from teltool.pro
;

common teltool_commonblock, atmstruct, indexmol, atmod

 ;sanity check
IF ABS(p[0]) GT 1. THEN p[0]=0.0

;ratio of observation over model should be minimized
obs=atmstruct.fit_flux
model=interpol(atmod[7,*], atmod[0,*], atmstruct.fit_wavenumber+p[0])
diff=TOTAL(ABS(obs/model-1.))
;print, p[0], diff
;plot, obs/model-1, title=STRING(p[0])+' '+STRING(diff), yr=[-0.1,0.1]
;stopcont
 
return, diff

end

;----------------------------------------------------------------

function teltool_readparamfile, paramfile
  
;TELTOOL_READPARAMFILE.PRO: IDL program to read parameter file (ASCII)
;                    containing input parameters for teltool.pro
;
;Calling: struct=teltool_readparamfile(paramfile STRING name of paramete file
;                              )   
;Output: IDL structure with all parameters in paramfile
;
;v0.0 03/07/2018 A. Boogert: created for IRTF/iSHELL.
;

  ;read paramfile, ignoring lines starting with '#', separating the
  ;parameter and its value on each side of '='

  str=''
  openr,1, paramfile
  i=0
  while not eof(1) DO BEGIN
     readf, 1, str
     ind_eq=STRPOS(str,'=')
     IF (STRMID(str,0,1) NE '#') AND (ind_eq NE -1) THEN BEGIN
        tag=STRTRIM(STRMID(str,0,ind_eq),2)
        value=STRCOMPRESS(STRTRIM(STRMID(str,ind_eq+1,STRLEN(str)-1),2),/rem)
        ;convert array that is single string to array of numbers
        IF STRMID(value,0,1) EQ '[' THEN value=[strsplit(STRMID(value,1,strlen(value)-2),',',/extract)]
        ;create structure of one tag and one field
        atmstruct1= create_struct(tag, value )
        ;combine tags in one structure
        IF i NE 0 THEN BEGIN
           atmstruct2=create_struct(atmstruct, atmstruct1)
           atmstruct=atmstruct2
        ENDIF ELSE atmstruct=atmstruct1
        i=i+1   
     ENDIF
  endwhile
  close, 1

  ;reserve parameters for later use that are not in the parameterfile
  atmstruct1={ $
               ;wavenumber range to get model spectra for (changes
               ;during code depending on what's fitted and what not)
               wavenumber : [0.,0], $
               ;wavenumber range the user requests to correct
               ;atmospheric absorption for (entire input spectrum or
               ;as defined in xr input parameter)
               wavenumber_corr : [0.,0], $                          
               ;number of apertures (usually, spectra of each nod
               ;position) in the data file
               naps : 'header' $
             }
  atmstruct2=create_struct(atmstruct, atmstruct1)
  atmstruct=atmstruct2

  ;add the name of the parameter file itself to atmstruct
  atmstruct1={ paramfile : paramfile }
  atmstruct2=create_struct(atmstruct, atmstruct1)
  atmstruct=atmstruct2
  
  return, atmstruct
  
end

;----------------------------------------------------------------

function teltool_readlinefile, linefile, atmstruct
  
;TELTOOL_READLINEFILE.PRO: IDL program to read file (ASCII) with
;                            atmospheric lines to use for column
;                            density and wavelength shift
;                            determination in teltool.pro
;
;Calling: struct=teltool_readlinefile(
;
;                  line file and directory STRING
;
;                  atmstruct STRUCTURE with other parameters  
;                                      )   
;
;Output: IDL structure with line parameters from linefile added to
;        input atmtruct
;
;v0.0 04/12/2018 A. Boogert: created for IRTF/iSHELL
;

  ;read paramfile, ignoring lines starting with '#', separating the
  ;parameter and its value on each side of '='

  str=''
  openr,1, linefile
  i=0
  while not eof(1) DO BEGIN
     readf, 1, str
     IF (STRMID(str,0,1) NE '#') THEN BEGIN
        row=[strsplit(str,' ',/extract)]
        IF i EQ 0 THEN BEGIN
           instrument=row[0]
           config=row[1]
           molecule=row[2]
           waven=float(row[3])
           reg=float(row[4])
           cont=float(row[5])
        ENDIF ELSE BEGIN
           instrument=[instrument,row[0]]
           config=[config,row[1]]
           molecule=[molecule,row[2]]
           waven=[waven,float(row[3])]
           reg=[reg,float(row[4])]
           cont=[cont,float(row[5])]
        ENDELSE
        i=i+1
     ENDIF
  endwhile
  close, 1

  ;initialize
  w_h2o=-1.
  r_h2o=-1.
  c_h2o=-1.
  w_co2=-1.
  r_co2=-1.
  c_co2=-1.
  w_ch4=-1.
  r_ch4=-1.
  c_ch4=-1.
  w_co=-1.
  r_co=-1.
  c_co=-1.
  w_n2o=-1.
  r_n2o=-1.
  c_n2o=-1.
  w_o3=-1.
  r_o3=-1.
  c_o3=-1.
  w_ocs=-1.
  r_ocs=-1.
  c_ocs=-1.
  ;assign values if found for this instrument and configuration
  ind=where(strlowcase(instrument) EQ strlowcase(atmstruct.instrument) AND $
            strlowcase(config) EQ strlowcase(atmstruct.configuration), count)
  IF count GE 0 THEN BEGIN
     indm=where(molecule[ind] EQ 'H2O', countm)
     IF countm GT 0 THEN BEGIN
        w_h2o=waven[ind[indm]]
        r_h2o=reg[ind[indm]]
        c_h2o=cont[ind[indm]]        
     ENDIF
     indm=where(molecule[ind] EQ 'CO2', countm)
     IF countm GT 0 THEN BEGIN
        w_co2=waven[ind[indm]]
        r_co2=reg[ind[indm]]
        c_co2=cont[ind[indm]]        
     ENDIF
     indm=where(molecule[ind] EQ 'CH4', countm)
     IF countm GT 0 THEN BEGIN
        w_ch4=waven[ind[indm]]
        r_ch4=reg[ind[indm]]
        c_ch4=cont[ind[indm]]        
     ENDIF
     indm=where(molecule[ind] EQ 'CO', countm)
     IF countm GT 0 THEN BEGIN
        w_co=waven[ind[indm]]
        r_co=reg[ind[indm]]
        c_co=cont[ind[indm]]        
     ENDIF
     indm=where(molecule[ind] EQ 'N2O', countm)
     IF countm GT 0 THEN BEGIN
        w_n2o=waven[ind[indm]]
        r_n2o=reg[ind[indm]]
        c_n2o=cont[ind[indm]]        
     ENDIF
     indm=where(molecule[ind] EQ 'O3', countm)
     IF countm GT 0 THEN BEGIN
        w_o3=waven[ind[indm]]
        r_o3=reg[ind[indm]]
        c_o3=cont[ind[indm]]        
     ENDIF
     indm=where(molecule[ind] EQ 'OCS', countm)
     IF countm GT 0 THEN BEGIN
        w_ocs=waven[ind[indm]]
        r_ocs=reg[ind[indm]]
        c_ocs=cont[ind[indm]]        
     ENDIF
  ENDIF
  
  ;create structure of one tag and one field
  atmstruct1={$
              w_h2o: w_h2o, $
              r_h2o: r_h2o, $
              c_h2o: c_h2o, $
              w_co2: w_co2, $
              r_co2: r_co2, $
              c_co2: c_co2, $
              w_ch4: w_ch4, $
              r_ch4: r_ch4, $
              c_ch4: c_ch4, $
              w_co:  w_co, $
              r_co:  r_co, $
              c_co:  c_co, $
              w_n2o: w_n2o, $
              r_n2o: r_n2o, $
              c_n2o: c_n2o, $
              w_o3:  w_o3, $
              r_o3:  r_o3, $
              c_o3:  c_o3, $
              w_ocs:  w_ocs, $
              r_ocs:  r_ocs, $
              c_ocs:  c_ocs}
  
  ;reserve parameters for later use that are not in the parameterfile
  atmstruct2=create_struct(atmstruct, atmstruct1)
  atmstruct=atmstruct2
  
  return, atmstruct
  
end

;----------------------------------------------------------------

function teltool_readdata, atmstruct, header=header, readdata=readdata

;TELTOOL_READDATA.PRO: read data and header into arrays
;
;Calling: out=teltool_readdata(atmstruct, IDL structure with file and
;                                           directory information
;
;                                header=.., STRARR output string array with
;                                           header variables
;
;                                readdata=.. DBLARR return the data
;                                            array exactly as read
;                                            with readfits [DEFAULT:
;                                            do not return original
;                                            data]
;
;                               )
;
; Output: DBLARR(n_spectralpoints, 5) array with spectra having these
;                                     columns, sorted by wavenumber:
;
;                                     wave=out[*,0]  in cm^-1
;                                     flux=out[*,1]
;                                     uncertainty=out[*,2]
;                                     flag=out[*,3]
;                                     order=out[*,4]    
;
; v0.0 03/09/2018 ACAB: created
; v0.1 05/23/2018 ACAB: take into account multiple extraction
;                       apertures in one file
;

;for proper reading, need to know the instrument first
print, '***reading data file: ', atmstruct.obsdir+atmstruct.obsfile
readdata=readfits(atmstruct.obsdir+atmstruct.obsfile, header,/silent)
IF atmstruct.instrument EQ 'header' THEN atmstruct.instrument=STRTRIM(sxpar(header, "INSTR"), 2)
IF atmstruct.instrument EQ 'iSHELL' THEN BEGIN
 n_orders=sxpar(header,'NORDERS')
 naps=sxpar(header,'NAPS')
 alldata=DBLARR(n_elements(readdata[*,0,0])*n_orders, 5)
 firstorder=STRMID(sxpar(header,'ORDERS'),0,(STRSPLIT((sxpar(header,'ORDERS')),','))[1]-1)
 ;TBD I think that the order number increases with wavelength, but
 ;check against manual
 orders=INDGEN(n_orders)+firstorder
 n_wave=n_elements(readdata[*,0,0])
 ;get data for each order, but only for one aperture (the first one);
 ;assuming for now the difference in atmospheric transmission between
 ;the nod positions is negligible
 get_ap=naps
 ;now loop over the orders
 FOR i=0,n_orders-1 DO BEGIN  
    alldata[i*n_wave:(i+1)*n_wave-1,0]=1.e4/readdata[*,0,i*get_ap]
    alldata[i*n_wave:(i+1)*n_wave-1,1]=readdata[*,1,i*get_ap]
    alldata[i*n_wave:(i+1)*n_wave-1,2]=readdata[*,2,i*get_ap]
    alldata[i*n_wave:(i+1)*n_wave-1,3]=readdata[*,3,i*get_ap]
    alldata[i*n_wave:(i+1)*n_wave-1,4]=orders[i]
 ENDFOR
 index=sort(alldata[*,0])
 alldata[*,0]=alldata[index,0]
 alldata[*,1]=alldata[index,1]
 alldata[*,2]=alldata[index,2]
 alldata[*,3]=alldata[index,3]
 alldata[*,4]=alldata[index,4] 

 ;remove wavelength 0 and NaN
 
 index=where(alldata[*,0] NE 0. AND finite(alldata[*,0]) EQ 1)
 alldata=alldata[index,*]
 ;print, alldata[*,0]
ENDIF ELSE print, '***data read for instrument '+atmstruct.instrument+' not defined'
 
return, alldata

end

;----------------------------------------------------------------

function teltool_header, atmstruct, header

;TELTOOL_HEADER.PRO: update atmstruct with header information
;                      obtained from fits file
;
;Calling: out=teltool_header(atmstruct, IDL structure to be updated
;
;                                header=.., STRARR string array with
;                                           header variables
;                               )
;
; Output: IDL structure updated with header information
;
; v0.0 03/09/2018 ACAB: created
; v0.1 21/05/2020 A. Boogert: make compatible with non-averaged
;                            spectra (that don't have AVE_DATE
;                            and AVE_TIME keywords)
;

  tags=tag_names(atmstruct)
  IF atmstruct.observatory EQ 'header' THEN BEGIN
     atmstruct.observatory=STRTRIM(sxpar(header, "TELESCOP"),2) ;for files produced by spextool
     IF !ERR EQ -1 THEN BEGIN
        IF atmstruct.instrument EQ 'iSHELL' THEN atmstruct.observatory="NASA IRTF" ;for files produced by xcombspec     
     ENDIF
  ENDIF
  IF atmstruct.configuration EQ 'header' THEN atmstruct.configuration=STRTRIM(sxpar(header, "MODE"), 2)
  FOR i=0,n_elements(tags)-1 DO BEGIN
     headervalue='noheadervaluefound'
     tagvalue=atmstruct.(i)
     type=size(tagvalue, /type)
     dimen=size(tagvalue, /dimen)
     IF type EQ 7 AND dimen LT 2 THEN BEGIN ;make sure that next if statement does not crash
       IF  tagvalue EQ 'header' THEN BEGIN
          IF atmstruct.instrument EQ 'iSHELL' THEN BEGIN
            ;IF tags[i] EQ 'ZA' THEN headervalue=90.-sxpar(header, "TCS_EL")
            IF tags[i] EQ 'AIRMASS' THEN BEGIN
               headervalue=sxpar(header, "AVE_AM") ;for files produced by xcombspec
               IF !ERR EQ -1 THEN headervalue=sxpar(header, "AM") ;for files produced by spextool
            ENDIF
            ;utc should be like this 2016-09-16T14:30:00.00 
            IF tags[i] EQ 'UTC' THEN BEGIN
               headervalue=sxpar(header, "AVE_TIME") ;test if files produced by xcombspec
               IF !ERR EQ -1 THEN BEGIN
                  ;for files produced by spextool                  
                  headervalue=sxpar(header, "DATE")+'T'+sxpar(header, "TIME")                  
               ENDIF ELSE BEGIN
                  ;for files produced by xcombspec
                  headervalue=sxpar(header, "AVE_DATE")+'T'+sxpar(header, "AVE_TIME")
               ENDELSE
            ENDIF
            IF tags[i] EQ 'SLIT_WIDTH' THEN headervalue=sxpar(header, "SLTW_ARC")
            IF tags[i] EQ 'RESOLVING_POWER' THEN BEGIN
               IF atmstruct.slit_width NE 'header' THEN sw=atmstruct.slit_width  ELSE sw=sxpar(header, "SLTW_ARC")
               IF float(sw) EQ 0.375 THEN headervalue=80000.
               IF float(sw) EQ 0.75 THEN headervalue=50000.
               IF float(sw) EQ 1.50 THEN headervalue=25000. ; TBD
            ENDIF
            IF tags[i] EQ 'NAPS' THEN headervalue=sxpar(header, "NAPS") ;number of apertures
            atmstruct.(i)=headervalue
          ENDIF
       ENDIF
    ENDIF
  ENDFOR

return, atmstruct

end

;----------------------------------------------------------------

pro atmod_get_full_spectrum, nlayers=nlayers

;ATMOD_GET_FULL_SPECTRUM.PRO: get full spectrum from server with all
;    the parameters the user entered in the parameter file or
;    teltool command line call. This separate routine is needed
;    because the servers have limited the wavelength range over which
;    one spectrum can be calculated to about 400 cm-1 (for a
;    resolution of 500,000...it's more at lower resolution for the
;    same number of points, but we do the convolution with the
;    instrument profile locally). So we have to calculate the separate
;    regions separately and then merge into the atmod array.
;
;Calling: atmod_get_full_spectrum,
;
;              /nlayers KEYWD just update number of atmospheric layers
;                             (which is linked to the number of lines
;                             to download) in atmstruct.nlayers. If
;                             this keyword is set no spectrum is
;                             calculated, just atmstruct.nlayers is
;                             updated.
; 
;
; v0.0 A. Boogert 03/14/18: created  
;
  
common teltool_commonblock, atmstruct, indexmol, atmod

;retrieve in chunks, because PSG can only handle a certain number of
;lines times number of atmospheric layers
;
;Allow for the possibility of small and large chunks because some
;wavelength regions contain many more lines (e.g. for ozone). For the
;small chunks, give the wavenumber ranges for which they apply.
;
chunk_large=200.
chunk_small=200.
small_chunk_wave=[-1.,-1.]

;unless defined in the parameter file, the number of layers will be
;updated here depending on the number of transitions to be used in the
;calculations.
;not working yet
;IF atmstruct.nlayers EQ -1 THEN BEGIN
;   update_nlayers=1
;   atmstruct.nlayers=72
;ENDIF
   
IF strlowcase(atmstruct.instrument) EQ 'ishell' THEN BEGIN
   ;iSHELL K3   
   IF strlowcase(atmstruct.configuration) EQ 'k3' THEN BEGIN
      chunk_large=200.           
      chunk_small=200.
      small_chunk_wave=[-1.,-1.]
   ENDIF
   ;iSHELL M-band
   IF strlowcase(atmstruct.configuration) EQ 'm1' OR $
      strlowcase(atmstruct.configuration) EQ 'm2' THEN BEGIN
      chunk_large=200.
      ;2070-2091: O3 still too many lines at 0.5 cm-1 chunks
      chunk_small=5.
      small_chunk_wave=[2040,2130]
      ;reduce number of layers because even 5 cm-1 chunk size not
      ;not working yet
      ;IF keyword_set(update_nlayers) THEN atmstruct.nlayers=36
   ENDIF
ENDIF

;something that works best in case very large wavenumber ranges need
;to be downloaded
IF atmstruct.orphan EQ 1 THEN chunk_large=50

IF NOT keyword_set(nlayers) THEN BEGIN

  ;print scaling factors used
  prmols=' '   
  prscalemols=' '
  FOR i=0,n_elements(atmstruct.molecules)-1 DO BEGIN
    prmols=prmols+STRTRIM(atmstruct.molecules[i],2)+' '
    prscalemols=prscalemols+STRTRIM(atmstruct.scale_molecules[i],2)+' '
  ENDFOR
  
  print, '***calculating entire model spectrum ('+STRTRIM(atmstruct.wavenumber[0],2)+'-'+STRTRIM(atmstruct.wavenumber[1],2)+' cm-1)'
  print, '   including only the molecules '+prmols
  print, '   which have column density scaling factors of '+prscalemols

  ;preserve full requested wavenumber range
  origwavenumber=atmstruct.wavenumber
  
  ;sanity check.. small_chunk_wave has to be within origwavenumber
  IF small_chunk_wave[0] NE -1 THEN BEGIN
    ind=where(small_chunk_wave GT origwavenumber[0] AND small_chunk_wave LT origwavenumber[1], count)
    IF count EQ 0 THEN small_chunk_wave=[-1.,-1.]
    IF count EQ 1 THEN BEGIN
       IF ind[0] EQ 0 THEN small_chunk_wave[1]=origwavenumber[1]
       IF ind[0] EQ 1 THEN small_chunk_wave[0]=origwavenumber[0]     
    ENDIF
  ENDIF
  
  IF small_chunk_wave[0] EQ -1 THEN BEGIN
    n_chunks=ceil((origwavenumber[1]-origwavenumber[0])/chunk_large)
    w_chunks=FINDGEN(n_chunks+1)*chunk_large+origwavenumber[0]
    w_chunks[n_chunks]=origwavenumber[1]
  ENDIF ELSE BEGIN
    ;large chunks
    n_chunks_large=ceil((origwavenumber[1]-origwavenumber[0])/chunk_large)
    w_chunks_large=FINDGEN(n_chunks_large+1)*chunk_large+origwavenumber[0]
    w_chunks_large[n_chunks_large]=origwavenumber[1]
    ;small chunks
    n_chunks_small=ceil((small_chunk_wave[1]-small_chunk_wave[0])/chunk_small)
    w_chunks_small=FINDGEN(n_chunks_small+1)*chunk_small+small_chunk_wave[0]
    w_chunks_small[n_chunks_small]=small_chunk_wave[1]
    ;combine small and large chunks
    ind=where(w_chunks_large LT MIN(w_chunks_small) OR $
              w_chunks_large GT MAX(w_chunks_small), count)
    IF count GT 0 THEN BEGIN
     w_chunks=[w_chunks_large[ind], w_chunks_small]
     ind=sort(w_chunks)
     w_chunks=w_chunks[ind]
    ENDIF ELSE w_chunks=w_chunks_small
  ENDELSE
  
  ;extend two resolution elements on both sides because convolution
  ;with instrumental resolution is done on each chunk
  resolution_element=median(atmstruct.wavenumber)/float(atmstruct.resolving_power)
  delta=2.*resolution_element
   
  ;cleanplot
  ;plot, [0,1], [0,1],xr=origwavenumber, yr=[0,1.1],/nodata
  ;color=[20,80,20,80,20,80]
  FOR n=0, n_elements(w_chunks)-2 DO BEGIN
    atmstruct.wavenumber=[w_chunks[n]-delta, w_chunks[n+1]+delta]
    ;wavenumber range needs to be at least 6 resolution elements for
    ;proper convolution with instrument profile (at least that's what's
    ;used in teltool_conv_instr.pro
    IF atmstruct.wavenumber[1]-atmstruct.wavenumber[0] LE 6.*resolution_element THEN BEGIN
      atmstruct.wavenumber[1]=atmstruct.wavenumber[0]+7.*resolution_element
    ENDIF
    IF n EQ 0 THEN pre='   retrieving chunk ' ELSE pre='                    '
    percen=STRTRIM((atmstruct.wavenumber[1]-atmstruct.wavenumber[0]-2.*delta)/(origwavenumber[1]-origwavenumber[0])*100.,2)
    print, pre+STRTRIM(atmstruct.wavenumber[0]+delta,2)+'-'+STRTRIM(atmstruct.wavenumber[1]-delta,2)+ ' cm-1 ('+percen+'% of total)'

    IF atmstruct.server EQ 'psg' THEN atmod1=get_psg(atmstruct)
    IF atmstruct.server EQ 'lblrtm' THEN atmod1=get_lblrtm(atmstruct)
    
    ;remove edges as they are compromised by convolution with
    ;instrumental profile
  
    ind=where(atmod1[0,*] GT atmstruct.wavenumber[0]+delta AND atmod1[0,*] LT atmstruct.wavenumber[1]-delta )
    atmod1=atmod1[*,ind]
    
    ;oplot, atmod1[0,*], atmod1[7,*], color=color[n]
    
    ;if the requested wavenumber space is covered within 0.5 cm-1, we're
    ;happy. Do this recursively until entire wavelength range is covered.
    ;WHILE min(atmod[0,*]) GT atmstruct.wavenumber[0]+0.5 DO BEGIN
    IF n GT 0 THEN BEGIN
      ;ind=where(atmod1[0,*] LT min(atmod[0,*]))
       atmod=[[atmod1], [atmod]]
    ENDIF ELSE atmod=atmod1
  ENDFOR
  atmstruct.wavenumber=origwavenumber
  
  ;stop
  
  ;sort in wavenumber space
  ind=sort(atmod[0,*])
  FOR i=0,(size(atmod,/dimen))[0]-1 DO BEGIN
    atmod[i,*]=atmod[i,ind]
  ENDFOR

ENDIF

;not working yet
;print, "***number of atmospheric layers: ", atmstruct.nlayers
  
end
   
;----------------------------------------------------------------   

pro teltool, paramfile, xr=xr, atm=readdata_atm, div=readdata_div, $
               do_plot=do_plot_in, order=order_in, molecules=molecules_in, $
               scale_molecules=scale_molecules_in, $
               fit_molecules=fit_molecules_in, stemfile=stemfile_in, $
               onlyfit=onlyfit_in, obsfile=obsfile_in, obsdir=obsdir_in, $
               orphan=orphan_in, plotfile=plotfile_in, teltool_vars
  
;TELTOOL.PRO: IDL program to correct spectra for atmospheric
;               absorption
;
;General workings:
;
;  -determine ...
;
;  -get atmospheric model spectra from a server
;
;Calling: teltool, paramfile STRING name of file with input
;                                     parameters
;  
;                  , teltool_vars undeclared variable that on output
;                                 will contain an IDL structure with
;                                 most variables that teltool uses.
;                                 [OPTIONAL]
;
;                  , xr=[..,..] FLTARR(2) wavenumber [cm^-1] range to
;                                    correct the atmosphere for
;                                    [DEFAULT: full spectrum or single
;                                    orders, depending on atmstruct.orders]
;
;                  , atm=DBLARR() output array with atmospheric model
;                                 in iSHELL format, only if all orders
;                                 processed separately, i.e.,
;                                 atmstruct.order=-1  [DEFAULT: no output]
;
;                  , div=DBLARR() output array with
;                                 observation/atmospheric model division
;                                 in iSHELL format, only if all orders
;                                 processed separately, i.e.,
;                                 atmstruct.order=-1 [DEFAULT: no output]
;
;                  , do_plot=..   Overrules 'do_plot' in input
;                                 parameter file:
;                                 0: no plots,
;                                 1: plot each fitted line
;                                 3: only plot best-fit spectrum,
;                                 4: plot each molecule separately
;                                    plus best fit spectrum
;                                 [DEFAULT: read do_plot from
;                                  parameter file]
;
;                  , plotfile=..  Overrules 'plotfile' in input
;                                 parameter file. 
;                                 [DEFAULT: plot on screen, not in file]
;
;                  , stemfile=..  Overrules 'stemfile' in input
;                                 parameter file. 
;                                 [DEFAULT: 'pumas']
;
;                  , onlyfit=..   Overrules 'onlyfit' in input
;                                 parameter file. 
;                                 [DEFAULT: read 'onlyfit' from
;                                  parameter file]
;
;                  , order=..     Overrules 'order' in input parameter
;                                 file:
;                                 -2: do not treat orders
;                                     individually, process entire
;                                     spectrum at once or specific
;                                     ranges as set by teltool 'xr'
;                                     input
;                                 -1: all orders, order by order
;                                     positive integers: specific orders
;                                 [DEFAULT: read 'order' from
;                                 parameter file]
;
;                  , molecules=.. Overrules 'molecules' in input
;                                 parameter file, e.g. ['H2O', 'CH4',
;                                 'CO2']
;                                 [DEFAULT: read
;                                 'molecules' from parameter file]
;
;                  , scale_molecules=.. Overrules 'scale_molecules' in
;                                 input parameter file. Same order and
;                                 number of elements as 'molecules'.
;                                 [DEFAULT:
;                                 read 'scale_molecules' from
;                                 parameter file]
;
;                  , fit_molecules=.. Overrules 'fit_molecules' in
;                                 input parameter file. Same order and
;                                 number of elements as 'molecules',
;                                 e.g., ['yes', 'no', 'yes']
;                                 [DEFAULT:
;                                 read 'fit_molecules' from parameter
;                                 file]
;
;                  , obsfile=..   Overrules 'obsfile' in input
;                                 parameter file.  [DEFAULT: read
;                                 'obsfile' from parameter file]
;
;                  , obsdir=..    Overrules 'obsdir' in input
;                                 parameter file.  [DEFAULT: read
;                                 'obsdir' from parameter file]
;
;                  , orphan=.. STRING meant to create an atmospheric
;                                  spectrum that is not precisely
;                                  related to the observed spectrum
;                                  nor meant to match it. This is most
;                                  importantly the wavenumber
;                                  range. It can be outside of the
;                                  observed wavenumber range. Other
;                                  parameters from the parameter file
;                                  and command line are still used
;                                  (resolution, observatory, etc.).
;                                  The value of orphan is the filename
;                                  with the atmospheric transmission
;                                  spectrum (per molecule and combined).
;  
;v0.0 01/31/2018 A. Boogert: created for IRTF/iSHELL
;v0.1 08/31/2018 A. Boogert: only fit to lines within the xr input if given
;v0.2 10/17/2018 A. Boogert: check for gaps in coverage (such as for iSHELL M-band)
;v0.3 11/30/2018 A. Boogert: only plot lines for those species for
;                            which column scaling factor or
;                            wavelengths were fitted
;v0.4 02/18/2020 A. Boogert: fixed bug chisq-->r_chisq in fit_atmgrid
;

common teltool_commonblock, atmstruct, indexmol, atmod

start_time=systime(1)

cleanplot, /silent

IF NOT keyword_set(xr) THEN xr=0

;Create structure with relevant information about observation and
;fitting parameters. It is easier to propagate a sructure than many
;individual parameters. Use variables with the same names as in
;TelTool parameter file (see lblrtm_param_ishell_k3.py for an
;example).

atmstruct=teltool_readparamfile(paramfile)

;update atmstruct with directory for auxiliary files
IF atmstruct.auxdir EQ 'auto' THEN BEGIN
   ;get directory where iSHELL parameter file is located.
   auxdir=teltool_directory()
   ;get directory where file 'teltool_directory' is located
   atmstruct.auxdir=auxdir
ENDIF

;rfile, atmstruct.obsdir+atmstruct.obsfile, rowcol=[100000, 5, 1],/indef, a,/quiet
;rfile, atmstruct.obsdir+atmstruct.obsfile, rowcol=[100000, 2, 0],/indef, a,/quiet

;read data and header into arrays, making sure that command line input
;has preference over parameter file
IF n_elements(obsfile_in) GT 0 THEN atmstruct.obsfile=obsfile_in
IF n_elements(obsdir_in) GT 0 THEN atmstruct.obsdir=obsdir_in
alldata=teltool_readdata(atmstruct, header=header, readdata=readdata)

;update atmstruct with header information
atmstruct=teltool_header(atmstruct, header)

;update atmstruct with weather information from IRTF weather archive
;atmstruct=teltool_irtfweather(atmstruct)

;update atmstruct with command line input
IF n_elements(do_plot_in) GT 0 THEN atmstruct.do_plot=do_plot_in
IF n_elements(plotfile_in) GT 0 THEN atmstruct.plotfile=plotfile_in
IF n_elements(onlyfit_in) GT 0 THEN atmstruct.onlyfit=onlyfit_in
IF n_elements(stemfile_in) GT 0 THEN atmstruct.stemfile=stemfile_in
IF n_elements(order_in) GT 0 THEN atmstruct.order=order_in
IF n_elements(molecules_in) GT 0 THEN BEGIN
   IF tag_exist(atmstruct, 'molecules') THEN BEGIN
      struct_delete_field, atmstruct, 'molecules'
   ENDIF
   atmstruct2={molecules : molecules_in}
   atmstruct1=atmstruct
   atmstruct=create_struct(atmstruct1, atmstruct2)
ENDIF
IF n_elements(scale_molecules_in) GT 0 THEN BEGIN
   IF tag_exist(atmstruct, 'scale_molecules') THEN BEGIN
      struct_delete_field, atmstruct, 'scale_molecules'
   ENDIF
   atmstruct2={scale_molecules : scale_molecules_in}
   atmstruct1=atmstruct
   atmstruct=create_struct(atmstruct1, atmstruct2)
ENDIF
IF n_elements(fit_molecules_in) GT 0 THEN BEGIN
   IF tag_exist(atmstruct, 'fit_molecules') THEN BEGIN
      struct_delete_field, atmstruct, 'fit_molecules'
   ENDIF
   atmstruct2={fit_molecules : fit_molecules_in}
   atmstruct1=atmstruct
   atmstruct=create_struct(atmstruct1, atmstruct2)
ENDIF
IF tag_exist(atmstruct, 'orphan') THEN BEGIN
    struct_delete_field, atmstruct, 'orphan'
ENDIF
atmstruct2={orphan : keyword_set(orphan_in)}
atmstruct1=atmstruct
atmstruct=create_struct(atmstruct1, atmstruct2)

;update atmstruct by reading file with PSG access key
IF atmstruct.psg_key_file EQ 'none' THEN psg_key='-1' ELSE BEGIN 
  IF file_test(atmstruct.psg_key_file,/read) eq 0 OR file_test(atmstruct.psg_key_file,/zero_length) EQ 1 THEN BEGIN
    psg_key='-1'
    print, '***PSG access key file ', atmstruct.psg_key_file, ' not found or empty'
    print, '   Accessing PSG without key'
 ENDIF ELSE BEGIN
    str=''
    openr, 1, atmstruct.psg_key_file
    readf,1,str
    close,1
    psg_key=str
 ENDELSE
ENDELSE
atmstruct2={psg_key : psg_key}
atmstruct1=atmstruct
atmstruct=create_struct(atmstruct1, atmstruct2)

;update atmstruct with number of atmospheric layers
;no spectrum will be calculated this time
;not working yet
;atmod_get_full_spectrum, /nlayers

;select the data to be processed (TBD: this will have to optimized to
;process the entire wavelength range efficiently)
;wave=1.e4/(a[0,sindex]/1.e4)         ;Angstrom to cm^-1
wave_in=alldata[*,0]
flux_in=alldata[*,1]
error_in=alldata[*,2]
flag_in=alldata[*,3]
order_in=alldata[*,4]

;define wavenumber range to correct atmosphere for
IF NOT keyword_set(xr) THEN $
      atmstruct.wavenumber_corr=[min(wave_in), max(wave_in)] ELSE atmstruct.wavenumber_corr=[min(xr),max(xr)]

;output atmospheric model array with flux column replaced by
;atmospheric model, and columns added with transmission for each
;molecule
IF atmstruct.order EQ -1 THEN BEGIN
 dimen_readdata=size(readdata,/dimen)
 IF atmstruct.server EQ 'psg' THEN nmols_from_code=12 ELSE nmols_from_code=11 ;irrespective of how many molecules are requested, the code returns this many
 readdata_atm=DBLARR(dimen_readdata[0], dimen_readdata[1]+nmols_from_code, dimen_readdata[2])
 readdata_atm[*,0,*]=readdata[*,0,*]
 readdata_atm[*,1,*]=0.
 readdata_atm[*,2,*]=0.
 readdata_atm[*,3,*]=0
ENDIF

;output division array with flux column replaced by flux divided by
;atmospheric model
IF atmstruct.order EQ -1 THEN BEGIN
 readdata_div=readdata
 readdata_div[*,1,*]=0.
 readdata_div[*,2,*]=0.
ENDIF

;do not take orders into account, select only on wavelength range
IF atmstruct.order EQ -2 THEN BEGIN
   begin_order=0
   end_order=0
ENDIF
;do all orders, order by order
IF atmstruct.order EQ -1 THEN BEGIN
   begin_order=min(order_in)
   end_order=max(order_in)
ENDIF
;do one individual order
IF atmstruct.order GT 0 THEN BEGIN
   begin_order=atmstruct.order*1 ;convert from string to integer
   end_order=atmstruct.order*1 ;convert from string to integer
ENDIF

;model optimization w.r.t. available weather parameters?
;IF atmstruct.fit_wave EQ 'yes' OR (where(atmstruct.fit_molecules EQ 'yes'))[0] NE -1 THEN dofit=1 ELSE dofit=0
IF (where(atmstruct.fit_molecules EQ 'yes'))[0] NE -1 THEN dofit=1 ELSE dofit=0

;correct for small wavelength calibration errors determined previously
IF atmstruct.fit_wave EQ 'no' OR NOT keyword_set(dofit) THEN wave_in=wave_in+atmstruct.waveshift

;If no fitting is needed, and all orders are corrected, then get model
;for entire observed wavelength range, instead of in small chunks for
;each order, avoiding many server calls.
IF NOT keyword_set(dofit) AND atmstruct.order LE 0 THEN BEGIN
      atmstruct.wavenumber=atmstruct.wavenumber_corr
      atmod_get_full_spectrum
ENDIF

;if fitting is needed, get the lines/regions to fit to
IF keyword_set(dofit) THEN $
    atmstruct=teltool_readlinefile(atmstruct.auxdir+'teltool_lines.dat', atmstruct)

;loop over orders
FOR o=begin_order, end_order DO BEGIN

   ;select order
   IF o EQ 0 THEN BEGIN
      wave=wave_in
      flux=flux_in
      error=error_in
      flag=flag_in
      order=order_in
      spectralordertitle=' '
   ENDIF ELSE BEGIN
      index=where(order_in EQ o, counto)
      IF counto GT 0 THEN BEGIN
         wave=wave_in[index]
         flux=flux_in[index]
         error=error_in[index]
         flag=flag_in[index]
         order=order_in[index]
         spectralordertitle='for Echelle Order '+STRTRIM(fix(o),2)

      ENDIF ELSE BEGIN
         print, '***order '+STRTRIM(o,2)+' not found'
         print, '   available orders', order_in[uniq(order_in, sort(order_in))]
         stop
      ENDELSE
   ENDELSE
   
   IF keyword_set(dofit) THEN BEGIN
       ;Molecules to consider for column and wavelength fitting.
       ;Will only happen if atmstruct.fit_molecules ot fit_wave are set.
       consider_fitting_molecules=['H2O', 'CH4', 'CO2', 'CO', 'N2O', 'O3', 'OCS']
   
       ;;use this to find suitable lines for column density correction factors
       ;atmstruct.wavenumber=[4390,4440]-400;[3937,4444]
       ;atmod=get_psg(atmstruct)
       ;stopcont

       ;save fitting plots in file
       ;atmstruct.plotfile='none'
       IF atmstruct.plotfile NE 'none' THEN BEGIN
          set_plot, 'ps'
          device, filename=atmstruct.plotfile, /color, bits_per_pixel=8,$
                  /landscape, scale_factor=1.0         
          print, atmstruct.plotfile+' will be written to disk'
       ENDIF    

       ;loop over molecules
       FOR i=0,n_elements(consider_fitting_molecules)-1 DO BEGIN
          molecule=consider_fitting_molecules[i]
          IF molecule EQ 'H2O' THEN BEGIN
             w_fit=atmstruct.w_h2o
             region=atmstruct.r_h2o ;wavelength region around line: w_fit+/-(region*resolution)
             contin=atmstruct.c_h2o ;start of continuum: w_fit+/-(contin*resolution)
          ENDIF
          IF molecule EQ 'CO2' THEN BEGIN
             w_fit=atmstruct.w_co2
             region=atmstruct.r_co2 ;wavelength region around line: w_fit+/-(region*resolution)
             contin=atmstruct.c_co2 ;start of continuum: w_fit+/-(contin*resolution)   
          ENDIF
          IF molecule EQ 'CH4' THEN BEGIN
             w_fit=atmstruct.w_ch4
             region=atmstruct.r_ch4 ;wavelength region around line: w_fit+/-(region*resolution)
             contin=atmstruct.c_ch4 ;start of continuum: w_fit+/-(contin*resolution)   
          ENDIF
          IF molecule EQ 'CO' THEN BEGIN
             w_fit=atmstruct.w_co
             region=atmstruct.r_co ;wavelength region around line: w_fit+/-(region*resolution)
             contin=atmstruct.c_co ;start of continuum: w_fit+/-(contin*resolution)   
          ENDIF
          IF molecule EQ 'N2O' THEN BEGIN
             w_fit=atmstruct.w_n2o
             region=atmstruct.r_n2o ;wavelength region around line: w_fit+/-(region*resolution)
             contin=atmstruct.c_n2o ;start of continuum: w_fit+/-(contin*resolution)   
          ENDIF
          IF molecule EQ 'O3' THEN BEGIN
             w_fit=atmstruct.w_o3
             region=atmstruct.r_o3 ;wavelength region around line: w_fit+/-(region*resolution)
             contin=atmstruct.c_o3 ;start of continuum: w_fit+/-(contin*resolution)   
          ENDIF
          IF molecule EQ 'OCS' THEN BEGIN
             w_fit=atmstruct.w_ocs
             region=atmstruct.r_ocs ;wavelength region around line: w_fit+/-(region*resolution)
             contin=atmstruct.c_ocs ;start of continuum: w_fit+/-(contin*resolution)   
          ENDIF
          ;column fitting for this molecule?
          indexmol=where(atmstruct.molecules EQ molecule AND atmstruct.fit_molecules EQ 'yes', do_fitmol)
          ;make sure that the lines to fit to are within the observed
          ;or requested wavelength range
          ;w_reg=[w_fit[nfit]-region[nfit]*w_fit[nfit]/float(atmstruct.resolving_power), w_fit[nfit]+region[nfit]*w_fit[nfit]/float(atmstruct.resolving_power)]
          index=where(w_fit GT min(atmstruct.wavenumber_corr) AND w_fit LT max(atmstruct.wavenumber_corr), count)
          IF count GT 0 THEN BEGIN
             w_fit=w_fit[index]
             region=region[index]
             contin=contin[index]
             ;and exclude lines that fall in a coverage gap, taking
             ;into account that some instrument settings have
             ;wavelength gaps (e.g., M-band iSHELL)
             zall=1000
             FOR z=0, count-1 DO BEGIN
                ind1=where(wave GT w_fit[z]-region[z]*w_fit[z]/float(atmstruct.resolving_power) AND $
                           wave LT w_fit[z]-contin[z]*w_fit[z]/float(atmstruct.resolving_power), count1)
                ind2=where(wave GT w_fit[z]+contin[z]*w_fit[z]/float(atmstruct.resolving_power) AND $
                           wave LT w_fit[z]+region[z]*w_fit[z]/float(atmstruct.resolving_power), count2)
                IF count1 GT 0 AND count2 GT 0 THEN zall=[zall,z]
             ENDFOR
             IF n_elements(zall) EQ 1 THEN w_fit=[-1] ELSE BEGIN
                ind=where(zall NE 1000)
                zall=zall[ind]
                w_fit=w_fit[zall]
                region=region[zall]
                contin=contin[zall]                
             ENDELSE
          ENDIF ELSE w_fit=[-1]
          ;start fitting loop
          IF w_fit[0] NE -1 THEN BEGIN
               ;molecule included in model?
               index=where(atmstruct.molecules EQ molecule, do_mol)             
               IF do_mol EQ 1 THEN BEGIN
                  ;loop over all lines to fit to
                  n_fitlines=n_elements(w_fit)
                  scale_molecules_all_lines=FLTARR(n_fitlines) ;collect scaling factors for all molecules, and then later use median et 'best fit'
                  FOR nfit=0,n_fitlines-1 DO BEGIN
                     ;optimize wavelength calibration using lines from this molecule?
                     IF atmstruct.fit_wave EQ 'yes' THEN do_fitwave=1 ELSE do_fitwave=0            

                     ;prepare sub-spectrum to be fitted to (select, divide over baseline)                     
                     normline=teltool_normlineflux(atmstruct, wave, flux, error, order, w_fit, region, contin, nfit, c_reg1=c_reg1, c_reg2=c_reg2, w_reg=w_reg, const=const_normline, rico=rico_normline)
                           
                     ;add sub-spectrum for this line to
                     ;structure, for column or wavelength fitting or just for plotting
                     IF tag_exist(atmstruct, 'fit_wavenumber') THEN BEGIN
                        struct_delete_field, atmstruct, 'fit_wavenumber'
                        struct_delete_field, atmstruct, 'fit_flux'
                        struct_delete_field, atmstruct, 'fit_error'
                     ENDIF
                     atmstruct2={fit_wavenumber : normline[0,*], $
                                 fit_flux :       normline[1,*], $
                                 fit_error :      normline[2,*]}
                     atmstruct1=atmstruct
                     atmstruct=create_struct(atmstruct1, atmstruct2)
   
                     IF do_fitmol EQ 1 OR do_fitwave EQ 1 THEN $
                        print, '***'+molecule+' ['+STRTRIM(w_fit[nfit],2)+ ' cm-1]'
                     
                     ;find best-fit column scaling factor (relative to
                     ;standard model)              
                     countmin=0
                     IF do_fitmol EQ 1 THEN BEGIN
                        ;first make sure other molecules to be fitted
                        ;have scaling factor 1.0, not -1.0. This
                        ;matters in case their lines overlap with the
                        ;line of the molecule to be fitted.                        
                        indexmin=where(atmstruct.scale_molecules EQ -1., countmin)
                        IF countmin GT 1 THEN BEGIN
                           indexmin=indexmin[where(indexmin NE indexmol)]
                           atmstruct.scale_molecules[indexmin]=1.                           
                        ENDIF
                        ;now fit the column
                        IF atmstruct.fit_method EQ 'local' THEN BEGIN
                          print, '   fitting column density scaling factor (locally) ... have patience'
                          ;IF atmstruct.debug EQ 'yes' THEN print, '      scaling', '   |min(obs)-min(model)|'                     
                          ;accur=0.01 ;for S/N=100, to be adjusted to data quality
                          ;fitparams=amoeba(accur, p0=[atmstruct.scale_molecules[indexmol]], $
                          ;                        scale=[0.9], $
                          ;                        function_name='fit_atmod_func')
                          ;atmstruct.scale_molecules[indexmol]=fitparams[0]
                          ;fit peak line depths
                          ;teltool_fitpeak_column, wave, flux, order, w_fit, region, contin, nfit, best_atm=best_atm
                          ; ...or fit line profiles
                          teltool_fitprofile_column, wave, flux, error, order, w_fit, region, contin, nfit, best_atm=best_atm
                        ENDIF ELSE BEGIN
                          print, '   fitting column density saling factor (remotely) ... have patience'
                          ;print, atmstruct.scale_molecules[indexmol]
                          ;print, indexmol
                          IF atmstruct.server EQ 'psg' THEN atmod=get_psg(atmstruct, molecule, atmstruct.scale_molecules[indexmol])
                          atmstruct.scale_molecules[indexmol]=get_param_from_stemfile(atmstruct.stemfile, 'RETRIEVAL-VALUES', atmstruct)
                          ;print, atmstruct.scale_molecules                        	
                          ;stopcont, 'fghjj'
                        ENDELSE
                        print, '   found column density scaling factor '+STRTRIM(atmstruct.scale_molecules[indexmol],2)
                        ;add to array of best fitting parameters for each line for each molecule
                        scale_molecules_all_lines[nfit]=atmstruct.scale_molecules[indexmol]
                     ENDIF
   
                     ;run best-fit model as determined by local or
                     ;remote molecular scaling factor fitting in order
                     ;to make a plot or calculate model with user
                     ;input values in case wavelength needs to be
                     ;fitted
                     IF do_fitwave EQ 1 OR (do_fitmol EQ 1 AND atmstruct.do_plot EQ 1) THEN BEGIN
                       print, '   calculating model spectrum with scaling factor '+STRTRIM(atmstruct.scale_molecules[where(atmstruct.molecules EQ molecule)], 2)
                       IF atmstruct.server EQ 'psg' THEN atmod=get_psg(atmstruct)
                       IF atmstruct.server EQ 'lblrtm' THEN atmod=get_lblrtm(atmstruct)
                       ;get local baseline for this line in case there
                       ;was no column fitting..just needed for
                       ;plotting
                       IF atmstruct.do_plot EQ 1 AND ((do_fitmol NE 1) OR (do_fitmol EQ 1 AND atmstruct.fit_method EQ 'remote')) THEN BEGIN
                          best_atm=teltool_normlineflux(atmstruct,atmod[0,*], atmod[7,*], INDGEN(n_elements(atmod[7,*]))*0., INDGEN(n_elements(atmod[7,*]))*0.+1., w_fit, region, contin, nfit)
                       ENDIF
                     ENDIF
      
                     ;improve wavelength calibration (fractions of a pixel
                     ;matter), but don't apply till later
                     IF do_fitwave EQ 1 THEN BEGIN
                        ;print, 'shift [cm-1]', '|obs/model|'
                        accur=0.01 ;TBD
                        fitparams=amoeba(accur, p0=[0.], $
                                                scale=[0.9], $
                                                function_name='fit_atmod_wave_func')
                        shift=fitparams[0]
                        print, '   found wavelength shift of '+STRTRIM(shift, 2)+' cm-1'
                        plotshift=shift
                        IF keyword_set(waveshift) THEN waveshift=[waveshift,shift] ELSE waveshift=shift
                     ENDIF ELSE BEGIN
                        waveshift=[0.]
                        plotshift=[0.]
                     ENDELSE

                     ;oplot,  atmod[0,*], atmod[7,*]/interpol(best_atm[5,*],best_atm[0,*],atmod[0,*]), color=3
                     ;stop

                     ;plot this line and model in case column scaling
                     ;factor or wavelength was fitted
                     IF atmstruct.do_plot EQ 1 AND (do_fitwave EQ 1 OR do_fitmol EQ 1) THEN BEGIN
                       loadct, 2, /silent
                       plot, atmstruct.fit_wavenumber, normline[3,*], xr=[w_reg[1],w_reg[0]],ps=10, title='teltool fit of atmospheric line of '+molecule
                       oplot, atmstruct.fit_wavenumber+plotshift, flux[index], ps=10, color=30
                       oplot, atmstruct.fit_wavenumber, normline[5,*], color=20
    
                       oplot,  atmod[0,*], atmod[9,*]/interpol(best_atm[5,*],best_atm[0,*],atmod[0,*])*poly(atmod[0,*],[const_normline,rico_normline]), color=180
                       oplot,  atmod[0,*], atmod[7,*]/interpol(best_atm[5,*],best_atm[0,*],atmod[0,*])*poly(atmod[0,*],[const_normline,rico_normline]), color=80
                     
                       oplot, c_reg1[0]*[1,1], [0,1.e6], linestyle=1
                       oplot, c_reg1[1]*[1,1], [0,1.e6], linestyle=1      
                       oplot, c_reg2[0]*[1,1], [0,1.e6], linestyle=1
                       oplot, c_reg2[1]*[1,1], [0,1.e6], linestyle=1

                       del=0.02
                       xyouts, 0.1, 0.40+del, 'observed spectrum (black histogram)', /normal
                       xyouts, 0.1, 0.35+del, 'baseline region (black vertical dots)', /normal
                       xyouts, 0.1, 0.30+del, 'baseline (green)', color=20, /normal                      
                       xyouts, 0.1, 0.25+del, 'model at R=500,000 (purple)', color=180, /normal
                       xyouts, 0.1, 0.20+del, 'model at R='+STRTRIM(round(float(atmstruct.resolving_power)),2)+' (red)', color=80, /normal
                       xyouts, 0.1, 0.15+del, 'applied column scaling factor '+STRTRIM(atmstruct.scale_molecules[where(atmstruct.molecules EQ molecule)], 2), /normal
                       xyouts, 0.1, 0.10+del, 'applied wavenumber shift '+STRTRIM(waveshift[0],2)+' cm-1', /normal
                       
                       ;no stops if plotting to PS file
                       IF atmstruct.plotfile EQ 'none' THEN BEGIN
                         print, '***type .c for next'
                         stop
                       ENDIF
                     ENDIF

                  ENDFOR        ;end of loop over all lines for one molecule
                  ;
                  ;take median over all lines per molecule as best
                  ;column density scaling factor. If
                  ;only 2, take mean.
                  IF do_fitmol EQ 1 THEN BEGIN
                   atmstruct.scale_molecules[indexmol]=MEDIAN(scale_molecules_all_lines,/even)
                   print, '***'+molecule+' median scaling factor across lines', atmstruct.scale_molecules[indexmol]
                  ENDIF
               ENDIF    

               ;make sure that remaining molecules to be fitted are
               ;recognized as such
               IF countmin GT 1 THEN BEGIN
                  indexmin=indexmin[where(indexmin NE indexmol)]
                  atmstruct.scale_molecules[indexmin]=1.                           
               ENDIF

          ENDIF ELSE BEGIN
             IF do_fitmol EQ 1 THEN BEGIN
                print, '!!!!!!!'
                print, 'WARNING: fitting of ', molecule, ' not possible: no lines defined in teltool_lines.dat'
                print, '         default scaling factor of 1.0 taken'
                print, '!!!!!!!'
                atmstruct.scale_molecules[indexmol]=1.0
             ENDIF
          ENDELSE
          
       ENDFOR ;end loop over molecules

       ;save fitting plots in file
       IF atmstruct.plotfile NE 'none' THEN BEGIN
          device, /close
          set_plot, 'x'       
       ENDIF    
       
       scale_molecules_in=atmstruct.scale_molecules
   ENDIF

   IF atmstruct.onlyfit NE 'yes' THEN BEGIN
   
     ;Now that we fitted all the parameters, get best-fit model for
     ;entire observed wavelength range. If no fitting was done, and all
     ;orders need to be corrected, do this before the order loop so the
     ;entire spectrum can be obtained at once, and the server does not
     ;have to be accessed many times.
     if atmstruct.order GT 0 OR keyword_set(dofit) THEN BEGIN
       atmstruct.wavenumber=atmstruct.wavenumber_corr
       atmod_get_full_spectrum
     ENDIF
       
     ;improve wavelength calibration, but only when shift determined by
     ;fitting above. User-supplied shifts are done at start of code.
     IF keyword_set(waveshift) THEN wave=wave+MEDIAN([waveshift])   

     ;save model and division in output arrays
     IF atmstruct.order EQ -1 THEN BEGIN
       ;output atmospheric model array with flux column replaced by
       ;atmospheric model
       ;make sure the data are sorted in micron
       ind=sort(1.e4/atmod[0,*])
       ;same atmosphere for each aperture
       FOR aper=0, atmstruct.naps-1, 1 DO BEGIN
        readdata_atm[*, 1, (o-begin_order)*atmstruct.naps+aper]=interpol(atmod[7,ind], 1.e4/atmod[0,ind], readdata_atm[*, 0, (o-begin_order)*atmstruct.naps+aper])
        ;and with transmission per molecule added as well
        FOR nm=0,nmols_from_code-1 DO BEGIN
           readdata_atm[*, 4+nm, (o-begin_order)*atmstruct.naps+aper]=interpol(atmod[10+nm,ind], 1.e4/atmod[0,ind], readdata_atm[*, 0, (o-begin_order)*atmstruct.naps+aper])       
        ENDFOR
        ;output division array with flux column replaced by flux divided by
        ;atmospheric model
        readdata_div[*, 1, (o-begin_order)*atmstruct.naps+aper]=readdata[*, 1, (o-begin_order)*atmstruct.naps+aper]/interpol(atmod[7,ind], 1.e4/atmod[0,ind], readdata_atm[*, 0, (o-begin_order)*atmstruct.naps+aper])
        readdata_div[*, 2, (o-begin_order)*atmstruct.naps+aper]=readdata[*, 2, (o-begin_order)*atmstruct.naps+aper]/interpol(atmod[7,ind], 1.e4/atmod[0,ind], readdata_atm[*, 0, (o-begin_order)*atmstruct.naps+aper])         
       ENDFOR
     ENDIF    
     
     ;plot data, best fit, and data/model division   
     IF atmstruct.do_plot GT 0 AND atmstruct.orphan EQ 0 THEN BEGIN
      ;for plotting, get the wavenumber range for this order
      atmstruct.wavenumber=atmstruct.wavenumber_corr
      !x.range=atmstruct.wavenumber
      !y.range=[0,2.*max(flux[where(wave GT atmstruct.wavenumber[0] AND wave LT atmstruct.wavenumber[1])])]  
      plot, wave, flux, ps=10, title='Best Fit Atmospheric Model '+spectralordertitle
      oplot,  atmod[0,*], atmod[7,*]*median(flux[where(wave GT atmstruct.wavenumber[0] AND wave LT atmstruct.wavenumber[1])]), color=80
      oplot, wave, flux/interpol(atmod[7,*], atmod[0,*], wave),ps=10, color=30
      IF o NE 0 THEN BEGIN
         print, '***echelle order '+STRTRIM(o)+' plotted'
         print, '   type .c for next'
         stop
      ENDIF
     ENDIF

   ENDIF ;end of onlyfit=no block
     
ENDFOR ;end of loop over orders

print, '***processing time [sec]: ', systime(1)-start_time

IF atmstruct.orphan EQ 1 THEN BEGIN
 ;
 path=atmstruct.obsdir
 close, 1
 print, '***writing atmospheric model to file', path+orphan_in
 openw, 1, path+orphan_in

 nr=n_elements(atmod(0,*))
 nc=n_elements(atmod(*,0))
 nh=6
 head=STRARR(1,nh)

 head[0,2]='wavenumber                                                     radiance  transm    radiance  transm    transm    transm    transm    transm    transm    transm    transm    transm    transm    transm    transm    transm'
 head[0,3]='                                                               unconvol  convol    convol    unconvol  convol    convol    convol    convol    convol    convol    convol    convol    convol    convol    convol    convol'
 head[0,4]='  cm-1                                                         all mol   all mol   all mol   all mol'
;           3030.405265  0.000000  0.000000  0.000000  0.000000  0.000000  0.032499  0.079434  0.032526  0.078950  0.079492  1.000000  0.999265  1.000000  1.000000  1.000000  1.000000  1.000000  0.999997  1.000000  1.000000  1.000000

 FOR i=0,n_elements(atmstruct.psg_molecules)-1 DO BEGIN
   head[0,4]=head[0,4]+'       '+(atmstruct.psg_molecules)[i]
 ENDFOR
 
; 0: Frequency vector
; 6: Fully resolved radiance
; 7: Convolved transmittance
; 8: Convolved radiance
; 9: Fully resolved transmittance
; 10-9+nmols: Convolved transmittance per molecule

 colrow=INTARR(3,1)
 colrow=[nr,nc,nh+2]

 FOR i=0,nc-1 DO BEGIN
  IF i EQ 0 THEN format='f12.6' ELSE format=format+','+'f10.6'
 ENDFOR
 format='('+format+')'

 printf, 1, colrow
 printf, 1, systime()
 printf, 1, orphan_in
 printf, 1, head
 printf, 1, atmod, FORMAT=format
 
 close, 1 
ENDIF

IF arg_present(teltool_vars) THEN teltool_vars=atmstruct

end

