;+
; NAME:
;     xtellcor_model
;    
; PURPOSE:
;     Runs iSHELL telluric correction using atmospheric models
;    
; CATEGORY:
;     Widget
;
; CALLING SEQUENCE:
;     xtellcor_model
;    
; INPUTS:
;     None
;    
; OPTIONAL INPUTS:
;     None
;
; KEYWORD PARAMETERS:
;     None
;     
; OUTPUTS:
;     Writes a iSHELL spectral FITS file to disk of the telluric
;     corrected spectra and the telluric correction spectrum.
;
; OPTIONAL OUTPUTS:
;     None
;
; COMMON BLOCKS:
;     None
;
; SIDE EFFECTS:
;     None
;
; RESTRICTIONS:
;     None
;
; PROCEDURE:
;     Follow the directions
;
; EXAMPLE:
;     
; MODIFICATION HISTORY:
;     2018-07-31 A. Boogert: adjusted from xtellcor written by
;                            M. Cushing (version from 2008)
;     2020-05-21 A. Boogert: make compatible with non-averaged spectra
;                            (that don't have AVE_DATE and AVE_TIME
;                            keywords)
;
;===============================================================================
;
; ------------------------------Event Handlers-------------------------------- 
;
;===============================================================================
;
pro xtellcor_model_event, event

  widget_control, event.id,  GET_UVALUE=uvalue
  if uvalue eq 'Quit' then begin
     
     widget_control, event.top, /DESTROY
     goto, getout
     
  endif
    
  widget_control, event.top, GET_UVALUE=state, /NO_COPY
  widget_control, /HOURGLASS
  
  case uvalue of     
     
     'Load Spectra': xtellcor_loadspec,state

     'Object Spectra Button': begin
        
        obj = dialog_pickfile(DIALOG_PARENT=state.xtellcor_base,$
                              PATH=state.objpath,/MUST_EXIST, $
                              GET_PATH=path,FILTER='*.fits')
        if obj eq '' then goto, cont
        widget_control,state.objspectra_fld[1],SET_VALUE=file_basename(obj)
        state.objpath = path
        mc_setfocus,state.objspectra_fld
        
     end
     
     'Get Atmospheric Model': xtellcor_fitatmmodel,state ;ACAB
     
     'Get Shifts': xtellcor_getshifts,state

     'Load Flat Spectra': xtellcor_loadflatspec,state     
     
     'Flat Spectrum Button': begin
        
        obj = dialog_pickfile(DIALOG_PARENT=state.xtellcor_base,$
                              PATH=state.objpath,/MUST_EXIST, $
                              GET_PATH=path,FILTER='*.fits')
        if obj eq '' then goto, cont
        widget_control,state.flatspectra_fld[1],SET_VALUE=file_basename(obj)
        state.flatpath = path
        mc_setfocus,state.flatspectra_fld
        
     end

     'Use Flat Field Blaze Profile': xtellcor_flatblaze,state

     'Use Stored Blaze Profile': xtellcor_storedblaze,state          
     
     'Skip Correct Blaze': xtellcor_blaze_skip,state     
     
     'Help': begin

        pre = (strupcase(!version.os_family) eq 'WINDOWS') ? 'start ':'open '
        
        spawn, pre+filepath(strlowcase(state.instrument)+'_spextoolmanual.pdf',$
                            ROOT=state.spextoolpath,$
                            SUBDIR='manuals')
        
     end
                    
     'Write File': xtellcor_writefile,state
     
     else:
     
  endcase
  
;  Put state variable into the user value of the top level base.
   
  cont: 
  widget_control, state.xtellcor_base, SET_UVALUE=state, /NO_COPY
  getout:
  
end
;
;===============================================================================
;
; ----------------------------Support procedures------------------------------ 
;
;===============================================================================
;
pro xtellcor_cleanup,base

  widget_control, base, GET_UVALUE = state, /NO_COPY
  if n_elements(state) ne 0 then begin
     
     ptr_free, state.shift
     ptr_free, state.blazeshift     
     ptr_free, state.objhdr
     ptr_free, state.objorders
     ptr_free, state.objspec
     ptr_free, state.objdisp
     ptr_free, state.tellspec
     ptr_free, state.blazespec
     ptr_free, state.blazecorspec     
     ptr_free, state.corspec          
     
  endif
  state = 0B
  
end
;
;===============================================================================
;
pro xtellcor_fitatmmodel,state

  if state.continue lt 1 then begin
     
     ok = dialog_message('Previous steps not complete.',/ERROR,$
                         DIALOG_PARENT=state.xtellcor_base)
     return
     
  endif

  IF state.column_scaleh2o[0] NE -100.  THEN scaleh2o=mc_cfld(state.column_scaleh2o, 4,/empty) ELSE scaleh2o=0.
  IF state.column_scalech4[0] NE -100.  THEN scalech4=mc_cfld(state.column_scalech4, 4,/empty) ELSE scalech4=0.
  IF state.column_scaleco2[0] NE -100.  THEN scaleco2=mc_cfld(state.column_scaleco2, 4,/empty) ELSE scaleco2=0.
  IF state.column_scaleco[0] NE -100.   THEN scaleco =mc_cfld(state.column_scaleco,  4,/empty) ELSE scaleco=0.
  IF state.column_scalen2o[0] NE -100.  THEN scalen2o=mc_cfld(state.column_scalen2o, 4,/empty) ELSE scalen2o=0.
  IF state.column_scaleo3[0] NE -100.   THEN scaleo3 =mc_cfld(state.column_scaleo3,  4,/empty) ELSE scaleo3=0.
  IF state.column_scaleocs[0] NE -100.  THEN scaleocs =mc_cfld(state.column_scaleocs,  4,/empty) ELSE scaleocs=0.
  IF state.column_scaleo2[0] NE -100.   THEN scaleo2 =mc_cfld(state.column_scaleo2,  4,/empty) ELSE scaleo2=0.
  IF state.column_scalec2h6[0] NE -100. THEN scalec2h6 =mc_cfld(state.column_scalec2h6,  4,/empty) ELSE scalec2h6=0.
  IF state.column_scalehcl[0] NE -100.  THEN scalehcl =mc_cfld(state.column_scalehcl,  4,/empty) ELSE scalehcl=0.
  IF state.column_scalen2[0] NE -100.   THEN scalen2 =mc_cfld(state.column_scalen2,  4,/empty) ELSE scalen2=0.

  ;fit to data if the scale parameter is -1
  IF scaleh2o EQ -1. THEN state.column_fitH2O=1 ELSE state.column_fitH2O=0
  IF scalech4 EQ -1. THEN state.column_fitCH4=1 ELSE state.column_fitCH4=0
  IF scaleco2 EQ -1. THEN state.column_fitCO2=1 ELSE state.column_fitCO2=0
  IF scaleco  EQ -1. THEN state.column_fitCO =1 ELSE state.column_fitCO =0
  IF scalen2o EQ -1. THEN state.column_fitN2O=1 ELSE state.column_fitN2O=0
  IF scaleo3  EQ -1. THEN state.column_fitO3 =1 ELSE state.column_fitO3 =0
  IF scaleocs EQ -1. THEN state.column_fitOCS=1 ELSE state.column_fitOCS=0
  IF scaleo2 EQ -1. THEN state.column_fitO2=1 ELSE state.column_fitO2=0
  IF scalec2h6 EQ -1. THEN state.column_fitC2H6=1 ELSE state.column_fitC2H6=0
  IF scalehcl EQ -1. THEN state.column_fitHCl=1 ELSE state.column_fitHCl=0
  IF scalen2 EQ -1. THEN state.column_fitN2=1 ELSE state.column_fitN2=0  
  
  IF state.column_fitH2O THEN column_fitH2O='yes' ELSE column_fitH2O='no'
  IF state.column_fitCH4 THEN column_fitCH4='yes' ELSE column_fitCH4='no'
  IF state.column_fitCO2 THEN column_fitCO2='yes' ELSE column_fitCO2='no' 
  IF state.column_fitCO  THEN column_fitCO ='yes' ELSE column_fitCO='no'
  IF state.column_fitN2O THEN column_fitN2O='yes' ELSE column_fitN2O='no'
  IF state.column_fitO3  THEN column_fitO3 ='yes' ELSE column_fitO3='no'
  IF state.column_fitOCS  THEN column_fitOCS ='yes' ELSE column_fitOCS='no'
  IF state.column_fitO2  THEN column_fitO2 ='yes' ELSE column_fitO2='no'
  IF state.column_fitC2H6  THEN column_fitC2H6 ='yes' ELSE column_fitC2H6='no'
  IF state.column_fitHCl  THEN column_fitHCl ='yes' ELSE column_fitHCl='no'
  IF state.column_fitN2  THEN column_fitN2 ='yes' ELSE column_fitN2='no'       
  ;print, 'Column scaling factors for H2O, CH4, and CO2: ', scaleh2o, scalech4, scaleco2
  ;print, 'Fitting H2O, CH4, and CO2: ', column_fitH2O, ' ', column_fitCH4, ' ', column_fitCO2  
  
  ;First, determine best column scaling factors by fitting to observed
  ;lines, if requested. Otherwise use the values entered by user.
  objfile=mc_cfld(state.objspectra_fld,7,/EMPTY,CANCEL=cancel)
  stemfile=STRMID(objfile,0,(strsplit(objfile, "."))[1]-1)
  ;get location of iSHELL parameter file
  teltool_paramdir=teltool_directory()
  teltool_paramfile=teltool_paramdir+'teltool_ishell.param' ;fixed parameters needed for running teltool
  state.molecules=['H2O','CO2','CH4','CO','N2O','O3','OCS', 'O2', 'C2H6', 'HCl', 'N2']
  IF state.column_fitH2O OR state.column_fitCH4 OR state.column_fitCO2 OR $
     state.column_fitCO  OR state.column_fitN2O OR state.column_fitO3 OR state.column_fitOCS OR $
     state.column_fitO2  OR state.column_fitC2H6 OR state.column_fitHCl OR state.column_fitN2 THEN BEGIN
       ;save fitted lines in PS file (because teltool has do_plot=1 option
       fitpsfile=state.objpath+'xtellcor_model_fittedlines_'+state.config+'_'+strmid(state.objfile,0,strpos(state.objfile,'.fi'))+'.ps' 
       scale_molecules_in=[scaleh2o, scaleco2, scalech4, scaleco, scalen2o, scaleo3, scaleocs, scaleo2, scalec2h6, scalehcl, scalen2]
       teltool, teltool_paramfile, do_plot=1, order=-2, onlyfit='yes', $
                  molecules=state.molecules, scale_molecules=scale_molecules_in, $
                  fit_molecules=[column_fitH2O, column_fitCO2, column_fitCH4,column_fitCO, column_fitN2O, column_fitO3, column_fitOCS, column_fitO2, column_fitC2H6, column_fitHCl, column_fitN2], $
                  stemfile=stemfile, obsfile=objfile, obsdir=state.objpath, plotfile=fitpsfile
       state.best_scale_molecules=scale_molecules_in
  ENDIF ELSE state.best_scale_molecules=[scaleh2o, scaleco2, scalech4, scaleco, scalen2o, scaleo3, scaleocs, scaleo2, scalec2h6, scalehcl, scalen2]

  debug_acab=0
  IF debug_acab NE 1 THEN BEGIN
  
     ;Then, get the best model over the full wavelength range
     teltool, teltool_paramfile, atm=atm, do_plot=0, order=-1, $
                molecules=state.molecules, scale_molecules=state.best_scale_molecules, $
                fit_molecules=['no', 'no', 'no','no', 'no', 'no', 'no', 'no', 'no', 'no', 'no'], stemfile=stemfile, onlyfit='no', $
                obsfile=objfile, obsdir=state.objpath, teltool_vars

     ;add selected teltool_vars elements to state structure
     state.server=teltool_vars.server
     state.resolving_power=teltool_vars.resolving_power
     state.auxdir=teltool_vars.auxdir
     IF tag_exist(state, 'psg_molecules') THEN struct_delete_field, state, 'psg_molecules'
     state2={psg_molecules : teltool_vars.psg_molecules}
     state1=state
     state=create_struct(state1, state2)
  ENDIF ELSE BEGIN
     atm=*state.objspec
     atm[*,1,*]=1.
  ENDELSE
  
  ;subsequent steps require inverse of transmission
  atm[*,1,*]=1./atm[*,1,*]

  ;put atmospheric model in iSHELL-type array (separated by order, etc)
  *state.tellspec=atm

  ;ACAB: put atm transmission model for all molecules in the relevant
  ;array for overplotting in getshifts step.
  ;                              
  ;remove order information because that's what xmc_findshifts_model
  ;wants

  dimen_atm=size(atm,/dimen)
  atm_glue=DBLARR(dimen_atm[1]-2,dimen_atm[0]*dimen_atm[2])
  ;loop over orders
  FOR o=0,dimen_atm[2]-1 DO BEGIN
     ;wavelength
     atm_glue[0,o*dimen_atm[0]:(o+1)*dimen_atm[0]-1]=atm[*,0,o]
     ;transmission summed over all molecules
     ;(taking into account that transmission was inverted above)
     atm_glue[1,o*dimen_atm[0]:(o+1)*dimen_atm[0]-1]=1./atm[*,1,o]
     ;loop over transmission for individual molecules
     FOR c=4,dimen_atm[1]-1 DO BEGIN
       atm_glue[c-2,o*dimen_atm[0]:(o+1)*dimen_atm[0]-1]=atm[*,c,o]
     ENDFOR  
  ENDFOR
  ;sort over wavelength
  ind=sort(atm_glue[0,*])
  FOR c=0,dimen_atm[1]-2-1 DO BEGIN
   atm_glue[c,*]=atm_glue[c,ind]
  ENDFOR

  *state.awave = atm_glue[0,*]
  *state.atrans = atm_glue[1:*,*]

  ;update the widget with the scaling factors found

  IF column_fitH2O  EQ 'yes' THEN WIDGET_CONTROL, state.column_scaleh2o[1], SET_VALUE=strtrim(state.best_scale_molecules[where(state.molecules EQ 'H2O')],2)
  IF column_fitCO2  EQ 'yes' THEN WIDGET_CONTROL, state.column_scaleco2[1], SET_VALUE=strtrim(state.best_scale_molecules[where(state.molecules EQ 'CO2')],2)
  IF column_fitCH4  EQ 'yes' THEN WIDGET_CONTROL, state.column_scalech4[1], SET_VALUE=strtrim(state.best_scale_molecules[where(state.molecules EQ 'CH4')],2)
  IF column_fitCO   EQ 'yes' THEN WIDGET_CONTROL, state.column_scaleco[1],  SET_VALUE=strtrim(state.best_scale_molecules[where(state.molecules EQ 'CO')],2)
  IF column_fitN2O  EQ 'yes' THEN WIDGET_CONTROL, state.column_scalen2o[1], SET_VALUE=strtrim(state.best_scale_molecules[where(state.molecules EQ 'N2O')],2)
  IF column_fitO3   EQ 'yes' THEN WIDGET_CONTROL, state.column_scaleo3[1],  SET_VALUE=strtrim(state.best_scale_molecules[where(state.molecules EQ 'O3')],2)
  IF column_fitOCS  EQ 'yes' THEN WIDGET_CONTROL, state.column_scaleocs[1], SET_VALUE=strtrim(state.best_scale_molecules[where(state.molecules EQ 'OCS')],2)
  IF column_fitO2   EQ 'yes' THEN WIDGET_CONTROL, state.column_scaleo2[1],  SET_VALUE=strtrim(state.best_scale_molecules[where(state.molecules EQ 'O2')],2)
  IF column_fitC2H6 EQ 'yes' THEN WIDGET_CONTROL, state.column_scalec2h6[1],SET_VALUE=strtrim(state.best_scale_molecules[where(state.molecules EQ 'C2H6')],2)
  IF column_fitHCl  EQ 'yes' THEN WIDGET_CONTROL, state.column_scalehcl[1], SET_VALUE=strtrim(state.best_scale_molecules[where(state.molecules EQ 'HCl')],2)
  IF column_fitN2   EQ 'yes' THEN WIDGET_CONTROL, state.column_scalen2[1],  SET_VALUE=strtrim(state.best_scale_molecules[where(state.molecules EQ 'N2')],2)
  
  state.continue = 2            ; (state.mode eq 0) ? 2:3

end
;===============================================================================
;
pro make_mols_widget, strongmols, intermediatemols, weakmols, absentmols, state

  ;MAKE_MOLS_WIDGET.PRO: produce the molecular selection part of the
  ;                      xtellcor_model widget, which depends on the
  ;                      wavelength range studied, i.e., which
  ;                      molecules absorb there.
  ;
  ; Calling: make_mols_widget, 
  ;
  ;                        strongmols STRARR list of molecules
  ;                           >10% absorption at R=75,000
  ;
  ;                        intermediatemols STRARR list of molecules
  ;                           1-10% absorption at R=75,000
  ;
  ;                        weakmols STRARR list of molecules
  ;                           0.1-1% absorption at R=75,000
  ;
  ;                        absentmols STRARR list of molecules
  ;                           no absorption at R=75,000  
  ;
  ;                        state STRUCTURE xtellcor_model structure
  ;

  ;convert input strength to a number
  ;strong: 1
  ;intermediate: 2
  ;weak: 3
  ;absent: 4
  mols=['H2O','CO2','CH4','CO','N2O', 'O2','O3','OCS', 'C2H6', 'HCl', 'N2']
  mols_strength=INTARR(n_elements(mols))
  FOR i=0,n_elements(mols)-1 DO BEGIN
     IF where(mols[i] EQ strongmols) NE -1 THEN mols_strength[i]=1
     IF where(mols[i] EQ intermediatemols) NE -1 THEN mols_strength[i]=2
     IF where(mols[i] EQ weakmols) NE -1 THEN mols_strength[i]=3
     IF where(mols[i] EQ absentmols) NE -1 THEN mols_strength[i]=4
  ENDFOR

  box_base_mols = widget_base(state.col1_base,$
                                  COLUMN=1, $
                                  FRAME=2)
           
  label = widget_label(box_base_mols,$
                       VALUE='2. Get Atmospheric Model',$
                       FONT=buttonfont,$
                       /ALIGN_LEFT)

  IF mols_strength[0] NE 4 THEN BEGIN
     IF mols_strength[0] EQ 1 THEN importance='*' ELSE importance=' '
     colscalh2o = coyote_field2(box_base_mols,$
                                  FIELDFONT=textfont,$
                                  LABELFONT=buttonfont,$                                         
                                  title=importance+'Scale H2O Column [-1: fit] ', $
                                  /FloatValue, $
                                 UVALUE='Scale Column',$
                                VALUE=1.0, $
                                 xsize=5, $
                                 /CR_only, $
                                 textID=textidh2o)
  ENDIF ELSE BEGIN
     colscalh2o=-100.
     textidh2o=-100.
  ENDELSE

  IF mols_strength[1] NE 4 THEN BEGIN
     IF mols_strength[1] EQ 1 THEN importance='*' ELSE importance=' '     
     colscalco2 = coyote_field2(box_base_mols,$
                                  FIELDFONT=textfont,$
                                  LABELFONT=buttonfont,$                                         
                                  title=importance+'Scale CO2 Column [-1: fit] ', $
                                  /FloatValue, $
                                 UVALUE='Scale Column',$
                                 VALUE=1.0, $
                                 xsize=5, $
                                 /CR_only, $
                                 textID=textidco2)
  ENDIF ELSE BEGIN
     colscalco2=-100.
     textidco2=-100
  ENDELSE
     
  IF mols_strength[2] NE 4 THEN BEGIN
     IF mols_strength[2] EQ 1 THEN importance='*' ELSE importance=' '     
     colscalch4 = coyote_field2(box_base_mols,$
                                  FIELDFONT=textfont,$
                                  LABELFONT=buttonfont,$                                         
                                  title=importance+'Scale CH4 column [-1: fit] ', $
                                  /FloatValue, $
                                 UVALUE='Scale Column',$
                                 VALUE=1.0, $
                                 xsize=5, $
                                 /CR_only, $
                                 textID=textidch4)
  ENDIF ELSE BEGIN
     colscalch4=-100.
     textidch4=-100
  ENDELSE
     
  IF mols_strength[3] NE 4 THEN BEGIN
     IF mols_strength[3] EQ 1 THEN importance='*' ELSE importance=' '     
     colscalco = coyote_field2(box_base_mols,$
                                  FIELDFONT=textfont,$
                                  LABELFONT=buttonfont,$                                         
                                  title=importance+'Scale CO Column [-1: fit]  ', $
                                  /FloatValue, $
                                 UVALUE='Scale Column',$
                                 VALUE=1.0, $
                                 xsize=5, $
                                 /CR_only, $
                                 textID=textidco)
  ENDIF ELSE BEGIN
     colscalco =-100.
     textidco =-100.
  ENDELSE
     
  IF mols_strength[4] NE 4 THEN BEGIN
     IF mols_strength[4] EQ 1 THEN importance='*' ELSE importance=' '     
     colscaln2o = coyote_field2(box_base_mols,$
                                  FIELDFONT=textfont,$
                                  LABELFONT=buttonfont,$                                         
                                  title=importance+'Scale N2O Column [-1: fit] ', $
                                  /FloatValue, $
                                 UVALUE='Scale Column',$
                                 VALUE=1.0, $
                                 xsize=5, $
                                 /CR_only, $
                                 textID=textidn2o)
  ENDIF ELSE BEGIN
     colscaln2o=-100.  
     textidn2o=-100.
  ENDELSE
     
  IF mols_strength[5] NE 4 THEN BEGIN
     IF mols_strength[5] EQ 1 THEN importance='*' ELSE importance=' '     
     colscalo2 = coyote_field2(box_base_mols,$
                                  FIELDFONT=textfont,$
                                  LABELFONT=buttonfont,$                                         
                                  title=importance+'Scale O2 Column [-1: fit]  ', $
                                  /FloatValue, $
                                 UVALUE='Scale Column',$
                                 VALUE=1.0, $
                                 xsize=5, $
                                 /CR_only, $
                                 textID=textido2)
  ENDIF ELSE BEGIN
     colscalo2 =-100.
     textido2 =-100.
  ENDELSE
     
  IF mols_strength[6] NE 4 THEN BEGIN
     IF mols_strength[6] EQ 1 THEN importance='*' ELSE importance=' '     
     colscalo3 = coyote_field2(box_base_mols,$
                                  FIELDFONT=textfont,$
                                  LABELFONT=buttonfont,$                                         
                                  title=importance+'Scale O3 Column [-1: fit]  ', $
                                  /FloatValue, $
                                 UVALUE='Scale Column',$
                                 VALUE=1.0, $
                                 xsize=5, $
                                 /CR_only, $
                                 textID=textido3)
  ENDIF ELSE BEGIN
     colscalo3 =-100.
     textido3 =-100.
  ENDELSE
     
  IF mols_strength[7] NE 4 THEN BEGIN
     IF mols_strength[7] EQ 1 THEN importance='*' ELSE importance=' '     
     colscalocs = coyote_field2(box_base_mols,$
                                  FIELDFONT=textfont,$
                                  LABELFONT=buttonfont,$                                         
                                  title=importance+'Scale OCS Column [-1: fit] ', $
                                  /FloatValue, $
                                 UVALUE='Scale Column',$
                                 VALUE=1.0, $
                                 xsize=5, $
                                 /CR_only, $
                                 textID=textidocs)
  ENDIF ELSE BEGIN
     colscalocs=-100.
     textidocs=-100.
  ENDELSE
     
  IF mols_strength[8] NE 4 THEN BEGIN
     IF mols_strength[8] EQ 1 THEN importance='*' ELSE importance=' '     
     colscalc2h6 = coyote_field2(box_base_mols,$
                                  FIELDFONT=textfont,$
                                  LABELFONT=buttonfont,$                                         
                                  title=importance+'Scale C2H6 Column [-1: fit] ', $
                                  /FloatValue, $
                                 UVALUE='Scale Column',$
                                 VALUE=1.0, $
                                 xsize=5, $
                                 /CR_only, $
                                 textID=textidc2h6)
  ENDIF ELSE BEGIN
     colscalc2h6=-100.
     textidc2h6=-100.
  ENDELSE
     
  IF mols_strength[9] NE 4 THEN BEGIN
     IF mols_strength[9] EQ 1 THEN importance='*' ELSE importance=' '     
     colscalhcl = coyote_field2(box_base_mols,$
                                  FIELDFONT=textfont,$
                                  LABELFONT=buttonfont,$                                         
                                  title=importance+'Scale HCl Column [-1: fit] ', $
                                  /FloatValue, $
                                 UVALUE='Scale Column',$
                                 VALUE=1.0, $
                                 xsize=5, $
                                 /CR_only, $
                                 textID=textidhcl)
  ENDIF ELSE BEGIN
     colscalhcl=-100.
     textidhcl=-100.
  ENDELSE
     
  IF mols_strength[10] NE 4 THEN BEGIN
     IF mols_strength[10] EQ 1 THEN importance='*' ELSE importance=' '     
     colscaln2 = coyote_field2(box_base_mols,$
                                  FIELDFONT=textfont,$
                                  LABELFONT=buttonfont,$                                         
                                  title=importance+'Scale N2 Column [-1: fit]  ', $
                                  /FloatValue, $
                                 UVALUE='Scale Column',$
                                 VALUE=1.0, $
                                 xsize=5, $
                                 /CR_only, $
                                 textID=textidn2)
  ENDIF ELSE BEGIN
     colscaln2=-100 
     textidn2=-100
  ENDELSE
     
  state.column_scaleh2o = [colscalh2o,textidh2o]
  state.column_scalech4 = [colscalch4,textidch4]
  state.column_scaleco2 = [colscalco2,textidco2]              
  state.column_scaleco = [colscalco,textidco]
  state.column_scalen2o = [colscaln2o,textidn2o]
  state.column_scaleo3 = [colscalo3,textido3]
  state.column_scaleocs = [colscalocs,textidocs]                            
  state.column_scaleo2 = [colscalo2,textido2]
  state.column_scalec2h6 = [colscalc2h6,textidc2h6]
  state.column_scalehcl = [colscalhcl,textidhcl]
  state.column_scalen2 = [colscaln2,textidn2]              

  button = widget_button(box_base_mols,$
                         VALUE='Retrieve Atmospheric Model',$
                         UVALUE='Get Atmospheric Model',$ ;this relates to an action in 'events'
                         FONT=buttonfont)

  IF keyword_set(state.box_base_mols) THEN WIDGET_CONTROL, state.box_base_mols, /destroy
  state.box_base_mols=box_base_mols
  ;complains about /redraw not allowed in IDL 7
  ;WIDGET_CONTROL, state.col1_base, /redraw
  WIDGET_CONTROL, state.col1_base, /update
  ;WIDGET_CONTROL, state.column_scaleh2o[1], SET_VALUE='1.'

end
;
;===============================================================================
;
pro xtellcor_loadspec,state

;  Get user inputs

  objfile = mc_cfld(state.objspectra_fld,7,/EMPTY,CANCEL=cancel)
  state.objfile=objfile
  if cancel then return
  objfile = mc_cfile(state.objpath+objfile,WIDGET_ID=state.xtellcor_base, $
                 CANCEL=cancel)
  if cancel then return
  
;  Read object spectra

  mc_readspec,objfile,objspec,objhdr,objobsmode,start,stop,objnorders,objnaps, $
              objorders,objxunits,objyunits,slith_pix,slith_arc,slitw_pix, $
              slitw_arc,objrp,objairmass,CANCEL=cancel
  if cancel then return  
  
  xtitle='xtitle'
  ytitle='ytitle'
  instr=' '
  
;  Store necessary information

  *state.objdisp   = mc_fxpar(objhdr,'DISP*')
  *state.objspec   = objspec
  *state.objorders = objorders
  *state.objhdr    = objhdr
  state.objairmass    = objairmass
  state.objobsmode    = objobsmode
  state.objnaps    = objnaps
  state.xtitle     = xtitle
  state.ytitle     = ytitle
  state.slitw_pix  = slitw_pix
  state.slitw_arc  = slitw_arc
    
;  Generate arrays and initialize things

  *state.shift   = fltarr(objnorders,objnaps)
  
;  Determine xtellcor mode and modify widget accordingly

  readcol, filepath('xtellcor_modeinfo.dat',ROOT_DIR=state.packagepath,$
                    SUBDIR='data'),imode,xmode,vegarp,method,output, $
           FORMAT='A,I,A,A,A',DELIMITER='|',COMMENT='#',/SILENT

  z = where(strtrim(imode,2) eq state.objobsmode,cnt)
  state.mode = xmode[z]

  widget_control, SENSITIVE=~state.mode
  widget_control,SENSITIVE=state.mode
  widget_control,SENSITIVE=state.mode
  
;  Color of upper widget bar. Green if OK. Red if there is an issue.

  color = 16 ;:17
  wset, state.message
  erase,COLOR=color

  state.config=STRTRIM(sxpar(*state.objhdr,'MODE'),2)

;  Check if data are an average over integrations, or a single
;  integration (added 05/21/20)

  IF sxpar(*state.objhdr, "AVE_TIME") EQ 0 THEN BEGIN
     datetime=STRTRIM(sxpar(*state.objhdr, "DATE"),2)+'T'+STRMID(STRTRIM(sxpar(*state.objhdr, "TIME"),2),0,8)
  ENDIF ELSE BEGIN
     datetime=STRTRIM(sxpar(*state.objhdr, "AVE_DATE"),2)+'T'+STRMID(STRTRIM(sxpar(*state.objhdr, "AVE_TIME"),2),0,8)        
  ENDELSE
     
  xyouts,10,8, state.instrument+' '+state.config+', Slit '+STRMID(STRTRIM(state.slitw_arc,2),0,5)+'", '+'Airmass '+string(objairmass,FORMAT='(f6.3)')+', '+datetime,/DEVICE,$
         CHARSIZE=1.1,FONT=0

  ;Now that we know the iSHELL configuration, set the default column
  ;scaling factors, i.e., 0.0 if that species has no transmission
  ;larger than 0.1% within that wavelength range. It would be good if
  ;we could also indicate this in a different way, e.g., by color or
  ;even removing that species. Similarly, we should try to highlight
  ;which species matter most for this configuration (>10% transmission).

  ;For each configuration, define the constributing molecules
  ;depending on maximum line depth at R=75,000.
  ;Defaults:
  strongmols=['H2O', 'CO2', 'CH4']
  intermediatemols=['O2', 'N2O', 'O3', 'CO']
  weakmols=['N2']
  absentmols=['C2H6', 'HCl', 'OCS']
  
  IF state.config EQ 'J0' THEN BEGIN
     ;molecules relevant for this configuration
     strongmols=['H2O']
     intermediatemols=['CH4']
     weakmols=['O2']
     absentmols=['CO2','CO','N2O','O3','OCS', 'C2H6', 'HCl', 'N2']

     ;WIDGET_CONTROL, state.column_scaleh2o[1], SET_VALUE='1.'
  ENDIF

  IF state.config EQ 'J1' THEN BEGIN
     ;molecules relevant for this configuration
     strongmols=['H2O', 'CO2']
     intermediatemols=['none']
     weakmols=['CH4']
     absentmols=['CO','N2O','O3','OCS', 'O2', 'C2H6', 'HCl', 'N2']

     ;WIDGET_CONTROL, state.column_scaleh2o[1], SET_VALUE='1.'
  ENDIF

  IF state.config EQ 'J2' THEN BEGIN
     ;molecules relevant for this configuration
     strongmols=['H2O', 'O2']
     intermediatemols=['CO2']
     weakmols=['none']
     absentmols=['CH4','CO','N2O','O3','OCS', 'C2H6', 'HCl', 'N2']

     ;WIDGET_CONTROL, state.column_scaleh2o[1], SET_VALUE='1.'
  ENDIF

  IF state.config EQ 'J3' THEN BEGIN
     ;molecules relevant for this configuration
     strongmols=['H2O', 'O2']
     intermediatemols=['CO2', 'CH4']
     weakmols=['none']
     absentmols=['CO','N2O','O3','OCS', 'C2H6', 'HCl', 'N2']

     ;WIDGET_CONTROL, state.column_scaleh2o[1], SET_VALUE='1.'
  ENDIF

  IF state.config EQ 'H1' THEN BEGIN
     ;molecules relevant for this configuration
     strongmols=['H2O', 'CO2', 'CH4']
     intermediatemols=['none']
     weakmols=['none']
     absentmols=['CO','N2O','O3','OCS', 'O2', 'C2H6', 'HCl', 'N2']

     ;WIDGET_CONTROL, state.column_scaleh2o[1], SET_VALUE='1.'
  ENDIF

  IF state.config EQ 'H2' THEN BEGIN
     ;molecules relevant for this configuration
     strongmols=['H2O', 'CO2', 'CH4']
     intermediatemols=['none']
     weakmols=['none']
     absentmols=['CO','N2O','O3','OCS', 'O2', 'C2H6', 'HCl', 'N2']

     ;WIDGET_CONTROL, state.column_scaleh2o[1], SET_VALUE='1.'
  ENDIF

  IF state.config EQ 'H3' THEN BEGIN
     ;molecules relevant for this configuration
     strongmols=['H2O', 'CO2', 'CH4']
     intermediatemols=['none']
     weakmols=['none']
     absentmols=['CO','N2O','O3','OCS', 'O2', 'C2H6', 'HCl', 'N2']

     ;WIDGET_CONTROL, state.column_scaleh2o[1], SET_VALUE='1.'
  ENDIF

  IF state.config EQ 'K1' THEN BEGIN
     ;molecules relevant for this configuration
     strongmols=['H2O', 'CO2', 'CH4']
     intermediatemols=['N2O']
     weakmols=['O3']
     absentmols=['CO','OCS', 'O2', 'C2H6', 'HCl', 'N2']

     ;WIDGET_CONTROL, state.column_scaleh2o[1], SET_VALUE='1.'
  ENDIF

  IF state.config EQ 'K2' THEN BEGIN
     ;molecules relevant for this configuration
     strongmols=['H2O', 'CH4']
     intermediatemols=['CO2','N2O','CO']
     weakmols=['O3']
     absentmols=['OCS', 'O2', 'C2H6', 'HCl', 'N2']

     ;WIDGET_CONTROL, state.column_scaleh2o[1], SET_VALUE='1.'
  ENDIF

  IF state.config EQ 'Kgas' THEN BEGIN
     ;molecules relevant for this configuration
     strongmols=['H2O', 'CH4']
     intermediatemols=['CO','N2O']
     weakmols=['CO2']
     absentmols=['O3','OCS', 'O2', 'C2H6', 'HCl', 'N2']

     ;WIDGET_CONTROL, state.column_scaleh2o[1], SET_VALUE='1.'
  ENDIF

  IF state.config EQ 'K3' THEN BEGIN
     ;molecules relevant for this configuration
     strongmols=['H2O', 'CH4']
     intermediatemols=['CO','N2O','O3']
     weakmols=['CO2']
     absentmols=['OCS', 'O2', 'C2H6', 'HCl', 'N2']

     ;WIDGET_CONTROL, state.column_scaleh2o[1], SET_VALUE='1.'
  ENDIF

  IF state.config EQ 'L1' THEN BEGIN
     ;molecules relevant for this configuration
     strongmols=['H2O', 'CO2', 'N2O']
     intermediatemols=['none']
     weakmols=['CH4', 'O3']
     absentmols=['CO','OCS', 'O2', 'C2H6', 'HCl', 'N2']

     ;WIDGET_CONTROL, state.column_scaleh2o[1], SET_VALUE='1.'
  ENDIF

  IF state.config EQ 'L2' THEN BEGIN
     ;molecules relevant for this configuration
     strongmols=['H2O', 'CO2', 'N2O', 'CH4']
     intermediatemols=['O3']
     weakmols=['none']
     absentmols=['CO','OCS', 'O2', 'C2H6', 'HCl', 'N2']

     ;WIDGET_CONTROL, state.column_scaleh2o[1], SET_VALUE='1.'
  ENDIF

  IF state.config EQ 'L3' THEN BEGIN
     ;molecules relevant for this configuration
     strongmols=['H2O', 'O3', 'CH4']
     intermediatemols=['C2H6', 'HCl']
     weakmols=['N2O', 'OCS']
     absentmols=['CO2','CO', 'O2', 'N2']

     ;WIDGET_CONTROL, state.column_scaleh2o[1], SET_VALUE='1.'
  ENDIF

  IF state.config EQ 'Lp1' THEN BEGIN
     ;molecules relevant for this configuration
     strongmols=['H2O', 'O3', 'CH4', 'N2O']
     intermediatemols=['C2H6', 'HCl']
     weakmols=['CO2', 'OCS']
     absentmols=['CO', 'O2', 'N2']

     ;WIDGET_CONTROL, state.column_scaleh2o[1], SET_VALUE='1.'
  ENDIF

  IF state.config EQ 'Lp2' THEN BEGIN
     ;molecules relevant for this configuration
     strongmols=['H2O', 'CH4', 'N2O']
     intermediatemols=['CO2', 'O3', 'HCl']
     weakmols=['none']
     absentmols=['CO','OCS', 'O2', 'C2H6', 'N2']

     ;WIDGET_CONTROL, state.column_scaleh2o[1], SET_VALUE='1.'
  ENDIF

  IF state.config EQ 'Lp3' THEN BEGIN
     ;molecules relevant for this configuration
     strongmols=['H2O', 'CH4', 'N2O']
     intermediatemols=['CO2', 'N2']
     weakmols=['none']
     absentmols=['CO','O3','OCS', 'O2', 'C2H6', 'HCl']

     ;WIDGET_CONTROL, state.column_scaleh2o[1], SET_VALUE='1.'
  ENDIF

  IF state.config EQ 'Lp4' THEN BEGIN
     ;molecules relevant for this configuration
     strongmols=['H2O', 'CH4', 'N2O']
     intermediatemols=['CO2', 'N2']
     weakmols=['none']
     absentmols=['CO','O3','OCS', 'O2', 'C2H6', 'HCl']

     ;WIDGET_CONTROL, state.column_scaleh2o[1], SET_VALUE='1.'
  ENDIF

  IF state.config EQ 'M1' OR state.config EQ 'M2' THEN BEGIN
     ;molecules relevant for this configuration
     strongmols=['H2O', 'CO2', 'N2O', 'O3', 'CO']
     intermediatemols=['OCS']
     weakmols=['CH4', 'N2']
     absentmols=['C2H6', 'HCl', 'O2']

     ;WIDGET_CONTROL, state.column_scaleh2o[1], SET_VALUE='1.'
  ENDIF
  
  ;...and produce the molecule part of the widget
  make_mols_widget, strongmols, intermediatemols, weakmols, absentmols, state
  
  state.continue=1
  
end
;
;===============================================================================
;
pro xtellcor_loadflatspec,state

  if state.continue lt 3 then begin
     
     ok = dialog_message('Previous steps not complete.',/ERROR,$
                         DIALOG_PARENT=state.xtellcor_base)
     return
     
  endif

;  Get user inputs

  flatfile = mc_cfld(state.flatspectra_fld,7,/EMPTY,CANCEL=cancel)
  if cancel then return
  flatfile = mc_cfile(state.flatpath+flatfile,WIDGET_ID=state.xtellcor_base, $
                 CANCEL=cancel)
  if cancel then return
  
  mc_readspec,flatfile,flatspec,flathdr,flatobsmode,start,stop,flatnorders,flatnaps, $
              flatorders,flatxunits,flatyunits,slith_pix,slith_arc,slitw_pix, $
              slitw_arc,flatrp,flatairmass,CANCEL=cancel
  if cancel then return  

  *state.flatspec=flatspec

  ;;check if order column of flatspec increases in same direction as objorders
  ;;(longer wavelengths have lower orders)
  ;FOR i=0,n_elements(*state.objorders)-1, 1 DO BEGIN
  ;  print, 'median wavelength: ', median((*state.flatspec)[*,0,i]), ' order: ', (*state.objorders)[i]
  ;ENDFOR
  ;stop 
  
  ;Find wavelength of the maximum blaze response, which tends to be a
  ;bit off the theoretical value of lambda=k/order with k a grating
  ;constant. This is critical for deriving blaze curves to divide over
  ;the data.

  state.config=STRTRIM(sxpar(flathdr,'MODE'),2)
  output=ishell_derive_blaze_peaks(*state.flatspec, *state.objorders, state.config, plot=state.objpath+'xtellcor_model_blazepeaks_'+state.config+'_'+strmid(state.objfile,0,strpos(state.objfile,'.fi'))+'.ps', dir=state.auxdir)
  *state.blazepeaks=output[1,*]
  *state.alpha=(output[2,*])[*]

  state.continue=3.1
  
end
;
;===============================================================================
;
pro xtellcor_blazedivide_noshift,state

  ;divide atmosphere-corrected spectrum over blaze function
  
  print, 'xtellcor_blazedivide_noshift'

  ;define array for blaze-, and telluric=corrected spectrum
  *state.blazecorspec=*state.corspec
  
  ;divide each observed echelle order over the same blaze function on
  ;the blaze x-axis scale

  n_orders=n_elements(*state.objorders)
  FOR i=0,n_orders-1,1 DO BEGIN
    w=(*state.corspec)[*,0,i]
    f=(*state.corspec)[*,1,i]
    e=(*state.corspec)[*,2,i]    
    x=!PI*(*state.alpha)[i]*(*state.objorders)[i]*(1.-(*state.blazepeaks)[i]/w)
    b=interpol((*state.blazespec)[1,*], (*state.blazespec)[0,*], x)    
    (*state.blazecorspec)[*,1,i]=f/b 
    (*state.blazecorspec)[*,2,i]=e/b   
  ENDFOR
  
end
;
;===============================================================================
;
pro xtellcor_flatblaze,state

  if state.continue lt 3.1 then begin
     
     ok = dialog_message('Previous steps not complete.',/ERROR,$
                         DIALOG_PARENT=state.xtellcor_base)
     return
     
  endif
  
 ;derive blaze spectra to be divided over the data, using the
 ;wavelengths of peak blaze response derived in xtellcor_loadflatspec
 ;(and ishell_derive_blaze_peaks.pro)
  *state.blazespec=ishell_derive_flatblaze(*state.flatspec, *state.blazepeaks, $
                                           *state.objorders, *state.alpha*1.00, state.config, $
                                           plot=state.objpath+'xtellcor_model_blazemedian_'+state.config+'_'+strmid(state.objfile,0,strpos(state.objfile,'.fi'))+'.ps', $
                                           write=state.objpath+'ishell_medianblaze_'+state.config+'_'+strmid(state.objfile,0,strpos(state.objfile,'.fi'))+'.dat', $
                                           dir=state.auxdir)

  ;actual divison of atmospheric-corrected spectrum over blaze function
  xtellcor_blazedivide_noshift, state
  
  state.continue=4
  
end
;
;===============================================================================
;
pro xtellcor_storedblaze,state

  if state.continue lt 3.1 then begin
     
     ok = dialog_message('Previous steps not complete.',/ERROR,$
                         DIALOG_PARENT=state.xtellcor_base)
     return
     
  endif

  print, 'xtellcor_storedblaze'

  ;*state.blazespec=ishell_read_flatblaze(*state.flatspec, *state.blazepeaks, $
  ;                                       *state.objorders, config)

  openr,1, state.auxdir+'ishell_medianblaze_K3_extractedstandard.dat'  
  ;first read 2 header lines
  dumstr='dum'
  readf, 1, dumstr
  readf, 1, dumstr
  ;then read the data
  dum=[0.,0.]
  x=[-100.]
  b=[-100.]
  while not eof(1) DO BEGIN
      readf, 1, dum
      x=[x,dum[0]]
      b=[b,dum[1]]     
  endwhile
  close, 1
  ind=where(x NE -100.)
  x=x[ind]
  b=b[ind]  
  arr=FLTARR(2,n_elements(x))
  arr[0,*]=x
  arr[1,*]=b  
  *state.blazespec=arr
  
  ;actual divison of atmospheric-corrected spectrum over blaze function
  xtellcor_blazedivide_noshift, state

  state.continue=4
  
end
;
;==============================
;
pro xtellcor_telldivide,state

;  Produce atmosphere-corrected object spectrum, i.e.,
;  wavelength-shift the observed spectrum by the amount determined by
;  the user, and then divide over the atmospheric model.
;
;  Note: add a 5th column with the atmospheric model used  

  ;create array for atmosphere-corrected spectrum
  *state.corspec  = *state.objspec

  ;loop over echelle orders
  for i = 0, n_elements(*state.objorders)-1 do begin

     ;loop over extraction apertures
     for j = 0, state.objnaps-1 do begin
        
        ;index for this order and this aperture
        k = i*state.objnaps+j
        
        ;  Interpolate telluric model spectrum onto object wavelength sampling.        
        mc_interpspec, (*state.tellspec)[*,0,k], (*state.tellspec)[*,1,k], $
                   (*state.objspec)[*,0,k], tellflux, CANCEL=cancel
        if cancel then return

        ;  Now shift observed spectrum on pixel scale (this assumes
        ;  any shift is due to wavelength calibration error of the
        ;  observation, not in the atmospheric model). Note that shift
        ;  determined by shifting model, but now we apply it to
        ;  observation in the opposite direction, so, "x-..."
        x = findgen(n_elements((*state.objspec)[*,0,k]))
        mc_interpspec, x-(*state.shift)[i,j], (*state.objspec)[*,1,k], x, objflux, objerror, $
                      IYERROR=(*state.objspec)[*,2,k], CANCEL=cancel
        if cancel then return

        ;  ....and shift flags too
        mc_interpflagspec, x-(*state.shift)[i,j], byte((*state.objspec)[*,2,k]), x, objflag, $
                          CANCEL=cancel
        if cancel then return

        ; ...and divide shifted object over telluric model
        ;    remembering that tellfux=1./atmospheric_absorption
        ;    as defined in xtellcor_fitatmmodel.
        (*state.corspec)[*,1,k] = tellflux*objflux
        ; ...propagating error bars (but note atm model has no uncertainty)
        (*state.corspec)[*,2,k] = tellflux * objerror
        ; ...and add flags
        (*state.corspec)[*,3,k] = objflag

        ;....TBD...what shall we put in 5th column?
        ;(*state.corspec)[*,4,k] = 1./ntellflux
        
     endfor
     
  endfor

  
end
;
;===============================================================================
;
pro xtellcor_getshifts,state

  if state.continue lt 2 then begin
     
     ok = dialog_message('Previous steps not complete.',/ERROR,$
                         DIALOG_PARENT=state.xtellcor_base)
     return
     
  endif

  shifts = xmc_findshifts_model(*state.objspec,*state.tellspec,*state.objorders, $
                          *state.awave,*state.atrans,state.xtitle,CANCEL=cancel)
  if ~cancel then begin

     *state.shift = shifts
     state.continue=3

     ;divide shifted object over atmospheric model, and put in 'corspec'
     xtellcor_telldivide,state     
  endif
  
end
;
;===============================================================================
;
pro xtellcor_blaze_skip,state

  if state.continue lt 3. then begin
     
     ok = dialog_message('Previous steps not complete.',/ERROR,$
                         DIALOG_PARENT=state.xtellcor_base)
     return
     
  endif

  *state.blazecorspec = *state.corspec

  state.continue=4  
  
end
;
;===============================================================================
;
pro xtellcor_writefile,state

; 10/18/18 A. Boogert: output fits file _tellspec no longer optional
  
  if state.continue lt 4 then begin
     
     ok = dialog_message('Previous steps not complete.',/ERROR,$
                         DIALOG_PARENT=state.xtellcor_base)
     return
     
  endif
  
  obj = mc_cfld(state.objoname_fld,7,/EMPTY,CANCEL=cancel)
  if cancel then return
  
;
;  Write the telluric correction spectrum to disk
;
  
;  Get info from obj hdr.
     
  orders    = fxpar(*state.objhdr,'ORDERS',COMMENT=corders)
  norders   = fxpar(*state.objhdr,'NORDERS',COMMENT=cnorders)
  obsmode   = fxpar(*state.objhdr,'MODE',COMMENT=cobsmode)
  xunits    = fxpar(*state.objhdr,'XUNITS',COMMENT=cxunits)
  xtitle    = fxpar(*state.objhdr,'XTITLE',COMMENT=cxtitle,COUNT=count)
     
;  For backwards compatibility
     
  if count eq 0 then begin
        
     xtitle  = '!7k!5 ('+strtrim(xunits,2)+')'
     cxtitle = 'IDL X Title'
        
  endif
     
;  Create hdr for the telluric correction spectrum
     
  fxhmake, hdr,*state.tellspec
  fxaddpar,hdr,'FILENAME',strtrim(obj+'_tellspec.fits',2)
  fxaddpar,hdr,'CREPROG','xtellcor_model', ' Creation IDL program'  
  fxaddpar,hdr,'ORDERS',orders,corders
  fxaddpar,hdr,'NORDERS',norders,cnorders
  fxaddpar,hdr,'NAPS',1, ' Number of apertures'
  fxaddpar,hdr,'AM',state.objairmass, 'Average airmass'
  fxaddpar,hdr,'TELMETH',strtrim(state.method,2), $
                          ' Telluric correction method'
  fxaddpar,hdr,'XUNITS',xunits,cxunits
  fxaddpar,hdr,'XTITLE',xtitle,cxtitle
  fxaddpar,hdr,'YUNITS',strcompress(state.units,/RE)+'/DNs-1', $
                          ' Units of the Y axis'
  fxaddpar,hdr,'YTITLE',state.nytitle+' / DN s!U-1!N)','IDL Y title'

;Add details of the telluric model
  ;column density scaling factors
  FOR m=0,n_elements(state.molecules)-1 DO BEGIN
     fxaddpar,hdr,'SC_'+STRTRIM((state.molecules)[m],2), $
              (state.best_scale_molecules)[m], $
              ' '+STRTRIM((state.molecules)[m],2)+' column scaling factor'
  ENDFOR

  ;origin of model spectrum (SPG, local, etc.)
  fxaddpar,hdr,'ATMMOD', strtrim(state.server,2), $
                          ' Origin atmospheric model'

  ;resolving power of model spectrum
  fxaddpar,hdr,'ATMRES', state.resolving_power, $
                          ' Resolving power atmospheric model'
     
;Add description of processing history

  history = 'atmospheric model used to correct for telluric lines in '+state.objpath+obj+'.fits'
  history = mc_splittext(history,68,CANCEL=cancel)
  if cancel then return
  sxaddhist,history,hdr

;Add description of columns
  ;total transmission
  fxaddpar,hdr,'COL_2', $ 
           'All', $
           ' total transm in a[*,1,*]'
  ;transmission per molecule
  IF tag_exist(state, 'psg_molecules') THEN BEGIN
    FOR m=0,n_elements(state.psg_molecules)-1 DO BEGIN
       fxaddpar,hdr,'COL_'+STRTRIM(m+5,2), $ ;take into account 1st molecule starts at column 5
                STRTRIM((state.psg_molecules)[m],2), $
                ' transm '+STRTRIM((state.psg_molecules)[m],2)+' in a[*,'+STRTRIM(m+4,2)+',*]'
    ENDFOR
  ENDIF
     
;  Write it  to disk

  ;  Make sure it's all transmission, not 1/transmission
  tellspec_out=*state.tellspec
  tellspec_out[*,1,*]=1./tellspec_out[*,1,*]
     
  writefits,state.objpath+obj+'_tellspec.fits', tellspec_out,hdr

  print
  print,'Wrote the telluric correction spectrum to '+ $
        strtrim(state.objpath+obj,2)+'_tellspec.fits'
  print
  
;  Write the telluric corrected object spectrum to disk.  Note that
;  the *state.blazecorspec contains the spectrum divided by the model and
;  the blaze function, and this was done in the xtellcor_telldivide
;  step.
    
;Copy header from input data
  
  hdr = *state.objhdr

  hdrinfo = mc_gethdrinfo(*state.objhdr,state.keywords,/IGNOREMISSING,$
                          CANCEL=cancel)
  if cancel then return

  fxhmake,newhdr,*state.blazecorspec
  ntags = n_tags(hdrinfo.vals)
  names = tag_names(hdrinfo.vals)

  for i = 0, ntags-1 do begin

     if names[i] eq 'HISTORY' then begin
        sxaddhist,(hdrinfo.vals.(i)),newhdr
        continue

     endif
     
     for k = 0,n_elements(hdrinfo.vals.(i))-1 do begin
        
        if size((hdrinfo.vals.(i))[k],/TYPE) eq 7 and $
           strlen((hdrinfo.vals.(i))[k]) gt 68 then begin
           
           fxaddpar,newhdr,names[i],(hdrinfo.vals.(i))[k],(hdrinfo.coms.(i))[k]
           
        endif else sxaddpar,newhdr,names[i],(hdrinfo.vals.(i))[k], $
                            (hdrinfo.coms.(i))[k]
        
     endfor
     
  endfor

;  Update keywords
  
  sxaddpar,newhdr,'FILENAME',strtrim(obj+'.fits',2)
  sxaddpar,newhdr,'CREPROG','xtellcor_model', ' Creation program'  
  sxaddpar,newhdr,'YUNITS',strcompress(state.units,/RE), ' Units of the Y axis'
  fxaddpar,newhdr,'YTITLE',state.nytitle+')',' IDL Y title'
  
;  Add xtellcor_model keywords

  if state.userkeywords[0] ne '' then before = state.userkeywords[0]

  sxaddpar,newhdr,'TELMETH',strtrim(state.method,2), $
           ' Telluric correction method',BEFORE=before

  sxaddpar,newhdr,'Telfile',obj+'_tellspec.fits', $
           ' The telluric correction file',BEFORE=before

  ;column density scaling factors
  FOR m=0,n_elements(state.molecules)-1 DO BEGIN
     fxaddpar,newhdr,'SC_'+STRTRIM((state.molecules)[m],2), $
              (state.best_scale_molecules)[m], $
              ' '+STRTRIM((state.molecules)[m],2)+' column scaling factor'
  ENDFOR

  ;origin of model spectrum (SPG, local, etc.)
  fxaddpar,newhdr,'ATMMOD', strtrim(state.server,2), $
                          ' Origin atmospheric model'

  ;resolving power of model spectrum
  fxaddpar,newhdr,'ATMRES', state.resolving_power, $
                          ' Resolving power atmospheric model'
  
;  Write new xtellcor_model history

  history1='Spectrum of this target divided by telluric model spectrum '+ state.objpath+obj+'_tellspec.fits. '
  if total(*state.shift) ne 0 then begin
     for i = 0, state.objnaps-1 do begin
        
        history = history1+'Target spectrum for aperture '+ $
                  string(i+1,FORMAT='(i2.2)')+' were shifted by '+ $
                  strjoin(string((*state.shift)[*,i],FORMAT='(f+5.2)'),', ')+ $
                  ' pixels.  '

     endfor

  endif else history=history1+'No wavelength shift appplied for spectrum of this target.'
    
  sxaddhist,' ',newhdr
  sxaddhist,'######################## Xtellcor_Model History ' + $
            '#########################',newhdr
  sxaddhist,' ',newhdr
  history = mc_splittext(history,67,CANCEL=cancel)
  if cancel then return
  sxaddhist,history,newhdr

;  Write to disk
  
  writefits,state.objpath+obj+'.fits',*state.blazecorspec,newhdr
  xvspec,state.objpath+obj+'.fits',/PLOTLINMAX
  
  print
  print,'Wrote the corrected spectrum to '+strtrim(state.objpath+obj,2)+'.fits'
  print

end
;
;===============================================================================
;
; ------------------------------Main Program---------------------------------- 
;
;===============================================================================
;
pro xtellcor_model

;  Get spextool and instrument information 
  
  mc_getspextoolinfo,spextoolpath,packagepath,spextool_keywords,instrinfo, $
                     notirtf,version,CANCEL=cancel
  if cancel then return
  
;  Load color table

  mc_mkct
  
;  Set the fonts

  mc_getfonts,buttonfont,textfont,CANCEL=cancel
  if cancel then return

;  Build three structures which will hold important info.

  mode = (strlowcase(instrinfo.instrument) eq 'ishell') ? 1:0
  method = 'teltool'
         
  state = {airmass:'',$
           awave:ptr_new(2),$
           atrans:ptr_new(fltarr(2)),$
           continue:0.0,$
           instrument:instrinfo.instrument,$
           keywords:[spextool_keywords,instrinfo.xtellcor_keywords,'HISTORY'],$
           maxdev:0.0,$
           message:0L,$
           method:method,$
           column_scaleh2o: fltarr(2),$
           column_scalech4: fltarr(2),$
           column_scaleco2: fltarr(2),$           
           column_scaleco: fltarr(2),$
           column_scalen2o: fltarr(2),$
           column_scaleo3: fltarr(2),$
           column_scaleocs: fltarr(2),$                      
           column_scaleo2: fltarr(2),$                      
           column_scalec2h6: fltarr(2),$                      
           column_scalehcl: fltarr(2),$                      
           column_scalen2: fltarr(2),$
           column_fit: [0,0,0,0,0,0],$
           column_fitH2O: 0,$
           column_fitCH4: 0,$
           column_fitCO2: 0,$                      
           column_fitCO: 0,$
           column_fitN2O: 0,$
           column_fitO3: 0,$                      
           column_fitOCS: 0,$                      
           column_fitO2: 0,$
           column_fitC2H6: 0,$
           column_fitHCl: 0,$                                 
           column_fitN2: 0,$                                 
           mode:mode,$
           nytitle:'',$
           objhdr:ptr_new(fltarr(2)),$
           objorders:ptr_new(fltarr(2)),$
           objnaps:0,$
           objpath:'',$ 
           objairmass:0,$          
           objfile:'',$
           objobsmode:'',$                      
           objoname_fld:[0L,0L],$
           objspec:ptr_new(fltarr(2)),$
           objspectra_fld:[0,0],$
           flatspec:ptr_new(fltarr(2)),$
           flatpath:'',$
           flatspectra_fld:[0,0],$           
           config:'', $
           alpha:ptr_new(fltarr(2)), $ ;a grating constant
           output_bg:0L,$
           packagepath:packagepath,$
           rmsdev:0.0,$
           scale:0.,$
           scales:ptr_new(fltarr(2)),$
           shift:ptr_new(fltarr(2)),$
           blazeshift:ptr_new(fltarr(2)),$
           blazepeaks:ptr_new(fltarr(2)),$           
           shiftobjap:0,$
           slitw_arc:0.,$
           slitw_pix:0.,$
           spextoolpath:spextoolpath,$
           objdisp:ptr_new(fltarr(2)),$
           tellspec:ptr_new(fltarr(2)),$
           blazespec:ptr_new(fltarr(2)),$
           blazecorspec:ptr_new(fltarr(2)),$           
           corspec:ptr_new(fltarr(2)),$                      
           units:'ergs s-1 cm-2 A-1',$
           userkeywords:instrinfo.xtellcor_keywords,$
           xtellcor_base:0L,$
           col2_base:0L,$           $
           col4_base:0L,$           $           
           xtitle:'',$
           ytitle:'', $
           box_base_mols:0L, $
           col1_base:0L, $
           server:'', $
           resolving_power:0., $
           auxdir:'', $
           molecules:strarr(11), $
           best_scale_molecules:fltarr(11) $
           ;psg_molecules ... defined in code
          }

  state.mode = 0
  
  title = 'Xtellcor_model '+version+' for '+state.instrument
  
  state.xtellcor_base = widget_base(TITLE=title,$
                                      EVENT_PRO='xtellcor_model_event',$
                                      /COLUMN)
  
     button = widget_button(state.xtellcor_base,$
                            FONT=buttonfont,$
                            VALUE='Done',$
                            UVALUE='Quit')
     
     message = widget_draw(state.xtellcor_base,$
                           FRAME=2,$
                           XSIZE=1,$
                           YSIZE=25)
     
     row_base = widget_base(state.xtellcor_base,$
                            /ROW)

        ;LEFT COLUMN
     
        col1_base = widget_base(row_base,$
                                /COLUMN)

           ;START OF BOX 1
        
           box1_base = widget_base(col1_base,$
                                   /COLUMN,$
                                   FRAME=2)

              label = widget_label(box1_base,$
                                   VALUE='1.  Load Spectrum',$
                                   FONT=buttonfont,$
                                   /ALIGN_LEFT)

              row = widget_base(box1_base,$
                                /ROW,$
                                /BASE_ALIGN_CENTER)
              
                 buton = widget_button(row,$
                                       FONT=buttonfont,$
                                       VALUE='Obj Spectrum',$
                                       UVALUE='Object Spectra Button')
                 
                 fld = coyote_field2(row,$
                                     LABELFONT=buttonfont,$
                                     FIELDFONT=textfont,$
                                     TITLE=':',$
                                     UVALUE='Object Spectra Field',$
;                                     VALUE='cspectra61-70.fits',$
                                     XSIZE=18,$
                                     /CR_ONLY,$
                                     TEXTID=textid)
                 state.objspectra_fld = [fld,textid]
                 
              load = widget_button(box1_base,$
                                   VALUE='Load Spectrum',$
                                   FONT=buttonfont,$
                                   UVALUE='Load Spectra')

              ;END OF BOX 1
              
              ;START OF BOX 2--Atmospheric model
              ;loadng
              
              box_base_mols = widget_base(col1_base,$
                                  COLUMN=1, $
                                  FRAME=2)
           
              label = widget_label(box_base_mols,$
                                   VALUE='2. Get Atmospheric Model',$
                                   FONT=buttonfont,$
                                   /ALIGN_LEFT)

              label = widget_label(box_base_mols,$
                                   VALUE='Load spectrum to see relevant molecules',$
                                   FONT=buttonfont,$
                                   /ALIGN_LEFT)


              
              state.box_base_mols=box_base_mols

              state.col1_base=col1_base
              
              ;END OF BOX 2
              
        ;END OF LEFT COLUMN

        ;START OF RIGHT COLUMN
              
        col2_base = widget_base(row_base,$
                                /COLUMN)

           ;START OF BOX 4
              
           box_base = widget_base(col2_base,$
                                   /COLUMN,$
                                   FRAME=2)
           
              label = widget_label(box_base,$
                                   VALUE='3. Plot Model and Determine Shifts',$
                                   FONT=buttonfont)
                          
              shift = widget_button(box_base,$
                                    VALUE='Find Shifts',$
                                    UVALUE='Get Shifts',$
                                    FONT=buttonfont)

           ;END OF BOX 4              

           ;START OF BOX 5
              
           box_base = widget_base(col2_base,$
                                   /COLUMN,$
                                   FRAME=2)
           
              label = widget_label(box_base,$
                                   VALUE='4.  Correct for Order Blaze Shape',$
                                   FONT=buttonfont, $
                                   /ALIGN_LEFT)

              shift = widget_button(box_base,$
                                    VALUE='Skip. Continue to Step 5',$
                                    UVALUE='Skip Correct Blaze',$
                                    FONT=buttonfont, $
                                    /ALIGN_LEFT)
              
               row = widget_base(box_base,$
                                /ROW,$
                                /BASE_ALIGN_CENTER)
                           

                 buton = widget_button(row,$
                                       FONT=buttonfont,$
                                       VALUE='Flat Spectrum',$
                                       UVALUE='Flat Spectrum Button')
                 
                 fld = coyote_field2(row,$
                                     LABELFONT=buttonfont,$
                                     FIELDFONT=textfont,$
                                     TITLE=':',$
                                     UVALUE='Flat Spectrum Field',$
;                                     VALUE='cspectra61-70.fits',$
                                     XSIZE=18,$
                                     /CR_ONLY,$
                                     TEXTID=textid)
                 state.flatspectra_fld = [fld,textid]
                 
              load = widget_button(box_base,$
                                   VALUE='Load Extracted Flat Spectrum',$
                                   FONT=buttonfont,$
                                   UVALUE='Load Flat Spectra')

              label = widget_label(box_base,$
                                   VALUE='Choose one of:',$
                                   FONT=buttonfont, $
                                    /ALIGN_LEFT)

              
              shift = widget_button(box_base,$
                                    VALUE='a) Use Flat Field Blaze Profile',$
                                    UVALUE='Use Flat Field Blaze Profile',$
                                    FONT=buttonfont, $
                                    /ALIGN_LEFT)

              shift = widget_button(box_base,$
                                    VALUE='b) Use Stored Blaze Profile',$
                                    UVALUE='Use Stored Blaze Profile',$
                                    FONT=buttonfont, $
                                    /ALIGN_LEFT)

           ;END OF BOX 5              

           ;START OF BOX 6

              
           box_base = widget_base(col2_base,$
                                  /COLUMN,$
                                  FRAME=2)
           
              label = widget_label(box_base,$
                                   VALUE='5.  Plot Result and Write to File',$
                                   FONT=buttonfont,$
                                   /ALIGN_LEFT)
              
              oname = coyote_field2(box_base,$
                                    LABELFONT=buttonfont,$
                                    FIELDFONT=textfont,$
                                    TITLE='File Name :',$
                                    UVALUE='Output File Name',$
                                    xsize=18,$
                                    textID=textid)
              state.objoname_fld = [oname,textid]
              
              write = widget_button(box_base,$
                                    VALUE='Plot and Write',$
                                    UVALUE='Write File',$
                                    FONT=buttonfont)

           ;END OF BOX 6                          
              
  ;END OF RIGHT COLUMN
              
  help = widget_button(state.xtellcor_base,$
                       VALUE='Help',$
                       UVALUE='Help',$
                       FONT=buttonfont)
  
; Get things running.  Center the widget using the Fanning routine.
            
  cgcentertlb,state.xtellcor_base
  widget_control, state.xtellcor_base, /REALIZE
  widget_control, message, GET_VALUE=x
  state.message = x

 widget_geom = widget_info(state.xtellcor_base, /GEOMETRY)
 widget_control, message, XSIZE=widget_geom.xsize-17
 erase, color=196
 
; Start the Event Loop. This will be a non-blocking program.

  XManager, 'xtellcor_model', $
            state.xtellcor_base, $
            CLEANUP='xtellcor_cleanup',$
            /NO_BLOCK
  
  widget_control, state.xtellcor_base, SET_UVALUE=state, /NO_COPY

end
